@@ -28,6 +28,7 @@ QUIET=''
 # link jemalloc (needed for tokudb). For now pxc doesn't build tokudb
 # if in future pxc supports tokudb we will need it.
 WITH_JEMALLOC=''
+DEBUG_EXTRA=''
 
 # Galera can be build locally as part of the script if the directory is placed
 # at said location. Alternatively, skip building galera instead just copy
@@ -35,9 +36,6 @@ WITH_JEMALLOC=''
 # script will install it at said location while making file tar.gz.
 COPYGALERA=0
 
-# suffix to add in case of debug build to tar.gz.
-DEBUG_EXTNAME=''
-
 WITH_MECAB_OPTION=''
 
 # build with ssl. configuration related to ssl.
@@ -49,7 +47,6 @@ CRYPTO_LIBRARY=''
 GALERA_SSL=''
 SSL_OPT=''
 STAG=${STAG:-}
-DEBUG_EXTRA=''
 
 # tag build with this name.
 TAG='1'
@@ -99,8 +96,8 @@ do
         shift
         CMAKE_BUILD_TYPE='Debug'
         BUILD_COMMENT="${BUILD_COMMENT:-}-debug"
-        DEBUG_EXTNAME="-DDEBUG_EXTNAME=OFF"
-        SCONS_ARGS+=' debug=1'
+        DEBUG_EXTRA="-DDEBUG_EXTNAME=ON"
+        SCONS_ARGS+=' debug=0'
         ;;
     -G | --copygalera )
         shift
@@ -298,32 +295,18 @@ if [[ $ENABLE_ASAN -eq 1 ]]; then
     fi
 fi
 
-#
-# TokuDB cmake flags [IGNORED FOR PXC]
-if test -d "$SOURCEDIR/storage/tokudb"
-then
-    CMAKE_OPTS="${CMAKE_OPTS:-} -DBUILD_TESTING=OFF -DUSE_GTAGS=OFF -DUSE_CTAGS=OFF -DUSE_ETAGS=OFF -DUSE_CSCOPE=OFF -DTOKUDB_BACKUP_PLUGIN_VERSION=${TOKUDB_BACKUP_VERSION}"
-
-    if test "x$CMAKE_BUILD_TYPE" != "xDebug"
-    then
-        CMAKE_OPTS="${CMAKE_OPTS:-} -DTOKU_DEBUG_PARANOID=OFF"
-    else
-        CMAKE_OPTS="${CMAKE_OPTS:-} -DTOKU_DEBUG_PARANOID=ON"
-    fi
-
-    if [[ $CMAKE_OPTS == *WITH_VALGRIND=ON* ]]
-    then
-        CMAKE_OPTS="${CMAKE_OPTS:-} -DUSE_VALGRIND=ON"
-    fi
-fi
-
-#
-
 COMMON_FLAGS="-DPERCONA_INNODB_VERSION=$PERCONA_SERVER_EXTENSION"
 export CFLAGS=" $COMMON_FLAGS -static-libgcc $MACHINE_SPECS_CFLAGS ${CFLAGS:-}"
 export CXXFLAGS=" $COMMON_FLAGS $MACHINE_SPECS_CFLAGS ${CXXFLAGS:-}"
 export MAKE_JFLAG="${MAKE_JFLAG:--j$PROCESSORS}"
 
+export DEBIAN_VERSION="$(lsb_release -sc)"
+echo $DEBIAN_VERSION
+if [[ "$CMAKE_BUILD_TYPE" == "Debug" ]] && ([[ "$DEBIAN_VERSION" == "yakkety" ]] || [[ "$DEBIAN_VERSION" == "zesty" ]]); then
+    export CFLAGS=" $CFLAGS -fno-strict-aliasing -Wno-unused-parameter -Wno-sign-compare -Wno-error=deprecated-declarations -Wno-error=nonnull-compare -Wno-error=shift-negative-value -Wno-error=misleading-indentation -Wno-error=literal-suffix -Wno-error=virtual-move-assign"
+    export CXXFLAGS=" $CFLAGS -fno-strict-aliasing -Wno-unused-parameter -Wno-sign-compare -Wno-error=deprecated-declarations -Wno-error=nonnull-compare -Wno-error=shift-negative-value -Wno-error=misleading-indentation -Wno-error=literal-suffix -Wno-error=virtual-move-assign"
+fi
+
 #
 # Test jemalloc directory
 if test "x$WITH_JEMALLOC" != "x"
@@ -398,7 +381,6 @@ fi
         cmake ../../ ${CMAKE_OPTS:-} -DBUILD_CONFIG=mysql_release \
             -DCMAKE_BUILD_TYPE=Debug \
             $DEBUG_EXTRA \
-            -DDEBUG_EXTNAME=ON \
             -DWITH_EMBEDDED_SERVER=OFF \
             -DFEATURE_SET=community \
             -DENABLE_DTRACE=OFF \
@@ -417,16 +399,16 @@ fi
             -DWITH_INNODB_MEMCACHED=ON \
             -DDOWNLOAD_BOOST=1 \
             -DWITH_BOOST="$TARGETDIR/libboost" \
+            -DWITH_SCALABILITY_METRICS=ON \
             $WITH_MECAB_OPTION $OPENSSL_INCLUDE $OPENSSL_LIBRARY $CRYPTO_LIBRARY
 
         (make $MAKE_JFLAG $QUIET) || exit 1
         (make install) || exit 1
-        cp -v sql/mysqld-debug $TARGETDIR/usr/local/$PRODUCT_FULL_NAME/bin/mysqld
+        (cp -v sql/mysqld-debug $TARGETDIR/usr/local/$PRODUCT_FULL_NAME/bin/mysqld) || true
         echo "mysqld in build in debug mode"
     else
         cmake ../../ ${CMAKE_OPTS:-} -DBUILD_CONFIG=mysql_release \
             -DCMAKE_BUILD_TYPE=${CMAKE_BUILD_TYPE:-RelWithDebInfo} \
-            $DEBUG_EXTNAME \
             -DWITH_EMBEDDED_SERVER=OFF \
             -DFEATURE_SET=community \
             -DENABLE_DTRACE=OFF \
@@ -444,6 +426,7 @@ fi
             -DWITH_INNODB_MEMCACHED=ON \
             -DDOWNLOAD_BOOST=1 \
             -DWITH_BOOST="$TARGETDIR/libboost" \
+            -DWITH_SCALABILITY_METRICS=ON \
             $WITH_MECAB_OPTION $OPENSSL_INCLUDE $OPENSSL_LIBRARY $CRYPTO_LIBRARY
 
         (make $MAKE_JFLAG $QUIET) || exit 1
