@@ -102,6 +102,11 @@ sub fix_port {
   return $self->{PORT}++;
 }
 
+sub fix_x_port {
+  my ($self, $config, $group_name, $group)= @_;
+  return $self->{ARGS}->{mysqlxbaseport}++;
+}
+
 sub fix_host {
   my ($self)= @_;
   'localhost'
@@ -140,6 +145,13 @@ sub fix_server_id {
   return $server_id;
 }
 
+sub fix_x_socket {
+  my $result = fix_socket(@_);
+  $result =~ s/mysqld\.([0-9]+)\.sock$/mysqlx\.$1\.sock/;
+
+  return $result;
+}
+
 sub fix_socket {
   my ($self, $config, $group_name, $group)= @_;
   # Put socket file in tmpdir
@@ -154,6 +166,7 @@ sub fix_socket {
     my $port = $group->value('port');
     $socket = "$dir/mysqld-$port.sock"; 
   }
+
   return $socket;
 }
 
@@ -261,6 +274,8 @@ my @mysqld_rules=
  { '#ist_port' => \&fix_port },
  { '#sst_port' => \&fix_port },
  { 'socket' => \&fix_socket },
+ { 'loose-mysqlx-port' => \&fix_x_port },
+ { 'loose-mysqlx-socket' => \&fix_x_socket },
  { '#log-error' => \&fix_log_error },
  { 'general-log' => 1 },
  { 'plugin-dir' => sub { $::plugindir } },
@@ -278,6 +293,7 @@ my @mysqld_rules=
  { 'loose-sha256_password_auto_generate_rsa_keys' => "0"},
   );
 
+
 if (IS_WINDOWS)
 {
   # For simplicity, we use the same names for shared memory and 
@@ -504,11 +520,21 @@ sub resolve_at_variable {
     $after = $';
     chop($group_name);
 
-  my $from_group= $config->group($group_name)
-    or croak "There is no group named '$group_name' that ",
-      "can be used to resolve '$option_name' for test '$self->{testname}'";
+    $group_name =~ s/^\@//; # Remove at
+    my $value;
+
+    if ($group_name =~ "env")
+    {
+      $value = $ENV{$option_name};
+    }
+    else
+    {
+      my $from_group= $config->group($group_name)
+        or croak "There is no group named '$group_name' that ",
+          "can be used to resolve '$option_name'";
+      $value= $from_group->value($option_name);
+    }
 
-    my $value= $from_group->value($option_name);
     $res .= $before.$value;
   }
   $res .= $after;
