@@ -868,10 +868,8 @@ sub collect_one_test_case {
   # ----------------------------------------------------------------------
   # Check for replicaton tests
   # ----------------------------------------------------------------------
-  if ( $suitedir =~ 'rpl' )
-  {
-    $tinfo->{'rpl_test'}= 1;
-  }
+  $tinfo->{'rpl_test'}= 1 if ($suitename =~ 'rpl');
+  $tinfo->{'grp_rpl_test'}= 1 if ($suitename =~ 'group_replication');
 
   # ----------------------------------------------------------------------
   # Check for disabled tests
@@ -1059,7 +1057,7 @@ sub collect_one_test_case {
     push(@{$tinfo->{'slave_opt'}}, "--loose-skip-log-bin");
   }
 
-  if ( $tinfo->{'rpl_test'} )
+  if ( $tinfo->{'rpl_test'} or $tinfo->{'grp_rpl_test'} )
   {
     if ( $skip_rpl )
     {
@@ -1098,6 +1096,18 @@ sub collect_one_test_case {
     }
   }
 
+  # Check for group replication tests
+  if ( $tinfo->{'grp_rpl_test'} )
+  {
+    $::group_replication= 1;
+  }
+
+  # Check for xplugin tests
+  if ( $tinfo->{'xplugin_test'} )
+  {
+    $::xplugin= 1;
+  }
+
   if ( $tinfo->{'not_windows'} && IS_WINDOWS )
   {
     $tinfo->{'skip'}= 1;
@@ -1188,6 +1198,12 @@ my @tags=
  ["include/have_ssl.inc", "need_ssl", 1],
  ["include/have_ssl_communication.inc", "need_ssl", 1],
  ["include/not_windows.inc", "not_windows", 1],
+
+ # Tests with below .inc file are considered to be group replication tests
+ ["have_group_replication_plugin_base.inc", "grp_rpl_test", 1],
+
+ # Tests with below .inc file are considered to be xplugin tests
+ ["include/have_mysqlx_plugin.inc", "xplugin_test", 1],
 );
 
 
