@@ -58,10 +58,14 @@ apply_trx_ws(void*                    recv_ctx,
 
                 if (err > 0)
                 {
+                    /* It is safe to call commit_cb w/o entering commit monitor
+                    as callback is with commit = false there-by invoking
+                    rollback action. */
                     wsrep_bool_t unused(false);
                     wsrep_cb_status const rcode(
                         commit_cb(
                             recv_ctx,
+                            NULL,
                             TrxHandle::trx_flags_to_wsrep_flags(trx.flags()),
                             &meta,
                             &unused,
@@ -143,6 +147,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
                              config_.get(Param::commit_order))),
     state_file_         (config_.get(BASE_DIR)+'/'+GALERA_STATE_FILE),
     st_                 (state_file_),
+    safe_to_bootstrap_  (true),
     trx_params_         (config_.get(BASE_DIR), -1,
                          KeySet::version(config_.get(Param::key_format)),
                          gu::from_string<int>(config_.get(
@@ -159,6 +164,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     unordered_cb_       (args->unordered_cb),
     sst_donate_cb_      (args->sst_donate_cb),
     synced_cb_          (args->synced_cb),
+    abort_cb_           (args->abort_cb),
     sst_donor_          (),
     sst_uuid_           (WSREP_UUID_UNDEFINED),
     sst_seqno_          (WSREP_SEQNO_UNDEFINED),
@@ -181,7 +187,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     ist_prepared_       (false),
     ist_senders_        (gcs_, gcache_),
     wsdb_               (),
-    cert_               (config_, service_thd_),
+    cert_               (config_, service_thd_, gcache_),
 #ifdef HAVE_PSI_INTERFACE
     local_monitor_      (WSREP_PFS_INSTR_TAG_LOCAL_MONITOR_MUTEX,
                          WSREP_PFS_INSTR_TAG_LOCAL_MONITOR_CONDVAR),
@@ -216,6 +222,15 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
 #endif /* HAVE_PSI_INTERFACE */
     wsrep_stats_        ()
 {
+    /*
+      Register the application callback that should be called
+      if the wsrep provider will teminated abnormally:
+    */
+    if (abort_cb_)
+    {
+        gu_abort_register_cb(abort_cb_);
+    }
+
     // @todo add guards (and perhaps actions)
     state_.add_transition(Transition(S_CLOSED,  S_DESTROYED));
     state_.add_transition(Transition(S_CLOSED,  S_CONNECTED));
@@ -239,6 +254,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     state_.add_transition(Transition(S_JOINED, S_CLOSING));
     state_.add_transition(Transition(S_JOINED, S_CONNECTED));
     state_.add_transition(Transition(S_JOINED, S_SYNCED));
+    // the following is possible if one desync() immediately follows another
     state_.add_transition(Transition(S_JOINED, S_DONOR));
 
     state_.add_transition(Transition(S_SYNCED, S_CLOSING));
@@ -254,7 +270,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     wsrep_uuid_t  uuid;
     wsrep_seqno_t seqno;
 
-    st_.get (uuid, seqno);
+    st_.get (uuid, seqno, safe_to_bootstrap_);
 
     if (0 != args->state_id &&
         args->state_id->uuid != WSREP_UUID_UNDEFINED &&
@@ -273,13 +289,16 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     // information provided on startup:
 
     update_state_uuid (uuid, seqno);
+    gcache_.seqno_reset(to_gu_uuid(uuid), seqno);
+    // update gcache position to one supplied by app.
 
     cc_seqno_ = seqno; // is it needed here?
-    apply_monitor_.set_initial_position(seqno);
 
+    // the following initialization is needed only to pass seqno to
+    // connect() call. Ideally this should be done only on receving conf change.
+    apply_monitor_.set_initial_position(seqno);
     if (co_mode_ != CommitOrder::BYPASS)
         commit_monitor_.set_initial_position(seqno);
-
     cert_.assign_initial_position(seqno, trx_proto_ver());
 
     build_stats_vars(wsrep_stats_);
@@ -287,7 +306,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
 
 galera::ReplicatorSMM::~ReplicatorSMM()
 {
-    log_info << "dtor state: " << state_();
+    log_debug << "dtor state: " << state_();
     switch (state_())
     {
     case S_CONNECTED:
@@ -317,19 +336,29 @@ wsrep_status_t galera::ReplicatorSMM::connect(const std::string& cluster_name,
 
     ssize_t err;
     wsrep_status_t ret(WSREP_OK);
-    wsrep_seqno_t const seqno(cert_.position());
+    wsrep_seqno_t const seqno(STATE_SEQNO());
     wsrep_uuid_t  const gcs_uuid(seqno < 0 ? WSREP_UUID_UNDEFINED :state_uuid_);
 
     log_info << "Setting initial position to " << gcs_uuid << ':' << seqno;
 
-    if ((err = gcs_.set_initial_position(gcs_uuid, seqno)) != 0)
+    if ((bootstrap == true || cluster_url == "gcomm://")
+        && safe_to_bootstrap_ == false)
+    {
+        log_error << "It may not be safe to bootstrap the cluster from this node. "
+                  << "It was not the last one to leave the cluster and may "
+                  << "not contain all the updates. To force cluster bootstrap "
+                  << "with this node, edit the grastate.dat file manually and "
+                  << "set safe_to_bootstrap to 1 .";
+        ret = WSREP_NODE_FAIL;
+    }
+
+    if (ret == WSREP_OK &&
+        (err = gcs_.set_initial_position(gcs_uuid, seqno)) != 0)
     {
         log_error << "gcs init failed:" << strerror(-err);
         ret = WSREP_NODE_FAIL;
     }
 
-    gcache_.reset();
-
     if (ret == WSREP_OK &&
         (err = gcs_.connect(cluster_name, cluster_url, bootstrap)) != 0)
     {
@@ -467,9 +496,14 @@ void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
     /* at this point any exception in apply_trx_ws() is fatal, not
      * catching anything. */
 
-    if (gu_likely(co_mode_ != CommitOrder::BYPASS))
+    TrxHandle* commit_trx_handle = trx;
+    if (gu_likely(co_mode_ != CommitOrder::BYPASS) && trx->is_toi())
     {
+        /* TOI action are fully serialized so it is make sense to
+        enforce commit ordering at this stage. For non-TOI action
+        commit ordering is delayed to take advantage of full parallelism. */
         gu_trace(commit_monitor_.enter(co));
+        commit_trx_handle = NULL;
     }
     trx->set_state(TrxHandle::S_COMMITTING);
 
@@ -477,6 +511,7 @@ void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
     wsrep_cb_status_t const rcode(
         commit_cb_(
             recv_ctx,
+            commit_trx_handle,
             TrxHandle::trx_flags_to_wsrep_flags(trx->flags()),
             &meta,
             &exit_loop,
@@ -485,9 +520,9 @@ void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
     if (gu_unlikely (rcode != WSREP_CB_SUCCESS))
         gu_throw_fatal << "Commit failed. Trx: " << trx;
 
-    if (gu_likely(co_mode_ != CommitOrder::BYPASS))
+    if (gu_likely(co_mode_ != CommitOrder::BYPASS) && trx->is_toi())
     {
-        commit_monitor_.leave(co);
+        gu_trace(commit_monitor_.leave(co));
     }
     trx->set_state(TrxHandle::S_COMMITTED);
 
@@ -666,6 +701,12 @@ wsrep_status_t galera::ReplicatorSMM::replicate(TrxHandle* trx,
     else
     {
         retval = WSREP_OK;
+        if (meta != 0)
+        {
+            meta->gtid.uuid  = state_uuid_;
+            meta->gtid.seqno = trx->global_seqno();
+            meta->depends_on = trx->depends_seqno();
+        }
     }
 
     assert(trx->last_seen_seqno() >= 0);
@@ -745,17 +786,32 @@ galera::ReplicatorSMM::abort_trx(TrxHandle* trx)
 wsrep_status_t galera::ReplicatorSMM::pre_commit(TrxHandle*        trx,
                                                  wsrep_trx_meta_t* meta)
 {
+    /* Replicate and pre-commit action are 2 different actions now.
+    This means transaction can get aborted on completion of replicate
+    before pre-commit action start. This condition capture that scenario
+    and ensure that resources are released. */
+    if (trx->state() == TrxHandle::S_MUST_ABORT ||
+        trx->state() == TrxHandle::S_ABORTING)
+    {
+        LocalOrder lo(*trx);
+        ApplyOrder ao(*trx);
+        CommitOrder co(*trx, co_mode_);
+
+        local_monitor_.self_cancel(lo);
+
+        gcache_.seqno_assign (trx->action(), trx->global_seqno(), -1);
+
+        apply_monitor_.self_cancel(ao);
+        if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.self_cancel(co);
+
+        return WSREP_PRECOMMIT_ABORT;
+    }
+
     assert(trx->state() == TrxHandle::S_REPLICATING);
     assert(trx->local_seqno()  > -1);
     assert(trx->global_seqno() > -1);
     assert(trx->last_seen_seqno() >= 0);
 
-    if (meta != 0)
-    {
-        meta->gtid.uuid  = state_uuid_;
-        meta->gtid.seqno = trx->global_seqno();
-        meta->depends_on = trx->depends_seqno();
-    }
     // State should not be checked here: If trx has been replicated,
     // it has to be certified and potentially applied. #528
     // if (state_() < S_JOINED) return WSREP_TRX_FAIL;
@@ -899,6 +955,7 @@ wsrep_status_t galera::ReplicatorSMM::replay_trx(TrxHandle* trx, void* trx_ctx)
             wsrep_cb_status_t rcode(
                 commit_cb_(
                     trx_ctx,
+                    NULL,
                     TrxHandle::trx_flags_to_wsrep_flags(trx->flags()),
                     &meta,
                     &unused,
@@ -937,7 +994,7 @@ wsrep_status_t galera::ReplicatorSMM::replay_trx(TrxHandle* trx, void* trx_ctx)
 }
 
 
-wsrep_status_t galera::ReplicatorSMM::post_commit(TrxHandle* trx)
+wsrep_status_t galera::ReplicatorSMM::interim_commit(TrxHandle* trx)
 {
     if (trx->state() == TrxHandle::S_MUST_ABORT)
     {
@@ -957,6 +1014,36 @@ wsrep_status_t galera::ReplicatorSMM::post_commit(TrxHandle* trx)
 
     CommitOrder co(*trx, co_mode_);
     if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.leave(co);
+    trx->mark_interim_committed(true);
+
+    return WSREP_OK;
+}
+
+
+wsrep_status_t galera::ReplicatorSMM::post_commit(TrxHandle* trx)
+{
+    if (trx->state() == TrxHandle::S_MUST_ABORT)
+    {
+        // This is possible in case of ALG: BF applier BF aborts
+        // trx that has already grabbed commit monitor and is committing.
+        // However, this should be acceptable assuming that commit
+        // operation does not reserve any more resources and is able
+        // to release already reserved resources.
+        log_debug << "trx was BF aborted during commit: " << *trx;
+        // manipulate state to avoid crash
+        trx->set_state(TrxHandle::S_MUST_REPLAY);
+        trx->set_state(TrxHandle::S_REPLAYING);
+    }
+    assert(trx->state() == TrxHandle::S_COMMITTING ||
+           trx->state() == TrxHandle::S_REPLAYING);
+    assert(trx->local_seqno() > -1 && trx->global_seqno() > -1);
+
+    if (!(trx->is_interim_committed()))
+    {
+        CommitOrder co(*trx, co_mode_);
+        if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.leave(co);
+    }
+    trx->mark_interim_committed(false);
 
     ApplyOrder ao(*trx);
     report_last_committed(cert_.set_trx_committed(trx));
@@ -1226,6 +1313,13 @@ galera::ReplicatorSMM::sst_sent(const wsrep_gtid_t& state_id, int const rcode)
     if (state_() != S_DONOR)
     {
         log_error << "sst sent called when not SST donor, state " << state_();
+        /* If sst-sent fails node should restore itself back to joined state.
+        sst-sent can fail commonly due to n/w error where-in DONOR may loose
+        connectivity to JOINER (or existing cluster) but on re-join it should
+        restore the original state (DONOR->JOINER->JOINED->SYNCED) without
+        waiting for JOINER. sst-failure on JOINER will gracefully shutdown the
+        joiner. */
+        gcs_.join_notification();
         return WSREP_CONN_FAIL;
     }
 
@@ -1419,6 +1513,11 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
 
     wsrep_seqno_t const upto(cert_.position());
 
+    if (view_info.status == WSREP_VIEW_PRIMARY)
+    {
+        safe_to_bootstrap_ = (view_info.memb_num == 1);
+    }
+
     apply_monitor_.drain(upto);
 
     if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.drain(upto);
@@ -1473,15 +1572,16 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
 
         // we have to reset cert initial position here, SST does not contain
         // cert index yet (see #197).
+        // Also this must be done before releasing GCache buffers.
         cert_.assign_initial_position(group_seqno, trx_params_.version_);
+
+        if (STATE_SEQNO() > 0) service_thd_.release_seqno(STATE_SEQNO());
+        // make sure all gcache buffers are released
+
         // at this point there is no ongoing master or slave transactions
         // and no new requests to service thread should be possible
-
         service_thd_.flush();             // make sure service thd is idle
 
-        if (STATE_SEQNO() > 0) gcache_.seqno_release(STATE_SEQNO());
-        // make sure all gcache buffers are released
-
         // record state seqno, needed for IST on DONOR
         cc_seqno_ = group_seqno;
 
@@ -1490,15 +1590,35 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
         if (st_required && app_wants_st)
         {
             // GCache::Seqno_reset() happens here
-            request_state_transfer (recv_ctx,
-                                    group_uuid, group_seqno, app_req,
-                                    app_req_len);
+            long ret =
+                request_state_transfer (recv_ctx,
+                                        group_uuid, group_seqno,
+                                        app_req, app_req_len);
+
+            if (ret < 0 || sst_state_ == SST_CANCELED)
+            {
+                // If the IST/SST request was canceled due to error
+                // at the GCS level or if request was canceled by another
+                // thread (by initiative of the server), and if the node
+                // remain in the S_JOINING state, then we must return it
+                // to the S_CONNECTED state (to the original state, which
+                // exist before the request_state_transfer started).
+                // In other words, if state transfer failed, then we
+                // need to move node back to the original state, because
+                // joining was canceled:
+
+                if (state_() == S_JOINING)
+                {
+                    state_.shift_to(S_CONNECTED);
+                }
+            }
         }
         else
         {
             if (view_info.view == 1 || !app_wants_st)
             {
                 update_state_uuid (group_uuid, group_seqno);
+                gcache_.seqno_reset(to_gu_uuid(group_uuid), group_seqno);
                 apply_monitor_.set_initial_position(group_seqno);
                 if (co_mode_ != CommitOrder::BYPASS)
                     commit_monitor_.set_initial_position(group_seqno);
@@ -1530,11 +1650,16 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
                 }
             }
 
-            st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED);
+            st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED, safe_to_bootstrap_);
         }
 
-        if (state_() == S_JOINING && sst_state_ != SST_NONE
-                                  && sst_state_ != SST_CANCELED)
+        // We should not try to joining the cluster at the GCS level,
+        // if the node is not in the S_JOINING state, or if we did not
+        // sent the IST/SST request, or if it is failed. In other words,
+        // any state other than SST_WAIT (f.e. SST_NONE or SST_CANCELED)
+        // not require us to sending the JOIN message at the GCS level:
+
+        if (sst_state_ == SST_WAIT && state_() == S_JOINING)
         {
             /* There are two reasons we can be here:
              * 1) we just got state transfer in request_state_transfer() above;
@@ -1555,7 +1680,7 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
         // Non-primary configuration
         if (state_uuid_ != WSREP_UUID_UNDEFINED)
         {
-            st_.set (state_uuid_, STATE_SEQNO());
+            st_.set (state_uuid_, STATE_SEQNO(), safe_to_bootstrap_);
         }
 
         if (next_state != S_CONNECTED && next_state != S_CLOSING)
@@ -1648,7 +1773,7 @@ wsrep_seqno_t galera::ReplicatorSMM::pause()
     }
 
     wsrep_seqno_t const ret(STATE_SEQNO());
-    st_.set(state_uuid_, ret);
+    st_.set(state_uuid_, ret, safe_to_bootstrap_);
 
     log_info << "Provider paused at " << state_uuid_ << ':' << ret
              << " (" << pause_seqno_ << ")";
@@ -1664,7 +1789,7 @@ void galera::ReplicatorSMM::resume()
         gu_throw_error(EALREADY) << "tried to resume unpaused provider";
     }
 
-    st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED);
+    st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED, safe_to_bootstrap_);
     log_info << "resuming provider at " << pause_seqno_;
     LocalOrder lo(pause_seqno_);
     pause_seqno_ = WSREP_SEQNO_UNDEFINED;
@@ -1915,7 +2040,7 @@ galera::ReplicatorSMM::update_state_uuid (const wsrep_uuid_t& uuid,
                 sizeof(state_uuid_str_));
     }
 
-    st_.set(uuid, seqno);
+    st_.set(uuid, seqno, safe_to_bootstrap_);
 }
 
 void
