@@ -101,9 +101,39 @@ namespace galera
         wsrep_status_t pre_commit(TrxHandle*  trx, wsrep_trx_meta_t*);
         wsrep_status_t replay_trx(TrxHandle* trx, void* replay_ctx);
 
+        wsrep_status_t interim_commit(TrxHandle* trx);
         wsrep_status_t post_commit(TrxHandle* trx);
         wsrep_status_t post_rollback(TrxHandle* trx);
 
+        wsrep_status_t applier_pre_commit(void* trx_handle)
+        {
+            TrxHandle* trx = reinterpret_cast<TrxHandle*>(trx_handle);
+            CommitOrder co(*trx, co_mode_);
+            commit_monitor_.enter(co);
+            return WSREP_OK;
+        }
+
+        wsrep_status_t applier_interim_commit(void* trx_handle)
+        {
+            TrxHandle* trx = reinterpret_cast<TrxHandle*>(trx_handle);
+            CommitOrder co(*trx, co_mode_);
+            commit_monitor_.leave(co);
+            trx->mark_interim_committed(true);
+            return WSREP_OK;
+        }
+
+        wsrep_status_t applier_post_commit(void* trx_handle)
+        {
+            TrxHandle* trx = reinterpret_cast<TrxHandle*>(trx_handle);
+            if (!(trx->is_interim_committed()))
+            {
+                CommitOrder co(*trx, co_mode_);
+                commit_monitor_.leave(co);
+            }
+            trx->mark_interim_committed(false);
+            return WSREP_OK;
+        }
+
         wsrep_status_t causal_read(wsrep_gtid_t*);
         wsrep_status_t to_isolation_begin(TrxHandle* trx, wsrep_trx_meta_t*);
         wsrep_status_t to_isolation_end(TrxHandle* trx);
@@ -364,7 +394,7 @@ namespace galera
                 case OOOC:
                     return true;
                 case LOCAL_OOOC:
-                    return trx_.is_local();
+                    if (trx_.is_local()) { return true; }
                     // in case of remote trx fall through
                 case NO_OOOC:
                     return (last_left + 1 == trx_.global_seqno());
@@ -487,7 +517,7 @@ namespace galera
 
         long send_state_request (const StateRequest* req, const bool unsafe);
 
-        void request_state_transfer (void* recv_ctx,
+        long request_state_transfer (void* recv_ctx,
                                      const wsrep_uuid_t& group_uuid,
                                      wsrep_seqno_t       group_seqno,
                                      const void*         sst_req,
@@ -553,6 +583,10 @@ namespace galera
         std::string           state_file_;
         SavedState            st_;
 
+        // boolean telling if the node is safe to use for bootstrapping
+        // a new primary component
+        bool safe_to_bootstrap_;
+
         // currently installed trx parameters
         TrxHandle::Params     trx_params_;
 
@@ -571,6 +605,7 @@ namespace galera
         wsrep_unordered_cb_t  unordered_cb_;
         wsrep_sst_donate_cb_t sst_donate_cb_;
         wsrep_synced_cb_t     synced_cb_;
+        wsrep_abort_cb_t      abort_cb_;
 
         // SST
         std::string   sst_donor_;
@@ -633,6 +668,10 @@ namespace galera
 #endif /* HAVE_PSI_INTERFACE */
 
         mutable std::vector<struct wsrep_stats_var> wsrep_stats_;
+
+        // Storage space for dynamic status strings
+        char                  interval_string_[64];
+        char                  ist_status_string_[128];
     };
 
     std::ostream& operator<<(std::ostream& os, ReplicatorSMM::State state);
