@@ -105,6 +105,7 @@ struct gcs_conn
 {
     long  my_idx;
     long  memb_num;
+    long  non_arb_memb_count;   /* count of non-arb members in cluster */
     char* my_name;
     char* channel;
     char* socket;
@@ -151,9 +152,23 @@ struct gcs_conn
     /* #603, #606 join control */
     bool        volatile need_to_join;
     gcs_seqno_t volatile join_seqno;
+    void        join_notification()
+    {
+        need_to_join = true;
+    }
 
     /* sync control */
-    bool         sync_sent;
+    bool         sync_sent_;
+    bool         sync_sent() const
+    {
+        assert(gu_fifo_locked(recv_q));
+        return sync_sent_;
+    }
+    void         sync_sent(bool const val)
+    {
+        assert(gu_fifo_locked(recv_q));
+        sync_sent_ = val;
+    }
 
     /* gcs_core object */
     gcs_core_t*  core; // the context that is returned by
@@ -267,9 +282,17 @@ gcs_create (gu_config_t* const conf, gcache_t* const gcache,
 
     {
         size_t recv_q_len = gu_avphys_bytes() / sizeof(struct gcs_recv_act) / 4;
-
-        gu_debug ("Requesting recv queue len: %zu", recv_q_len);
-        conn->recv_q = gu_fifo_create (recv_q_len, sizeof(struct gcs_recv_act));
+        if (recv_q_len == 0)
+        {
+            gu_error ("Requesting recv queue len: %zu", recv_q_len);
+            gu_error ("Available system memory is running low: %zu",
+                      gu_avphys_bytes());
+        }
+        else
+        {
+            gu_debug ("Requesting recv queue len: %zu", recv_q_len);
+            conn->recv_q = gu_fifo_create (recv_q_len, sizeof(struct gcs_recv_act));
+        }
     }
     if (!conn->recv_q) {
         gu_error ("Failed to create recv_q.");
@@ -474,9 +497,15 @@ static inline bool
 gcs_send_sync_begin (gcs_conn_t* conn)
 {
     if (gu_unlikely(GCS_CONN_JOINED == conn->state)) {
-        if (conn->lower_limit >= conn->queue_len && !conn->sync_sent) {
+        if (conn->lower_limit >= conn->queue_len && !conn->sync_sent()) {
             // tripped lower slave queue limit, send SYNC message
-            conn->sync_sent = true;
+            conn->sync_sent(true);
+#if 0
+            gu_info ("Sending SYNC: state = %s, queue_len = %ld, "
+                     "lower_limit = %ld, sync_sent = %s",
+                     gcs_conn_state_str[conn->state], conn->queue_len,
+                     conn->lower_limit, conn->sync_sent() ? "true" : "false");
+#endif
             return true;
         }
 #if 0
@@ -484,7 +513,7 @@ gcs_send_sync_begin (gcs_conn_t* conn)
             gu_info ("Not sending SYNC: state = %s, queue_len = %ld, "
                      "lower_limit = %ld, sync_sent = %s",
                      gcs_conn_state_str[conn->state], conn->queue_len,
-                     conn->lower_limit, conn->sync_sent ? "true" : "false");
+                     conn->lower_limit, conn->sync_sent() ? "true" : "false");
         }
 #endif
     }
@@ -505,7 +534,9 @@ gcs_send_sync_end (gcs_conn_t* conn)
         ret = 0;
     }
     else {
-        conn->sync_sent = false;
+        gu_fifo_lock(conn->recv_q);
+        conn->sync_sent(false);
+        gu_fifo_release(conn->recv_q);
     }
 
     ret = gcs_check_error (ret, "Failed to send SYNC signal");
@@ -516,7 +547,11 @@ gcs_send_sync_end (gcs_conn_t* conn)
 static inline long
 gcs_send_sync (gcs_conn_t* conn)
 {
-    if (gcs_send_sync_begin (conn)) {
+    gu_fifo_lock(conn->recv_q);
+    bool const send_sync(gcs_send_sync_begin (conn));
+    gu_fifo_release(conn->recv_q);
+
+    if (send_sync) {
         return gcs_send_sync_end (conn);
     }
     else {
@@ -692,6 +727,9 @@ _release_sst_flow_control (gcs_conn_t* conn)
         if (conn->stop_sent > 0) {
             ret = gcs_send_fc_event (conn, GCS_FC_CONT);
             conn->stop_sent -= (ret >= 0);
+
+            if (ret >= 0)
+                gu_info("SST leaving flow control");
         }
     }
     while (ret < 0 && -EAGAIN == ret); // we need to send CONT here at all costs
@@ -734,8 +772,12 @@ gcs_become_joined (gcs_conn_t* conn)
 static void
 gcs_become_synced (gcs_conn_t* conn)
 {
-    gcs_shift_state (conn, GCS_CONN_SYNCED);
-    conn->sync_sent = false;
+    gu_fifo_lock(conn->recv_q);
+    {
+        gcs_shift_state (conn, GCS_CONN_SYNCED);
+        conn->sync_sent(false);
+    }
+    gu_fifo_release(conn->recv_q);
     gu_debug("Become synced, FC offset %ld", conn->fc_offset);
     conn->fc_offset = 0;
 }
@@ -747,7 +789,7 @@ _set_fc_limits (gcs_conn_t* conn)
     /* Killing two birds with one stone: flat FC profile for master-slave setups
      * plus #440: giving single node some slack at some math correctness exp.*/
     double const fn
-        (conn->params.fc_master_slave ? 1.0 : sqrt(double(conn->memb_num)));
+        (conn->params.fc_master_slave ? 1.0 : sqrt(double(conn->non_arb_memb_count)));
 
     conn->upper_limit = conn->params.fc_base_limit * fn + .5;
     conn->lower_limit = conn->upper_limit * conn->params.fc_resume_factor + .5;
@@ -834,11 +876,27 @@ gcs_handle_act_conf (gcs_conn_t* conn, const void* action)
     {
         /* reset flow control as membership is most likely changed */
         if (!gu_mutex_lock (&conn->fc_lock)) {
+
             conn->stop_sent   = 0;
             conn->stop_count  = 0;
             conn->conf_id     = conf->conf_id;
             conn->memb_num    = conf->memb_num;
 
+            // Count the number of non-arb members, this will be
+            // used for the fc_limit calculations
+            long    non_arb_memb_count = 0;
+            const char *  ptr = &conf->data[0];
+            for (long i=0; i < conf->memb_num; i++)
+            {
+                ptr += strlen(ptr) + 1;     // move past the ID
+                ptr += strlen(ptr) + 1;     // move past the name
+                if (*ptr)                   // 0-length IP addr indicates an ARB
+                    non_arb_memb_count += 1;
+                ptr += strlen(ptr) + 1;     // move past the IP address
+                ptr += sizeof(gcs_seqno_t); // move past the gcs_seqno_t
+            }
+            conn->non_arb_memb_count = non_arb_memb_count;
+
             _set_fc_limits (conn);
 
             gu_mutex_unlock (&conn->fc_lock);
@@ -848,7 +906,7 @@ gcs_handle_act_conf (gcs_conn_t* conn, const void* action)
             abort();
         }
 
-        conn->sync_sent = false;
+        conn->sync_sent(false);
 
         // need to wake up send monitor if it was paused during CC
         gcs_sm_continue(conn->sm);
@@ -922,19 +980,9 @@ gcs_handle_act_conf (gcs_conn_t* conn, const void* action)
     switch (conn->state) {
     case GCS_CONN_JOINED:
         /* One of the cases when the node can become SYNCED */
-    {
-        bool send_sync = false;
-
-        gu_fifo_lock(conn->recv_q);
-        {
-            send_sync = gcs_send_sync_begin(conn);
-        }
-        gu_fifo_release (conn->recv_q);
-
-        if (send_sync && (ret = gcs_send_sync_end (conn))) {
+        if ((ret = gcs_send_sync(conn)) < 0) {
             gu_warn ("CC: sending SYNC failed: %ld (%s)", ret, strerror (-ret));
         }
-    }
     break;
     case GCS_CONN_JOINER:
     case GCS_CONN_DONOR:
@@ -981,8 +1029,9 @@ gcs_handle_state_change (gcs_conn_t*     conn,
 
     if (buf) {
         memcpy (buf, act->buf, act->buf_len);
-        /* initially act->buf points to internal static recv buffer. No leak here */
-        act->buf = buf;
+        /* Initially act->buf points to internal static recv buffer.
+         * No leak here. */
+        ((struct gcs_act*)act)->buf = buf;
         return 1;
     }
     else {
@@ -1025,8 +1074,15 @@ gcs_handle_actions (gcs_conn_t*          conn,
             gcs_become_joined (conn);
         break;
     case GCS_ACT_SYNC:
-        ret = gcs_handle_state_change (conn, &rcvd->act);
-        gcs_become_synced (conn);
+        if (rcvd->id < 0) {
+            gu_fifo_lock(conn->recv_q);
+            conn->sync_sent(false);
+            gu_fifo_release(conn->recv_q);
+            gcs_send_sync(conn);
+        } else {
+            ret = gcs_handle_state_change (conn, &rcvd->act);
+            gcs_become_synced (conn);
+        }
         break;
     default:
         break;
@@ -1080,6 +1136,8 @@ _check_recv_queue_growth (gcs_conn_t* conn, ssize_t size)
             if ((ret = gcs_send_fc_event (conn, GCS_FC_STOP)) >= 0) {
                 conn->stop_sent++;
                 ret = 0;
+
+                gu_info("SST entering flow control");
             }
             else {
                 ret = gcs_check_error (ret, "Failed to send SST FC_STOP.");
@@ -1202,9 +1260,7 @@ static void *gcs_recv_thread (void *arg)
         struct gcs_repl_act** repl_act_ptr;
         struct gcs_act_rcvd   rcvd;
 
-        bool* sync_sent = (conn->state == GCS_CONN_DONOR
-                           ? &conn->sync_sent : NULL);
-        ret = gcs_core_recv (conn->core, &rcvd, conn->timeout, sync_sent);
+        ret = gcs_core_recv (conn->core, &rcvd, conn->timeout);
 
         if (gu_unlikely(ret <= 0)) {
 
@@ -1391,7 +1447,7 @@ long gcs_open (gcs_conn_t* conn, const char* channel, const char* url,
                 gcs_fifo_lite_open(conn->repl_q);
                 gu_fifo_open(conn->recv_q);
                 gcs_shift_state (conn, GCS_CONN_OPEN);
-                gu_info ("Opened channel '%s'", channel);
+                gu_debug ("Opened channel '%s'", channel);
                 conn->inner_close_count = 0;
                 conn->outer_close_count = 0;
                 goto out;
@@ -2001,6 +2057,11 @@ gcs_get_stats (gcs_conn_t* conn, struct gcs_stats* stats)
 
     stats->fc_sent     = conn->stats_fc_sent;
     stats->fc_received = conn->stats_fc_received;
+
+    stats->fc_lower_limit = conn->lower_limit;
+    stats->fc_upper_limit = conn->upper_limit;
+
+    stats->fc_status = conn->stop_sent > 0 ? 1 : 0;
 }
 
 void
@@ -2261,3 +2322,9 @@ const char* gcs_param_get (gcs_conn_t* conn, const char* key)
 
     return NULL;
 }
+
+
+void gcs_join_notification(gcs_conn_t* conn)
+{
+     conn->need_to_join = true;
+}
