@@ -170,7 +170,7 @@ group_redo_last_applied (gcs_group_t* group)
 {
     long       n;
     long       last_node    = -1;
-    gu_seqno_t last_applied = GU_LONG_LONG_MAX;
+    gu_seqno_t last_applied = GU_LLONG_MAX;
 
     for (n = 0; n < group->num; n++) {
         const gcs_node_t* const node = &group->nodes[n];
@@ -256,7 +256,7 @@ group_check_donor (gcs_group_t* group)
         gu_warn ("Donor %s is no longer in the group. State transfer cannot "
                  "be completed, need to abort. Aborting...", donor_id);
 
-        gu_abort();
+        // gu_abort();
     }
 
     return;
@@ -353,7 +353,7 @@ group_post_state_exchange (gcs_group_t* group)
              "\n\tversion    = %u,"
              "\n\tcomponent  = %s,"
              "\n\tconf_id    = %lld,"
-             "\n\tmembers    = %d/%d (joined/total),"
+             "\n\tmembers    = %d/%d (primary/total),"
              "\n\tact_id     = %lld,"
              "\n\tlast_appl. = %lld,"
              "\n\tprotocols  = %d/%d/%d (gcs/repl/appl),"
@@ -741,6 +741,25 @@ gcs_group_handle_join_msg  (gcs_group_t* group, const gcs_recv_msg_t* msg)
 
             if (from_donor && peer_idx == group->my_idx &&
                 GCS_NODE_STATE_JOINER == group->nodes[peer_idx].status) {
+
+                // If there is an ENODATA error code, then it is indication
+                // that state on the donor node was moved forward too much
+                // and we need to initiate a full SST instead of IST.
+                // We should not treat this condition as a fatal error
+                // (decision on this matter should take higher levels,
+                // for example, Replicator or the server):
+
+                if (seqno == -ENODATA)
+                {
+                    gu_fatal ("State transfer request failed unrecoverably "
+                              "because the donor seqno had gone forward "
+                              "during IST, but SST request was not prepared "
+                              "from our side due to selected state transfer "
+                              "method (which do not supports SST during "
+                              "node operation). Restart required.");
+                    return -ENOTRECOVERABLE;
+                }
+
                 // this node will be waiting for SST forever. If it has only
                 // one recv thread there is no (generic) way to wake it up.
                 gu_fatal ("Will never receive state. Need to abort.");
@@ -821,24 +840,23 @@ gcs_group_handle_sync_msg  (gcs_group_t* group, const gcs_recv_msg_t* msg)
         return (sender_idx == group->my_idx);
     }
     else {
-        int ret = 0;
-
-        if (GCS_NODE_STATE_DONOR == sender->status) {
-            gu_info ("SYNC message ignored as node %d.%d (%s) was"
-                     " re-transitioned to DONOR mode before it synced.",
-                     sender_idx, sender->segment, sender->name);
-            ret = -1;
-        }
-        else if (GCS_NODE_STATE_SYNCED != sender->status) {
-            gu_warn ("SYNC message sender from non-JOINED %d.%d (%s). Ignored.",
-                     sender_idx, sender->segment, sender->name);
-        }
-        else {
+        if (GCS_NODE_STATE_SYNCED == sender->status) {
             gu_debug ("Redundant SYNC message from %d.%d (%s).",
                       sender_idx, sender->segment, sender->name);
         }
+        else if (GCS_NODE_STATE_DONOR == sender->status) {
+            // this is possible with quick succession of desync()/resync() calls
+            gu_debug ("SYNC message from %d.%d (%s, DONOR). Ignored.",
+                      sender_idx, sender->segment, sender->name);
+        }
+        else {
+            gu_warn ("SYNC message from non-JOINED %d.%d (%s, %s). Ignored.",
+                     sender_idx, sender->segment, sender->name,
+                     gcs_node_state_to_str(sender->status));
+        }
 
-        return ret;
+        /* signal sender that it didn't work */
+        return -ERESTART * (sender_idx == group->my_idx);
     }
 }
 
@@ -1144,7 +1162,8 @@ group_find_ist_donor (const gcs_group_t* const group,
                       int joiner_idx,
                       const char* str, int str_len,
                       gcs_seqno_t ist_seqno,
-                      gcs_node_state_t status)
+                      gcs_node_state_t status,
+                      const bool ist_only)
 {
     int idx = -1;
 
@@ -1158,7 +1177,11 @@ group_find_ist_donor (const gcs_group_t* const group,
     gcs_seqno_t const max_cached_range = conf_seqno - lowest_cached_seqno;
     gcs_seqno_t safety_gap = max_cached_range >> 7; /* 1.0 / 128 ~= 0.008 */
     safety_gap = safety_gap < (1 << 20) ? safety_gap : (1 << 20); /* Be sensible and don't reserve more than 1M */
-    gcs_seqno_t safe_ist_seqno = lowest_cached_seqno + safety_gap;
+
+    // We should ignore safety_gap heuristic if the request contains
+    // only a IST part, since otherwise we cannot fulfill the request:
+
+    gcs_seqno_t safe_ist_seqno = lowest_cached_seqno + (ist_only ? 0 : safety_gap);
 
     gu_debug("ist_seqno[%lld], lowest_cached_seqno[%lld],"
              "conf_seqno[%lld], safe_ist_seqno[%lld]",
@@ -1189,7 +1212,8 @@ gcs_group_find_donor(const gcs_group_t* group,
                      int const str_version,
                      int const joiner_idx,
                      const char* const donor_string, int const donor_len,
-                     const gu_uuid_t* ist_uuid, gcs_seqno_t ist_seqno)
+                     const gu_uuid_t* ist_uuid, gcs_seqno_t ist_seqno,
+                     const bool ist_only)
 {
     static gcs_node_state_t const min_donor_state = GCS_NODE_STATE_SYNCED;
 
@@ -1205,7 +1229,8 @@ gcs_group_find_donor(const gcs_group_t* group,
                                          joiner_idx,
                                          donor_string, donor_len,
                                          ist_seqno,
-                                         min_donor_state);
+                                         min_donor_state,
+                                         ist_only);
     }
     if (donor_idx < 0)
     {
@@ -1235,7 +1260,8 @@ group_select_donor (gcs_group_t* group,
                     int const joiner_idx,
                     const char* const donor_string,
                     const gu_uuid_t* ist_uuid, gcs_seqno_t ist_seqno,
-                    bool const desync)
+                    const bool desync,
+                    const bool ist_only)
 {
     static gcs_node_state_t const min_donor_state = GCS_NODE_STATE_SYNCED;
     int  donor_idx;
@@ -1261,7 +1287,8 @@ group_select_donor (gcs_group_t* group,
                                          str_version,
                                          joiner_idx,
                                          donor_string, donor_len,
-                                         ist_uuid, ist_seqno);
+                                         ist_uuid, ist_seqno,
+                                         ist_only);
     }
 
     if (donor_idx >= 0) {
@@ -1391,10 +1418,54 @@ gcs_group_handle_state_request (gcs_group_t*         group,
         }
     }
 
+    // We need to perform a partial analysis of the application
+    // request to find out that it contains the SST part (not just IST):
+
+    int app_req_len = act->act.buf_len - (donor_name_len + 1);
+    char * app_req =
+        reinterpret_cast<char*>(const_cast<void*>(act->act.buf)) +
+        (donor_name_len + 1);
+
+    // Requests of the "zero" version contain only the SST part.
+    // They can be distinguished from the first version of the requests
+    // by the lack of the "magic" prefix:
+
+    int sst_len = app_req_len;
+    int magic_len = strlen("STRv1");
+
+    if (app_req_len > magic_len && !strncmp(app_req, "STRv1", magic_len))
+    {
+        // Request of the first version contain an explicit length
+        // of the SST part:
+
+        int offset = magic_len + 1;
+        sst_len = gtohl(*(reinterpret_cast<uint32_t*>(app_req + offset)));
+
+        // To facilitate debugging, we may need involuntary stimulation
+        // of the lack of SST part in the request:
+
+#ifdef GU_DBUG_ON
+        GU_DBUG_EXECUTE("simulate_ist_only_request", {
+           *(reinterpret_cast<uint32_t*>(app_req + offset)) = 0;
+           act->act.buf_len -= sst_len;
+           memmove(app_req     + offset + sizeof(uint32_t),
+                   app_req     + offset + sizeof(uint32_t) + sst_len,
+                   app_req_len - offset - sizeof(uint32_t) - sst_len);
+           sst_len = 0;
+        });
+#endif
+    }
+
+    bool ist_only = true;
+    if (sst_len)
+    {
+        ist_only = false;
+    }
+
     donor_idx = group_select_donor(group,
                                    str_version,
                                    joiner_idx, donor_name,
-                                   &ist_uuid, ist_seqno, desync);
+                                   &ist_uuid, ist_seqno, desync, ist_only);
 
     assert (donor_idx != joiner_idx || desync  || donor_idx < 0);
     assert (donor_idx == joiner_idx || !desync || donor_idx < 0);
