@@ -25,28 +25,12 @@ WSREP_SST_OPT_DATA=""
 WSREP_SST_OPT_AUTH=${WSREP_SST_OPT_AUTH:-}
 WSREP_SST_OPT_USER=${WSREP_SST_OPT_USER:-}
 WSREP_SST_OPT_PSWD=${WSREP_SST_OPT_PSWD:-}
-WSREP_SST_OPT_VERSION=""
+WSREP_LOG_DEBUG=0
 
 while [ $# -gt 0 ]; do
 case "$1" in
     '--address')
         readonly WSREP_SST_OPT_ADDR="$2"
-        #
-        # Break address string into host:port/path parts
-        #
-        if echo $WSREP_SST_OPT_ADDR | grep -qe '^\[.*\]'
-        then
-            # IPv6 notation
-            readonly WSREP_SST_OPT_HOST=${WSREP_SST_OPT_ADDR/\]*/\]}
-            readonly WSREP_SST_OPT_HOST_UNESCAPED=$(echo $WSREP_SST_OPT_HOST | \
-                 cut -d '[' -f 2 | cut -d ']' -f 1)
-        else
-            # "traditional" notation
-            readonly WSREP_SST_OPT_HOST=${WSREP_SST_OPT_ADDR%%[:/]*}
-        fi
-        readonly WSREP_SST_OPT_PORT=$(echo $WSREP_SST_OPT_ADDR | \
-                cut -d ']' -f 2 | cut -s -d ':' -f 2 | cut -d '/' -f 1)
-        readonly WSREP_SST_OPT_PATH=${WSREP_SST_OPT_ADDR#*/}
         shift
         ;;
     '--bypass')
@@ -96,10 +80,6 @@ case "$1" in
         WSREP_SST_OPT_USER="$2"
         shift
         ;;
-    '--mysqld-version')
-        readonly WSREP_SST_OPT_VERSION="$2"
-        shift
-        ;;
     '--gtid')
         readonly WSREP_SST_OPT_GTID="$2"
         shift
@@ -119,6 +99,44 @@ readonly WSREP_SST_OPT_BYPASS
 readonly WSREP_SST_OPT_BINLOG
 readonly WSREP_SST_OPT_CONF_SUFFIX
 
+
+#
+# user can specify xtrabackup specific settings that will be used during sst
+# process like encryption, etc.....
+# parse such configuration option. (group for xb settings is [sst] in my.cnf
+#
+# 1st param: group : name of the config file section, e.g. mysqld
+# 2nd param: var : name of the variable in the section, e.g. server-id
+# 3rd param: - : default value for the param
+parse_cnf()
+{
+    local group=$1
+    local var=$2
+    local reval=""
+
+    # print the default settings for given group using my_print_default.
+    # normalize the variable names specified in cnf file (user can use _ or - for example log-bin or log_bin)
+    # then grep for needed variable
+    # finally get the variable value (if variables has been specified multiple time use the last value only)
+
+    # look in group+suffix
+    if [[ -n $WSREP_SST_OPT_CONF_SUFFIX ]]; then
+        reval=$($MY_PRINT_DEFAULTS -c $WSREP_SST_OPT_CONF "${group}${WSREP_SST_OPT_CONF_SUFFIX}" | awk -F= '{if ($1 ~ /_/) { gsub(/_/,"-",$1); print $1"="$2 } else { print $0 }}' | grep -- "--$var=" | cut -d= -f2- | tail -1)
+    fi
+
+    # look in group
+    if [[ -z $reval ]]; then
+        reval=$($MY_PRINT_DEFAULTS -c $WSREP_SST_OPT_CONF $group | awk -F= '{if ($1 ~ /_/) { gsub(/_/,"-",$1); print $1"="$2 } else { print $0 }}' | grep -- "--$var=" | cut -d= -f2- | tail -1)
+    fi
+
+    # use default if we haven't found a value
+    if [[ -z $reval ]]; then
+        [[ -n $3 ]] && reval=$3
+    fi
+    echo $reval
+}
+
+
 # try to use my_print_defaults, mysql and mysqldump that come with the sources
 # (for MTR suite)
 SCRIPTS_DIR="$(cd $(dirname "$0"); pwd -P)"
@@ -177,12 +195,27 @@ else
     SST_PROGRESS_FILE=""
 fi
 
+# Determine the timezone to use for error logging
+LOGGING_TZ=$(parse_cnf mysqld log-timestamps "")
+if [ -z "$LOGGING_TZ" ]; then
+    LOGGING_TZ=$(parse_cnf mysqld log_timestamps "")
+fi
+if [ -z "$LOGGING_TZ" ]; then
+    LOGGING_TZ="UTC"
+fi
+# Set the options to the 'date' command based on the timezone
+if [[ "$LOGGING_TZ" == "SYSTEM" ]]; then
+    LOG_DATE_COMMAND="+%Y-%m-%dT%H:%M:%S.%6N%:z"
+else
+    LOG_DATE_COMMAND="-u +%Y-%m-%dT%H:%M:%S.%6NZ"
+fi
+
 wsrep_log()
 {
     # echo everything to stderr so that it gets into common error log
     # deliberately made to look different from the rest of the log
-    local readonly tst="$(date +%Y-%m-%d\ %H:%M:%S)"
-    echo "WSREP_SST: $* ($tst)" >&2
+    local readonly tst="$(date $LOG_DATE_COMMAND)"
+    echo -e "\t$tst WSREP_SST: $*" >&2
 }
 
 wsrep_log_error()
@@ -200,6 +233,13 @@ wsrep_log_info()
     wsrep_log "[INFO] $*"
 }
 
+wsrep_log_debug()
+{
+    if [[ -n "$WSREP_LOG_DEBUG" ]]; then
+        wsrep_log "[DEBUG] $*"
+    fi
+}
+
 wsrep_cleanup_progress_file()
 {
     [ -n "${SST_PROGRESS_FILE:-}" ] && rm -f "$SST_PROGRESS_FILE" 2>/dev/null || true
@@ -229,38 +269,3 @@ wsrep_check_programs()
     return $ret
 }
 
-#
-# user can specify xtrabackup specific settings that will be used during sst
-# process like encryption, etc.....
-# parse such configuration option. (group for xb settings is [sst] in my.cnf
-#
-# 1st param: group : name of the config file section, e.g. mysqld
-# 2nd param: var : name of the variable in the section, e.g. server-id
-# 3rd param: - : default value for the param
-parse_cnf()
-{
-    local group=$1
-    local var=$2
-    local reval=""
-
-    # print the default settings for given group using my_print_default.
-    # normalize the variable names specified in cnf file (user can use _ or - for example log-bin or log_bin)
-    # then grep for needed variable
-    # finally get the variable value (if variables has been specified multiple time use the last value only)
-
-    # look in group+suffix
-    if [[ -n $WSREP_SST_OPT_CONF_SUFFIX ]]; then
-        reval=$($MY_PRINT_DEFAULTS -c $WSREP_SST_OPT_CONF "${group}${WSREP_SST_OPT_CONF_SUFFIX}" | awk -F= '{st=index($0,"="); cur=$0; if ($1 ~ /_/) { gsub(/_/,"-",$1);} if (st != 0) { print $1"="substr(cur,st+1) } else { print cur }}' | grep -- "--$var=" | cut -d= -f2- | tail -1)
-    fi
-
-    # look in group
-    if [[ -z $reval ]]; then
-        reval=$($MY_PRINT_DEFAULTS -c $WSREP_SST_OPT_CONF "${group}" | awk -F= '{st=index($0,"="); cur=$0; if ($1 ~ /_/) { gsub(/_/,"-",$1);} if (st != 0) { print $1"="substr(cur,st+1) } else { print cur }}' | grep -- "--$var=" | cut -d= -f2- | tail -1)
-    fi
-
-    # use default if we haven't found a value
-    if [[ -z $reval ]]; then
-        [[ -n $3 ]] && reval=$3
-    fi
-    echo $reval
-}
