@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2013, 2016, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2013, 2017, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -33,7 +33,13 @@
 #include <sys/un.h>
 #endif
 #ifdef HAVE_LIBWRAP
+#ifndef HAVE_LIBWRAP_PROTOTYPES
+extern "C" {
 #include <tcpd.h>
+}
+#else
+#include <tcpd.h>
+#endif
 #include <syslog.h>
 #endif
 
@@ -43,9 +49,6 @@ ulong Mysqld_socket_listener::connection_errors_select= 0;
 ulong Mysqld_socket_listener::connection_errors_accept= 0;
 ulong Mysqld_socket_listener::connection_errors_tcpwrap= 0;
 
-#ifdef HAVE_PSI_STATEMENT_INTERFACE
-PSI_statement_info stmt_info_new_packet;
-#endif
 
 void net_before_header_psi(struct st_net *net, void *user_data, size_t /* unused: count */)
 {
@@ -533,6 +536,7 @@ public:
     (void) mysql_sock_set_nonblocking(listener_socket);
 #endif
 
+// TODO: PXC....is this needed ?
 #if defined(WITH_WSREP) && defined(HAVE_FCNTL) && defined(FD_CLOEXEC)
   (void) fcntl(mysql_socket_getfd(listener_socket), F_SETFD, FD_CLOEXEC);
 #endif /* WITH_WSREP */
@@ -953,6 +957,7 @@ Channel_info* Mysqld_socket_listener::listen_for_connection_event()
     return NULL;
   }
 
+// TODO: PXC .... is this needed ?
 #if defined(WITH_WSREP) && defined(HAVE_FCNTL) && defined(FD_CLOEXEC)
     (void) fcntl(mysql_socket_getfd(connect_sock), F_SETFD, FD_CLOEXEC);
 #endif /* WITH_WSREP */
@@ -987,9 +992,15 @@ Channel_info* Mysqld_socket_listener::listen_for_connection_event()
       syslog(LOG_AUTH | m_deny_severity,
              "refused connect from %s", eval_client(&req));
 
+#ifdef HAVE_LIBWRAP_PROTOTYPES
+      // Some distros have patched tcpd.h to have proper prototypes
       if (req.sink)
         (req.sink)(req.fd);
-
+#else
+      // Some distros have not patched tcpd.h
+      if (req.sink)
+        ((void (*)(int))req.sink)(req.fd);
+#endif
       mysql_socket_shutdown(listen_sock, SHUT_RDWR);
       mysql_socket_close(listen_sock);
       /*
