@@ -1,29 +1,20 @@
 /* Copyright (c) 2005, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
 
-#define MYSQL_LEX 1
-#include "my_global.h"                          /* NO_EMBEDDED_ACCESS_CHECKS */
-#include "sql_priv.h"
-#include "unireg.h"
+#include "event_data_objects.h"
+
 #include "sql_parse.h"                          // parse_sql
 #include "strfunc.h"                           // find_string_in_array
 #include "sql_db.h"                        // get_default_db_collation
@@ -31,23 +22,34 @@
                                            // date_add_interval,
                                            // calc_time_diff
 #include "tztime.h"     // my_tz_find, my_tz_OFFSET0, struct Time_zone
-#include "sql_acl.h"    // EVENT_ACL, SUPER_ACL
+#include "auth_common.h"                   // EVENT_ACL, SUPER_ACL
 #include "sp.h"         // load_charset, load_collation
 #include "events.h"
-#include "event_data_objects.h"
 #include "event_db_repository.h"
+#include "event_parse_data.h"
 #include "sp_head.h"
 #include "sql_show.h"                // append_definer, append_identifier
+#include "log.h"
 
-#ifdef WITH_WSREP
-#include "wsrep_thd.h"
-#endif
+#include "mysql/psi/mysql_sp.h"
 
 /**
   @addtogroup Event_Scheduler
   @{
 */
 
+#ifdef HAVE_PSI_INTERFACE
+void init_scheduler_psi_keys()
+{
+  const char *category= "scheduler";
+
+  PSI_server->register_statement(category, & Event_queue_element_for_exec::psi_info, 1);
+}
+
+PSI_statement_info Event_queue_element_for_exec::psi_info=
+{ 0, "event", 0};
+#endif
+
 /*************************************************************************/
 
 /**
@@ -119,8 +121,8 @@ Event_creation_ctx::load_from_db(THD *thd,
   {
     sql_print_warning("Event '%s'.'%s': invalid value "
                       "in column mysql.event.character_set_client.",
-                      (const char *) db_name,
-                      (const char *) event_name);
+                      db_name,
+                      event_name);
 
     invalid_creation_ctx= TRUE;
   }
@@ -132,8 +134,8 @@ Event_creation_ctx::load_from_db(THD *thd,
   {
     sql_print_warning("Event '%s'.'%s': invalid value "
                       "in column mysql.event.collation_connection.",
-                      (const char *) db_name,
-                      (const char *) event_name);
+                      db_name,
+                      event_name);
 
     invalid_creation_ctx= TRUE;
   }
@@ -145,8 +147,8 @@ Event_creation_ctx::load_from_db(THD *thd,
   {
     sql_print_warning("Event '%s'.'%s': invalid value "
                       "in column mysql.event.db_collation.",
-                      (const char *) db_name,
-                      (const char *) event_name);
+                      db_name,
+                      event_name);
 
     invalid_creation_ctx= TRUE;
   }
@@ -183,9 +185,11 @@ Event_creation_ctx::load_from_db(THD *thd,
 bool
 Event_queue_element_for_exec::init(LEX_STRING db, LEX_STRING n)
 {
-  if (!(dbname.str= my_strndup(db.str, dbname.length= db.length, MYF(MY_WME))))
+  if (!(dbname.str= my_strndup(key_memory_Event_queue_element_for_exec_names,
+                               db.str, dbname.length= db.length, MYF(MY_WME))))
     return TRUE;
-  if (!(name.str= my_strndup(n.str, name.length= n.length, MYF(MY_WME))))
+  if (!(name.str= my_strndup(key_memory_Event_queue_element_for_exec_names,
+                             n.str, name.length= n.length, MYF(MY_WME))))
   {
     my_free(dbname.str);
     return TRUE;
@@ -193,6 +197,11 @@ Event_queue_element_for_exec::init(LEX_STRING db, LEX_STRING n)
   return FALSE;
 }
 
+void Event_queue_element_for_exec::claim_memory_ownership()
+{
+  my_claim(dbname.str);
+  my_claim(name.str);
+}
 
 /*
   Destructor
@@ -219,7 +228,7 @@ Event_basic::Event_basic()
 {
   DBUG_ENTER("Event_basic::Event_basic");
   /* init memory root */
-  init_sql_alloc(&mem_root, 256, 512);
+  init_sql_alloc(key_memory_event_basic_root, &mem_root, 256, 512);
   dbname.str= name.str= NULL;
   dbname.length= name.length= 0;
   time_zone= NULL;
@@ -377,8 +386,10 @@ Event_timed::init()
 {
   DBUG_ENTER("Event_timed::init");
 
-  definer_user.str= definer_host.str= body.str= comment.str= NULL;
-  definer_user.length= definer_host.length= body.length= comment.length= 0;
+  definer_user= NULL_CSTR;
+  definer_host= NULL_CSTR;
+  body= NULL_STR;
+  comment= NULL_STR;
 
   sql_mode= 0;
 
@@ -618,7 +629,7 @@ Event_timed::load_from_row(THD *thd, TABLE *table)
                                        table, &creation_ctx))
   {
     push_warning_printf(thd,
-                        Sql_condition::WARN_LEVEL_WARN,
+                        Sql_condition::SL_WARNING,
                         ER_EVENT_INVALID_CREATION_CTX,
                         ER(ER_EVENT_INVALID_CREATION_CTX),
                         (const char *) dbname.str,
@@ -662,7 +673,7 @@ Event_timed::load_from_row(THD *thd, TABLE *table)
 static
 my_time_t
 add_interval(MYSQL_TIME *ltime, const Time_zone *time_zone,
-             interval_type scale, INTERVAL interval)
+             interval_type scale, Interval interval)
 {
   if (date_add_interval(ltime, scale, interval))
     return 0;
@@ -763,7 +774,7 @@ bool get_next_time(const Time_zone *time_zone, my_time_t *next,
     time_zone->gmt_sec_to_TIME(&local_now, time_now);
   }
 
-  INTERVAL interval;
+  Interval interval;
   memset(&interval, 0, sizeof(interval));
   my_time_t next_time= 0;
 
@@ -1205,7 +1216,7 @@ Event_timed::get_create_event(THD *thd, String *buf)
     DBUG_RETURN(EVEX_MICROSECOND_UNSUP);
 
   buf->append(STRING_WITH_LEN("CREATE "));
-  append_definer(thd, buf, &definer_user, &definer_host);
+  append_definer(thd, buf, definer_user, definer_host);
   buf->append(STRING_WITH_LEN("EVENT "));
   append_identifier(thd, buf, name.str, name.length);
 
@@ -1368,13 +1379,13 @@ Event_job_data::execute(THD *thd, bool drop)
     mysql_change_db will be invoked anyway later, to activate the
     procedure database before it's executed.
   */
-  thd->set_db(dbname.str, dbname.length);
+  thd->set_db(to_lex_cstring(dbname));
 
   lex_start(thd);
 
 #ifndef NO_EMBEDDED_ACCESS_CHECKS
   if (event_sctx.change_security_context(thd,
-                                         &definer_user, &definer_host,
+                                         definer_user, definer_host,
                                          &dbname, &save_sctx))
   {
     sql_print_error("Event Scheduler: "
@@ -1418,7 +1429,8 @@ Event_job_data::execute(THD *thd, bool drop)
 
   {
     Parser_state parser_state;
-    if (parser_state.init(thd, thd->query(), thd->query_length()))
+
+    if (parser_state.init(thd, thd->query().str, thd->query().length))
       goto end;
 
     thd->m_digest= NULL;
@@ -1450,30 +1462,18 @@ Event_job_data::execute(THD *thd, bool drop)
     sphead->set_creation_ctx(creation_ctx);
     sphead->optimize();
 
+    sphead->m_type= SP_TYPE_EVENT;
+#ifdef HAVE_PSI_SP_INTERFACE
+    sphead->m_sp_share= MYSQL_GET_SP_SHARE(SP_TYPE_EVENT,
+                                           dbname.str, dbname.length,
+                                           name.str, name.length);
+#endif
+
     ret= sphead->execute_procedure(thd, &empty_item_list);
     /*
       There is no pre-locking and therefore there should be no
       tables open and locked left after execute_procedure.
     */
-
-#ifdef WITH_WSREP
-    /*
-      If the thread has been killed, rollback the operation
-      properly (for WSREP).
-    */
-    if (WSREP(thd))
-    {
-      mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-      if (thd->wsrep_conflict_state == MUST_ABORT)
-      {
-        wsrep_client_rollback(thd);
-
-        WSREP_DEBUG("abort the event in exec query state, avoiding autocommit");
-      }
-      mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
-    }
-#endif /* WITH_WSREP */
-
   }
 
 end:
@@ -1506,51 +1506,42 @@ end:
         Temporarily reset it to read-write.
       */
 
-      saved_master_access= thd->security_ctx->master_access;
-      thd->security_ctx->master_access |= SUPER_ACL;
+      saved_master_access= thd->security_context()->master_access();
+      thd->security_context()->set_master_access(saved_master_access |
+                                                 SUPER_ACL);
       bool save_tx_read_only= thd->tx_read_only;
       thd->tx_read_only= false;
 
 #ifdef WITH_WSREP
-      /* 
-         This code is processing event execution and does not have client 
-         connection. Here, event execution will now execute a prepared 
-         DROP EVENT statement, but thd->lex->sql_command is set to 
-         SQLCOM_CREATE_PROCEDURE 
-         DROP EVENT will be logged in binlog, and we have to
-         replicate it to make all nodes have consistent event definitions
-         Wsrep DDL replication is triggered inside Events::drop_event(), 
-         and here we need to prepare the THD so that DDL replication is
-         possible, essentially it requires setting sql_command to 
-         SQLCOMM_DROP_EVENT, we will switch sql_command for the duration
-         of DDL replication only.
-      */
-      enum_sql_command sql_command_save= SQLCOM_END;
-      if (WSREP(thd))
-      {
-        sql_command_save= thd->lex->sql_command;
-        thd->lex->sql_command = SQLCOM_DROP_EVENT;
+      LEX* original_saved_lex= thd->lex;
+      if (WSREP(thd)) {
+          // sql_print_information("sizeof(LEX) = %d", sizeof(struct LEX));
+          // sizeof(LEX) = 4512, so it's relatively safe to allocate it on stack.
+          LEX lex;
+          lex.sql_command = SQLCOM_DROP_EVENT;
+          thd->lex = &lex;
+          WSREP_TO_ISOLATION_BEGIN(WSREP_MYSQL_DB, NULL, NULL);
+          thd->lex= original_saved_lex;
       }
-#endif
-
+#endif /* WITH_WSREP */
+     
       ret= Events::drop_event(thd, dbname, name, FALSE);
 
 #ifdef WITH_WSREP
-      if (WSREP(thd))
-      {
-        WSREP_TO_ISOLATION_END;
-        thd->lex->sql_command = sql_command_save;
-      }
-#endif
+      WSREP_TO_ISOLATION_END;
+error:
+      if (thd->lex != original_saved_lex)
+        thd->lex= original_saved_lex;
+#endif /* WITH_WSREP */
       thd->tx_read_only= save_tx_read_only;
-      thd->security_ctx->master_access= saved_master_access;
+      thd->security_context()->set_master_access(saved_master_access);
     }
   }
 #ifndef NO_EMBEDDED_ACCESS_CHECKS
   if (save_sctx)
     event_sctx.restore_security_context(thd, save_sctx);
 #endif
-  thd->lex->unit.cleanup();
+  thd->lex->unit->cleanup(true);
   thd->end_statement();
   thd->cleanup_after_query();
   /* Avoid races with SHOW PROCESSLIST */
