@@ -186,9 +186,11 @@ lock_tables_check(THD *thd, TABLE **tables, size_t count, uint flags)
 
     /*
       Prevent modifications to base tables if READ_ONLY is activated.
-      In any case, read only does not apply to temporary tables.
+      In any case, read only does not apply to temporary tables and
+      performance_schema tables.
     */
-    if (!(flags & MYSQL_LOCK_IGNORE_GLOBAL_READ_ONLY) && !t->s->tmp_table)
+    if (!(flags & MYSQL_LOCK_IGNORE_GLOBAL_READ_ONLY) && !t->s->tmp_table &&
+        !is_perfschema_db(t->s->db.str, t->s->db.length))
     {
       if (t->reginfo.lock_type >= TL_WRITE_ALLOW_WRITE &&
         check_readonly(thd, true))
@@ -1206,7 +1208,6 @@ void Global_read_lock::unlock_global_read_lock(THD *thd)
     }
 #endif /* WITH_WSREP */
   }
-
   thd->mdl_context.release_lock(m_mdl_global_shared_lock);
   my_atomic_add32(&Global_read_lock::m_active_requests, -1);
   m_mdl_global_shared_lock= NULL;
@@ -1273,7 +1274,7 @@ bool Global_read_lock::make_global_read_lock_block_commit(THD *thd)
   else if (WSREP(thd) && !provider_desynced_paused)
   {
     int rcode;
-    WSREP_DEBUG("running implicit desync for node");
+    WSREP_DEBUG("Running implicit desync for node from FTWRL");
     rcode = wsrep->desync(wsrep);
     if (rcode != WSREP_OK)
     {
@@ -1342,7 +1343,7 @@ wsrep_status_t Global_read_lock::wsrep_resume()
 
   if (ret != WSREP_OK)
   {
-    WSREP_WARN("Failed to resume provider: %d", ret);
+    WSREP_WARN("failed to resume provider: %d", ret);
   }
   else
   {
@@ -1447,35 +1448,19 @@ void Global_backup_lock::release(THD *thd)
 
 void Global_backup_lock::set_explicit_locks_duration(THD *thd)
 {
-  bool should_own;
-
   DBUG_ENTER("Global_backup_lock::set_explicit_lock_duration");
 
   if (m_lock)
-  {
-    should_own= true;
     thd->mdl_context.set_lock_duration(m_lock, MDL_EXPLICIT);
-  }
-  else
-  {
-    should_own= false;
-  }
 
-  DBUG_ASSERT(should_own ==
+  DBUG_ASSERT((m_lock != NULL) ==
               thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "", "",
                                                            MDL_SHARED));
 
   if (m_prot_lock)
-  {
-    should_own= true;
     thd->mdl_context.set_lock_duration(m_prot_lock, MDL_EXPLICIT);
-  }
-  else
-  {
-    should_own= false;
-  }
 
-  DBUG_ASSERT(should_own ==
+  DBUG_ASSERT((m_prot_lock != NULL) ==
               thd->mdl_context.owns_equal_or_stronger_lock(m_namespace, "", "",
                                                      MDL_INTENTION_EXCLUSIVE));
 
