@@ -402,11 +402,6 @@ static inline int my_thread_set_THR_THD(THD *thd)
 
 C_MODE_START
 
-#ifdef WITH_WSREP
-extern PSI_mutex_key key_LOCK_wsrep_thd;
-extern PSI_cond_key  key_COND_wsrep_thd;
-#endif /* WITH_WSREP */
-
 extern PSI_mutex_key key_LOCK_tc;
 
 #ifdef HAVE_OPENSSL
@@ -467,11 +462,17 @@ extern PSI_mutex_key key_mts_gaq_LOCK;
 extern PSI_mutex_key key_thd_timer_mutex;
 extern PSI_mutex_key key_LOCK_offline_mode;
 extern PSI_mutex_key key_LOCK_default_password_lifetime;
+extern PSI_mutex_key key_LOCK_group_replication_handler;
 
 #ifdef HAVE_REPLICATION
 extern PSI_mutex_key key_commit_order_manager_mutex;
 extern PSI_mutex_key key_mutex_slave_worker_hash;
 #endif
+#ifdef WITH_WSREP
+extern PSI_mutex_key key_LOCK_wsrep_thd;
+extern PSI_cond_key  key_COND_wsrep_thd;
+extern PSI_thread_key key_thread_handle_wsrep;
+#endif /* WITH_WSREP */
 
 extern PSI_rwlock_key key_rwlock_LOCK_grant, key_rwlock_LOCK_logger,
   key_rwlock_LOCK_sys_init_connect, key_rwlock_LOCK_sys_init_slave,
@@ -659,6 +660,9 @@ extern PSI_memory_key key_memory_Quick_ranges;
 extern PSI_memory_key key_memory_File_query_log_name;
 extern PSI_memory_key key_memory_Table_trigger_dispatcher;
 extern PSI_memory_key key_memory_show_slave_status_io_gtid_set;
+#ifdef WITH_WSREP
+extern PSI_memory_key key_memory_wsrep;
+#endif /* WITH_WSREP */
 extern PSI_memory_key key_memory_write_set_extraction;
 extern PSI_memory_key key_memory_thd_timer;
 extern PSI_memory_key key_memory_THD_Session_tracker;
@@ -786,13 +790,26 @@ extern PSI_stage_info stage_restoring_secondary_keys;
 extern PSI_stage_info stage_wsrep_writing_rows;
 extern PSI_stage_info stage_wsrep_deleting_rows;
 extern PSI_stage_info stage_wsrep_updating_rows;
+
 extern PSI_stage_info stage_wsrep_applying_writeset;
+extern PSI_stage_info stage_wsrep_applied_writeset;
+
 extern PSI_stage_info stage_wsrep_committing;
+extern PSI_stage_info stage_wsrep_committed;
+
 extern PSI_stage_info stage_wsrep_rolling_back;
+extern PSI_stage_info stage_wsrep_rolled_back;
+
 extern PSI_stage_info stage_wsrep_replicating_commit;
+extern PSI_stage_info stage_wsrep_write_set_replicated;
 extern PSI_stage_info stage_wsrep_waiting_on_replaying;
+extern PSI_stage_info stage_wsrep_replicate;
 extern PSI_stage_info stage_wsrep_pre_commit;
+extern PSI_stage_info stage_wsrep_pre_commit_cert_passed;
+
 extern PSI_stage_info stage_wsrep_preparing_for_TO_isolation;
+extern PSI_stage_info stage_wsrep_completed_TO_isolation;
+
 extern PSI_stage_info stage_wsrep_replaying_trx;
 extern PSI_stage_info stage_wsrep_applier_idle;
 extern PSI_stage_info stage_wsrep_in_rollback_thread;
@@ -889,6 +906,7 @@ extern mysql_rwlock_t LOCK_system_variables_hash;
 extern mysql_rwlock_t LOCK_consistent_snapshot;
 extern mysql_cond_t COND_manager;
 extern int32 thread_running;
+extern mysql_mutex_t LOCK_group_replication_handler;
 
 extern char *opt_ssl_ca, *opt_ssl_capath, *opt_ssl_cert, *opt_ssl_cipher,
             *opt_ssl_key, *opt_ssl_crl, *opt_ssl_crlpath, *opt_tls_version;
@@ -1059,7 +1077,7 @@ static inline THD *_current_thd(void)
 #define ER(X)         ER_THD(current_thd,X)
 
 #ifdef WITH_WSREP
-void* start_wsrep_THD(void*);
+extern "C" void *start_wsrep_THD(void*);
 #endif /* WITH_WSREP */
 
 #endif /* MYSQLD_INCLUDED */
