@@ -1,4 +1,4 @@
-/* Copyright (c) 2015, 2016, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -274,7 +274,8 @@ int channel_create(const char* channel,
   if (channel_info->auto_position)
   {
     lex_mi->auto_position= LEX_MASTER_INFO::LEX_MI_ENABLE;
-    if (mi && mi->is_auto_position())
+    if ((mi && mi->is_auto_position()) ||
+        channel_info->auto_position == RPL_SERVICE_SERVER_DEFAULT)
     {
       //So change master allows new configurations with a running SQL thread
       lex_mi->auto_position= LEX_MASTER_INFO::LEX_MI_UNCHANGED;
@@ -455,6 +456,8 @@ int channel_stop(const char* channel,
 
   int thread_mask= 0;
   int server_thd_mask= 0;
+  int error= 0;
+  bool thd_init= false;
   lock_slave_threads(mi);
 
   init_thread_mask(&server_thd_mask, mi, 0 /* not inverse*/);
@@ -472,16 +475,15 @@ int channel_stop(const char* channel,
 
   if (thread_mask == 0)
   {
-    mi->channel_unlock();
-    channel_map.unlock();
-    DBUG_RETURN(0);
+    goto end;
   }
 
-  bool thd_init= init_thread_context();
+  thd_init= init_thread_context();
 
-  int error= terminate_slave_threads(mi, thread_mask, timeout, false);
+  error= terminate_slave_threads(mi, thread_mask, timeout, false);
+
+end:
   unlock_slave_threads(mi);
-
   mi->channel_unlock();
   channel_map.unlock();
 
@@ -757,7 +759,7 @@ int channel_queue_packet(const char* channel,
 }
 
 int channel_wait_until_apply_queue_applied(const char* channel,
-                                           long long timeout)
+                                           double timeout)
 {
   DBUG_ENTER("channel_wait_until_apply_queue_applied(channel, timeout)");
 
@@ -917,4 +919,56 @@ int channel_get_retrieved_gtid_set(const char* channel,
   DBUG_RETURN(error);
 }
 
+bool channel_is_stopping(const char* channel,
+                         enum_channel_thread_types thd_type)
+{
+  bool is_stopping= false;
+  DBUG_ENTER("channel_is_stopping(channel, thd_type");
+
+  channel_map.rdlock();
+  Master_info *mi= channel_map.get_mi(channel);
+  if (mi == NULL)
+  {
+    channel_map.unlock();
+    DBUG_RETURN(false);
+  }
+
+  mi->channel_rdlock();
+
+  switch(thd_type)
+  {
+    case CHANNEL_NO_THD:
+      break;
+    case CHANNEL_RECEIVER_THREAD:
+      is_stopping= likely(mi->is_stopping.atomic_get());
+      break;
+    case CHANNEL_APPLIER_THREAD:
+      is_stopping= likely(mi->rli->is_stopping.atomic_get());
+      break;
+    default:
+      DBUG_ASSERT(0);
+  }
+
+  mi->channel_unlock();
+  channel_map.unlock();
+
+  DBUG_RETURN(is_stopping);
+}
+
+bool is_partial_transaction_on_channel_relay_log(const char *channel)
+{
+  DBUG_ENTER("is_partial_transaction_on_channel_relay_log(channel)");
+  channel_map.rdlock();
+  Master_info *mi= channel_map.get_mi(channel);
+  if (mi == NULL)
+  {
+    channel_map.unlock();
+    DBUG_RETURN(false);
+  }
+  mi->channel_rdlock();
+  bool ret= mi->transaction_parser.is_inside_transaction();
+  mi->channel_unlock();
+  channel_map.unlock();
+  DBUG_RETURN(ret);
+}
 #endif /* HAVE_REPLICATION */
