@@ -1,4 +1,4 @@
-/* Copyright (c) 2011, 2015, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2011, 2017, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -21,6 +21,7 @@
 #include "sql_class.h"             // THD
 #include "debug_sync.h"            // DEBUG_SYNC
 
+PSI_memory_key key_memory_Gtid_state_group_commit_sidno;
 
 int Gtid_state::clear(THD *thd)
 {
@@ -42,7 +43,7 @@ int Gtid_state::clear(THD *thd)
     thd->clear_error();
     ret= 0;
   }
-
+  next_free_gno= 1;
   DBUG_RETURN(ret);
 }
 
@@ -141,6 +142,58 @@ void Gtid_state::broadcast_owned_sidnos(const THD *thd)
 }
 
 
+void Gtid_state::update_commit_group(THD *first_thd)
+{
+  DBUG_ENTER("Gtid_state::update_commit_group");
+
+  /*
+    We are going to loop in all sessions of the group commit in order to avoid
+    being taking and releasing the global_sid_lock and sidno_lock for each
+    session.
+  */
+  DEBUG_SYNC(first_thd, "update_gtid_state_before_global_sid_lock");
+  global_sid_lock->rdlock();
+  DEBUG_SYNC(first_thd, "update_gtid_state_after_global_sid_lock");
+
+  update_gtids_impl_lock_sidnos(first_thd);
+
+  for (THD *thd= first_thd; thd != NULL; thd= thd->next_to_commit)
+  {
+    bool is_commit= (thd->commit_error != THD::CE_COMMIT_ERROR);
+
+    if (update_gtids_impl_do_nothing(thd) ||
+        (!is_commit && update_gtids_impl_check_skip_gtid_rollback(thd)))
+      continue;
+
+    bool more_trx_with_same_gtid_next= update_gtids_impl_begin(thd);
+
+    if (thd->owned_gtid.sidno == THD::OWNED_SIDNO_GTID_SET)
+    {
+      update_gtids_impl_own_gtid_set(thd, is_commit);
+    }
+    else if (thd->owned_gtid.sidno > 0)
+    {
+      update_gtids_impl_own_gtid(thd, is_commit);
+    }
+    else if (thd->owned_gtid.sidno == THD::OWNED_SIDNO_ANONYMOUS)
+    {
+      update_gtids_impl_own_anonymous(thd, &more_trx_with_same_gtid_next);
+    }
+    else
+    {
+      update_gtids_impl_own_nothing(thd);
+    }
+
+    update_gtids_impl_end(thd, more_trx_with_same_gtid_next);
+  }
+
+  update_gtids_impl_broadcast_and_unlock_sidnos();
+
+  global_sid_lock->unlock();
+
+  DBUG_VOID_RETURN;
+}
+
 void Gtid_state::update_on_commit(THD *thd)
 {
   DBUG_ENTER("Gtid_state::update_on_commit");
@@ -156,25 +209,8 @@ void Gtid_state::update_on_rollback(THD *thd)
 {
   DBUG_ENTER("Gtid_state::update_on_rollback");
 
-  /*
-    The administrative commands [CHECK|REPAIR|OPTIMIZE|ANALYZE] TABLE
-    are written to the binary log even when they fail.  When the
-    commands fail, they will call update_on_rollback; later they will
-    write the binary log.  But we must not do any of the things in
-    update_gtids_impl if we are going to write the binary log.  So
-    these statements set the skip_gtid_rollback flag, which tells
-    update_on_rollback to return early.  When the statements are
-    written to the binary log they will call update_on_commit as
-    usual.
-  */
-  if (thd->skip_gtid_rollback)
-  {
-    DBUG_PRINT("info", ("skipping gtid rollback because "
-                        "thd->skip_gtid_rollback is set"));
-    DBUG_VOID_RETURN;
-  }
-
-  update_gtids_impl(thd, false);
+  if (!update_gtids_impl_check_skip_gtid_rollback(thd))
+    update_gtids_impl(thd, false);
 
   DBUG_VOID_RETURN;
 }
@@ -184,65 +220,10 @@ void Gtid_state::update_gtids_impl(THD *thd, bool is_commit)
 {
   DBUG_ENTER("Gtid_state::update_gtids_impl");
 
-  /*
-    If we don't own anything, there is nothing to do, so we do an
-    early return.  Except if there is a GTID consistency violation;
-    then we need to decrement the counter, so then we go ahead and
-    call update_gtids_impl.
-  */
-  if (thd->owned_gtid.is_empty() && !thd->has_gtid_consistency_violation)
-  {
-    if (thd->variables.gtid_next.type == GTID_GROUP)
-      thd->variables.gtid_next.set_undefined();
-    DBUG_PRINT("info", ("skipping update_gtids_impl because "
-                        "thread does not own anything and does not violate "
-                        "gtid consistency"));
+  if (update_gtids_impl_do_nothing(thd))
     DBUG_VOID_RETURN;
-  }
 
-  /*
-    This variable is true for anonymous transactions, when the
-    'transaction' has been split into multiple transactions in the
-    binlog, and the present transaction is not the last one.
-
-    This means two things:
-
-    - We should not release anonymous ownership in case
-      gtid_next=anonymous.  If we did, it would be possible for user
-      to set GTID_MODE=ON from a concurrent transaction, making it
-      impossible to commit the current transaction.
-
-    - We should not decrease the counters for GTID-violating
-      statements.  If we did, it would be possible for a concurrent
-      client to set ENFORCE_GTID_CONSISTENCY=ON despite there is an
-      ongoing transaction that violates GTID consistency.
-
-    The flag is set in two cases:
-
-     1. We are committing the statement cache when there are more
-        changes in the transaction cache.
-
-        This happens either because a single statement in the
-        beginning of a transaction updates both transactional and
-        non-transactional tables, or because we are committing a
-        non-transactional update in the middle of a transaction when
-        binlog_direct_non_transactional_updates=1.
-
-        In this case, the flag is set further down in this function.
-
-     2. The statement is one of the special statements that may
-        generate multiple transactions: CREATE...SELECT, DROP TABLE,
-        DROP DATABASE. See comment for THD::owned_gtid in
-        sql/sql_class.h.
-
-        In this case, the THD::is_commit_in_middle_of_statement flag
-        is set by the caller and the flag becomes true here.
-  */
-  bool more_transactions_with_same_gtid_next=
-    thd->is_commit_in_middle_of_statement;
-  DBUG_PRINT("info", ("query='%s' thd->is_commit_in_middle_of_statement=%d",
-                      thd->query().str,
-                      thd->is_commit_in_middle_of_statement));
+  bool more_trx_with_same_gtid_next= update_gtids_impl_begin(thd);
 
   DEBUG_SYNC(thd, "update_gtid_state_before_global_sid_lock");
   global_sid_lock->rdlock();
@@ -250,138 +231,27 @@ void Gtid_state::update_gtids_impl(THD *thd, bool is_commit)
 
   if (thd->owned_gtid.sidno == THD::OWNED_SIDNO_GTID_SET)
   {
-#ifdef HAVE_GTID_NEXT_LIST
-    rpl_sidno prev_sidno= 0;
-    Gtid_set::Gtid_iterator git(&thd->owned_gtid_set);
-    Gtid g= git.get();
-    while (g.sidno != 0)
-    {
-      if (g.sidno != prev_sidno)
-        sid_locks.lock(g.sidno);
-      owned_gtids.remove_gtid(g);
-      git.next();
-      g= git.get();
-      if (is_commit)
-        executed_gtids._add_gtid(g);
-    }
-
-    if (is_commit && !thd->owned_gtid_set.is_empty())
-      thd->rpl_thd_ctx.session_gtids_ctx().
-        notify_after_gtid_executed_update(thd);
-
-    thd->variables.gtid_next.set_undefined();
-    thd->owned_gtid.dbug_print(NULL,
-                               "set owned_gtid (clear; old was gtid_set) "
-                               "in update_gtids_impl");
-    thd->clear_owned_gtids();
-#else
-    DBUG_ASSERT(0);
-#endif
+    update_gtids_impl_own_gtid_set(thd, is_commit);
   }
   else if (thd->owned_gtid.sidno > 0)
   {
-    lock_sidno(thd->owned_gtid.sidno);
-    DBUG_ASSERT(!executed_gtids.contains_gtid(thd->owned_gtid));
-    owned_gtids.remove_gtid(thd->owned_gtid);
-
-    if (is_commit)
-    {
-      DBUG_EXECUTE_IF(
-        "rpl_gtid_update_on_commit_simulate_out_of_memory",
-        DBUG_SET("+d,rpl_gtid_get_free_interval_simulate_out_of_memory"););
-      /*
-        Any session adds transaction owned GTID into global executed_gtids.
-
-        If binlog is disabled, we report @@GLOBAL.GTID_PURGED from
-        executed_gtids, since @@GLOBAL.GTID_PURGED and @@GLOBAL.GTID_EXECUTED
-        are always same, so we did not save gtid into lost_gtids for every
-        transaction for improving performance.
-
-        If binlog is enabled and log_slave_updates is disabled, slave
-        SQL thread or slave worker thread adds transaction owned GTID
-        into global executed_gtids, lost_gtids and gtids_only_in_table.
-      */
-      executed_gtids._add_gtid(thd->owned_gtid);
-      thd->rpl_thd_ctx.session_gtids_ctx().
-        notify_after_gtid_executed_update(thd);
-      if (thd->slave_thread && opt_bin_log && !opt_log_slave_updates)
-      {
-        lost_gtids._add_gtid(thd->owned_gtid);
-        gtids_only_in_table._add_gtid(thd->owned_gtid);
-      }
-    }
-
-    broadcast_owned_sidnos(thd);
-    unlock_owned_sidnos(thd);
-
-    thd->clear_owned_gtids();
-    if (thd->variables.gtid_next.type == GTID_GROUP)
-    {
-      DBUG_ASSERT(!more_transactions_with_same_gtid_next);
-      thd->variables.gtid_next.set_undefined();
-    }
-    else
-    {
-      /*
-        Can be UNDEFINED for statements where
-        gtid_pre_statement_checks skips the test for undefined,
-        e.g. ROLLBACK.
-      */
-      DBUG_ASSERT(thd->variables.gtid_next.type == AUTOMATIC_GROUP ||
-                  thd->variables.gtid_next.type == UNDEFINED_GROUP);
-    }
+    rpl_sidno sidno= thd->owned_gtid.sidno;
+    update_gtids_impl_lock_sidno(sidno);
+    update_gtids_impl_own_gtid(thd, is_commit);
+    update_gtids_impl_broadcast_and_unlock_sidno(sidno);
   }
   else if (thd->owned_gtid.sidno == THD::OWNED_SIDNO_ANONYMOUS)
   {
-    DBUG_ASSERT(thd->variables.gtid_next.type == ANONYMOUS_GROUP ||
-                thd->variables.gtid_next.type == AUTOMATIC_GROUP);
-    /*
-      If there is more in the transaction cache, set
-      more_transactions_with_same_gtid_next to indicate this.
-
-      See comment for the declaration of
-      more_transactions_with_same_gtid_next.
-    */
-    if (opt_bin_log)
-    {
-      // Needed before is_binlog_cache_empty.
-      thd->binlog_setup_trx_data();
-      if (!thd->is_binlog_cache_empty(true))
-      {
-        more_transactions_with_same_gtid_next= true;
-        DBUG_PRINT("info", ("Transaction cache is non-empty: setting "
-                            "more_transaction_with_same_gtid_next="
-                            "true."));
-      }
-    }
-    if (!(more_transactions_with_same_gtid_next &&
-          thd->variables.gtid_next.type == ANONYMOUS_GROUP))
-    {
-      release_anonymous_ownership();
-      thd->clear_owned_gtids();
-    }
+    update_gtids_impl_own_anonymous(thd, &more_trx_with_same_gtid_next);
   }
   else
   {
-    /*
-      Nothing is owned.  There are two cases when this happens:
-
-      - Normally, it is a rollback of an automatic transaction, so
-        then is_commit is false and gtid_next=automatic.
-
-      - There is also a corner case. A transaction with an empty gtid
-        should call gtid_end_transaction(...) to check a possible
-        violation of gtid consistency on commit, if it has set
-        has_gtid_consistency_violation to true.
-    */
-    DBUG_ASSERT(!is_commit || thd->has_gtid_consistency_violation);
-    DBUG_ASSERT(thd->variables.gtid_next.type == AUTOMATIC_GROUP);
+    update_gtids_impl_own_nothing(thd);
   }
 
   global_sid_lock->unlock();
 
-  if (!more_transactions_with_same_gtid_next)
-    end_gtid_violating_transaction(thd);
+  update_gtids_impl_end(thd, more_trx_with_same_gtid_next);
 
   thd->owned_gtid.dbug_print(NULL,
                              "set owned_gtid (clear) in update_gtids_impl");
@@ -442,20 +312,20 @@ bool Gtid_state::wait_for_gtid(THD *thd, const Gtid &gtid,
 
 
 bool Gtid_state::wait_for_gtid_set(THD *thd, Gtid_set* wait_for,
-                                   longlong timeout)
+                                   double timeout)
 {
   struct timespec abstime;
   DBUG_ENTER("Gtid_state::wait_for_gtid_set");
   DEBUG_SYNC(thd, "begin_wait_for_executed_gtid_set");
   wait_for->dbug_print("Waiting for");
-  DBUG_PRINT("info", ("Timeout %lld", timeout));
+  DBUG_PRINT("info", ("Timeout %f", timeout));
 
   global_sid_lock->assert_some_rdlock();
 
   DBUG_ASSERT(wait_for->get_sid_map() == global_sid_map);
 
   if (timeout > 0)
-    set_timespec(&abstime, timeout);
+    set_timespec_nsec(&abstime, (ulonglong) timeout * 1000000000ULL);
 
   /*
     Algorithm:
@@ -563,7 +433,32 @@ rpl_gno Gtid_state::get_automatic_gno(rpl_sidno sidno) const
 {
   DBUG_ENTER("Gtid_state::get_automatic_gno");
   Gtid_set::Const_interval_iterator ivit(&executed_gtids, sidno);
-  Gtid next_candidate= { sidno, 1 };
+  /*
+    When assigning new automatic GTIDs, we can optimize the assignment by start
+    searching an available GNO from the "supposed" next free one instead of
+    starting from 1.
+
+    This is useful mostly on systems having many transactions committing in
+    group asking for automatic GTIDs. When a GNO is assigned to be owned by a
+    transaction, it is not removed from the free intervals, but will be added
+    to the owned_gtids set. In this way, picking up the actual first free GNO
+    would often lead to getting a GNO already owned by other thread. This can
+    lead to many "tries" of getting a free and not owned yet GNO (a thread
+    would try N times, N being the sum of transactions in the FLUSH stage plus
+    the transactions in the COMMIT stage that didn't released their ownership
+    yet).
+
+    The optimization just set next_free_gno variable to the last assigned
+    GNO + 1, as this would be the common case without having transactions
+    rolling back. This is done at Gtid_state::generate_automatic_gtid.
+
+    In order to fill the gaps of GTID_EXECUTED when a transaction rolls back
+    releasing the ownership of a GTID, we check if the released GNO is smaller
+    than the next_free_gno at Gtid_state::update_gtids_impl_own_gtid function
+    to set next_free_gno with the "released" GNO in this case.
+  */
+  Gtid next_candidate= { sidno,
+                         sidno == get_server_sidno() ? next_free_gno : 1 };
   while (true)
   {
     const Gtid_set::Interval *iv= ivit.get();
@@ -571,6 +466,7 @@ rpl_gno Gtid_state::get_automatic_gno(rpl_sidno sidno) const
     while (next_candidate.gno < next_interval_start &&
            DBUG_EVALUATE_IF("simulate_gno_exhausted", false, true))
     {
+      DBUG_PRINT("debug",("Checking availability of gno= %llu", next_candidate.gno));
       if (owned_gtids.get_owner(next_candidate) == 0)
         DBUG_RETURN(next_candidate.gno);
       next_candidate.gno++;
@@ -581,7 +477,8 @@ rpl_gno Gtid_state::get_automatic_gno(rpl_sidno sidno) const
       my_error(ER_GNO_EXHAUSTED, MYF(0));
       DBUG_RETURN(-1);
     }
-    next_candidate.gno= iv->end;
+    if (next_candidate.gno <= iv->end)
+      next_candidate.gno= iv->end;
     ivit.next();
   }
 }
@@ -602,7 +499,8 @@ rpl_gno Gtid_state::get_last_executed_gno(rpl_sidno sidno) const
 
 enum_return_status Gtid_state::generate_automatic_gtid(THD *thd,
                                                        rpl_sidno specified_sidno,
-                                                       rpl_gno specified_gno)
+                                                       rpl_gno specified_gno,
+                                                       rpl_sidno *locked_sidno)
 {
   DBUG_ENTER("Gtid_state::generate_automatic_gtid");
   enum_return_status ret= RETURN_STATUS_OK;
@@ -612,7 +510,13 @@ enum_return_status Gtid_state::generate_automatic_gtid(THD *thd,
   DBUG_ASSERT(specified_gno >= 0);
   DBUG_ASSERT(thd->owned_gtid.is_empty());
 
-  sid_lock->rdlock();
+  bool locked_sidno_was_passed_null = (locked_sidno == NULL);
+
+  if (locked_sidno_was_passed_null)
+    sid_lock->rdlock();
+  else
+    /* The caller must lock the sid_lock when locked_sidno is passed */
+    sid_lock->assert_some_lock();
 
   // If GTID_MODE = ON_PERMISSIVE or ON, generate a new GTID
   if (get_gtid_mode(GTID_MODE_LOCK_SID) >= GTID_MODE_ON_PERMISSIVE)
@@ -631,17 +535,45 @@ enum_return_status Gtid_state::generate_automatic_gtid(THD *thd,
       automatic_gtid.sidno= get_server_sidno();
 #endif /* WITH_WSREP */
 
-    lock_sidno(automatic_gtid.sidno);
+    /*
+      We need to lock the sidno if locked_sidno wasn't passed as paramenter
+      or the already locked sidno doesn't match the one to generate the new
+      automatic GTID.
+    */
+    bool need_to_lock_sidno= (locked_sidno_was_passed_null ||
+                              *locked_sidno != automatic_gtid.sidno);
+    if (need_to_lock_sidno)
+    {
+      /*
+        When locked_sidno contains a value greater than zero we must release
+        the current locked sidno. This should not happen with current code, as
+        the server only generates automatic GTIDs with server's UUID as sid.
+      */
+      if (!locked_sidno_was_passed_null && *locked_sidno != 0)
+        unlock_sidno(*locked_sidno);
+      lock_sidno(automatic_gtid.sidno);
+      /* Update the locked_sidno, so the caller would know what to unlock */
+      if (!locked_sidno_was_passed_null)
+        *locked_sidno= automatic_gtid.sidno;
+    }
 
     if (automatic_gtid.gno == 0)
+    {
       automatic_gtid.gno= get_automatic_gno(automatic_gtid.sidno);
+      if (automatic_gtid.sidno == get_server_sidno() &&
+          automatic_gtid.gno != -1)
+        next_free_gno= automatic_gtid.gno + 1;
+    }
 
     if (automatic_gtid.gno != -1)
       acquire_ownership(thd, automatic_gtid);
     else
       ret= RETURN_STATUS_REPORTED_ERROR;
 
-    unlock_sidno(automatic_gtid.sidno);
+    /* The caller will unlock the sidno_lock if locked_sidno was passed */
+    if (locked_sidno_was_passed_null)
+      unlock_sidno(automatic_gtid.sidno);
+
   }
   else
   {
@@ -654,7 +586,9 @@ enum_return_status Gtid_state::generate_automatic_gtid(THD *thd,
                                "set owned_gtid (anonymous) in generate_automatic_gtid");
   }
 
-  sid_lock->unlock();
+  /* The caller will unlock the sid_lock if locked_sidno was passed */
+  if (locked_sidno_was_passed_null)
+    sid_lock->unlock();
 
   gtid_set_performance_schema_values(thd);
 
@@ -708,6 +642,7 @@ enum_return_status Gtid_state::ensure_sidno()
     PROPAGATE_REPORTED_ERROR(lost_gtids.ensure_sidno(sidno));
     PROPAGATE_REPORTED_ERROR(owned_gtids.ensure_sidno(sidno));
     PROPAGATE_REPORTED_ERROR(sid_locks.ensure_index(sidno));
+    PROPAGATE_REPORTED_ERROR(ensure_commit_group_sidnos(sidno));
     sidno= sid_map->get_max_sidno();
     DBUG_ASSERT(executed_gtids.get_max_sidno() >= sidno);
     DBUG_ASSERT(gtids_only_in_table.get_max_sidno() >= sidno);
@@ -715,6 +650,7 @@ enum_return_status Gtid_state::ensure_sidno()
     DBUG_ASSERT(lost_gtids.get_max_sidno() >= sidno);
     DBUG_ASSERT(owned_gtids.get_max_sidno() >= sidno);
     DBUG_ASSERT(sid_locks.get_max_index() >= sidno);
+    DBUG_ASSERT(commit_group_sidnos.size() >= (unsigned int)sidno);
   }
   RETURN_OK;
 }
@@ -760,7 +696,7 @@ int Gtid_state::init()
 {
   DBUG_ENTER("Gtid_state::init()");
 
-  global_sid_lock->assert_some_lock();
+  global_sid_lock->assert_some_wrlock();
 
   rpl_sid server_sid;
   if (server_sid.parse(server_uuid) != 0)
@@ -769,7 +705,7 @@ int Gtid_state::init()
   if (sidno <= 0)
     DBUG_RETURN(1);
   server_sidno= sidno;
-
+  next_free_gno= 1;
   DBUG_RETURN(0);
 }
 
@@ -874,3 +810,245 @@ bool Gtid_state::warn_or_err_on_modify_gtid_table(THD *thd, TABLE_LIST *table)
   DBUG_RETURN(ret);
 }
 #endif
+
+bool Gtid_state::update_gtids_impl_check_skip_gtid_rollback(THD *thd)
+{
+  if (thd->skip_gtid_rollback)
+  {
+    DBUG_PRINT("info", ("skipping gtid rollback because "
+                        "thd->skip_gtid_rollback is set"));
+    return true;
+  }
+  return false;
+}
+
+bool Gtid_state::update_gtids_impl_do_nothing(THD *thd)
+{
+  if (thd->owned_gtid.is_empty() && !thd->has_gtid_consistency_violation)
+  {
+    if (thd->variables.gtid_next.type == GTID_GROUP)
+      thd->variables.gtid_next.set_undefined();
+    DBUG_PRINT("info", ("skipping update_gtids_impl because "
+                        "thread does not own anything and does not violate "
+                        "gtid consistency"));
+
+    return true;
+  }
+  return false;
+}
+
+bool Gtid_state::update_gtids_impl_begin(THD *thd)
+{
+#ifndef DBUG_OFF
+  if (current_thd != thd)
+    mysql_mutex_lock(&thd->LOCK_thd_query);
+  DBUG_PRINT("info", ("query='%s' thd->is_commit_in_middle_of_statement=%d",
+                      thd->query().str,
+                      thd->is_commit_in_middle_of_statement));
+  if (current_thd != thd)
+    mysql_mutex_unlock(&thd->LOCK_thd_query);
+#endif
+  return thd->is_commit_in_middle_of_statement;
+}
+
+void Gtid_state::update_gtids_impl_own_gtid_set(THD *thd, bool is_commit)
+{
+#ifdef HAVE_GTID_NEXT_LIST
+  rpl_sidno prev_sidno= 0;
+  Gtid_set::Gtid_iterator git(&thd->owned_gtid_set);
+  Gtid g= git.get();
+  while (g.sidno != 0)
+  {
+    if (g.sidno != prev_sidno)
+      sid_locks.lock(g.sidno);
+    owned_gtids.remove_gtid(g);
+    git.next();
+    g= git.get();
+    if (is_commit)
+      executed_gtids._add_gtid(g);
+  }
+
+  if (is_commit && !thd->owned_gtid_set.is_empty())
+    thd->rpl_thd_ctx.session_gtids_ctx().
+      notify_after_gtid_executed_update(thd);
+
+  thd->variables.gtid_next.set_undefined();
+  thd->owned_gtid.dbug_print(NULL,
+                             "set owned_gtid (clear; old was gtid_set) "
+                             "in update_gtids_impl");
+  thd->clear_owned_gtids();
+#else
+  DBUG_ASSERT(0);
+#endif
+}
+
+void Gtid_state::update_gtids_impl_lock_sidno(rpl_sidno sidno)
+{
+  DBUG_ASSERT(sidno > 0);
+  DBUG_PRINT("info",("Locking sidno %d", sidno));
+  lock_sidno(sidno);
+}
+
+void Gtid_state::update_gtids_impl_lock_sidnos(THD *first_thd)
+{
+  /* Define which sidnos should be locked to be updated */
+  for (THD *thd= first_thd; thd != NULL; thd= thd->next_to_commit)
+  {
+    if (thd->owned_gtid.sidno > 0)
+    {
+      DBUG_PRINT("info",("Setting sidno %d to be locked",
+                         thd->owned_gtid.sidno));
+      commit_group_sidnos[thd->owned_gtid.sidno]= true;
+    }
+    else if (thd->owned_gtid.sidno == THD::OWNED_SIDNO_GTID_SET)
+#ifdef HAVE_GTID_NEXT_LIST
+      for (rpl_sidno i= 1; i < thd->owned_gtid_set.max_sidno; i++)
+        if (owned_gtid_set.contains_sidno(i))
+          commit_group_sidnos[i]= true;
+#else
+      DBUG_ASSERT(0);
+#endif
+  }
+
+  /* Take the sidno_locks in order */
+  for (rpl_sidno i= 1; i < (rpl_sidno)commit_group_sidnos.size(); i++)
+    if (commit_group_sidnos[i])
+      update_gtids_impl_lock_sidno(i);
+}
+
+void Gtid_state::update_gtids_impl_own_gtid(THD *thd, bool is_commit)
+{
+  assert_sidno_lock_owner(thd->owned_gtid.sidno);
+  DBUG_ASSERT(!executed_gtids.contains_gtid(thd->owned_gtid));
+  owned_gtids.remove_gtid(thd->owned_gtid);
+
+  if (is_commit)
+  {
+    DBUG_EXECUTE_IF(
+      "rpl_gtid_update_on_commit_simulate_out_of_memory",
+      DBUG_SET("+d,rpl_gtid_get_free_interval_simulate_out_of_memory"););
+    /*
+      Any session adds transaction owned GTID into global executed_gtids.
+
+      If binlog is disabled, we report @@GLOBAL.GTID_PURGED from
+      executed_gtids, since @@GLOBAL.GTID_PURGED and @@GLOBAL.GTID_EXECUTED
+      are always same, so we did not save gtid into lost_gtids for every
+      transaction for improving performance.
+
+      If binlog is enabled and log_slave_updates is disabled, slave
+      SQL thread or slave worker thread adds transaction owned GTID
+      into global executed_gtids, lost_gtids and gtids_only_in_table.
+    */
+    executed_gtids._add_gtid(thd->owned_gtid);
+    thd->rpl_thd_ctx.session_gtids_ctx().
+      notify_after_gtid_executed_update(thd);
+    if (thd->slave_thread && opt_bin_log && !opt_log_slave_updates)
+    {
+      lost_gtids._add_gtid(thd->owned_gtid);
+      gtids_only_in_table._add_gtid(thd->owned_gtid);
+    }
+  }
+  else
+  {
+    if (thd->owned_gtid.sidno == server_sidno &&
+        next_free_gno > thd->owned_gtid.gno)
+      next_free_gno= thd->owned_gtid.gno;
+  }
+
+  thd->clear_owned_gtids();
+  if (thd->variables.gtid_next.type == GTID_GROUP)
+  {
+    DBUG_ASSERT(!thd->is_commit_in_middle_of_statement);
+    thd->variables.gtid_next.set_undefined();
+  }
+  else
+  {
+    /*
+      Can be UNDEFINED for statements where
+      gtid_pre_statement_checks skips the test for undefined,
+      e.g. ROLLBACK.
+    */
+    DBUG_ASSERT(thd->variables.gtid_next.type == AUTOMATIC_GROUP ||
+                thd->variables.gtid_next.type == UNDEFINED_GROUP);
+  }
+}
+
+void Gtid_state::update_gtids_impl_broadcast_and_unlock_sidno(rpl_sidno sidno)
+{
+  DBUG_PRINT("info",("Unlocking sidno %d", sidno));
+  broadcast_sidno(sidno);
+  unlock_sidno(sidno);
+}
+
+void Gtid_state::update_gtids_impl_broadcast_and_unlock_sidnos()
+{
+  for (rpl_sidno i= 1; i < (rpl_sidno)commit_group_sidnos.size(); i++)
+    if (commit_group_sidnos[i])
+    {
+      update_gtids_impl_broadcast_and_unlock_sidno(i);
+      commit_group_sidnos[i]= false;
+    }
+}
+
+void Gtid_state::update_gtids_impl_own_anonymous(THD* thd,
+                                                 bool *more_trx)
+{
+  DBUG_ASSERT(thd->variables.gtid_next.type == ANONYMOUS_GROUP ||
+              thd->variables.gtid_next.type == AUTOMATIC_GROUP);
+  /*
+    If there is more in the transaction cache, set more_trx to indicate this.
+
+    See comment for the update_gtids_impl_begin function.
+  */
+  if (opt_bin_log)
+  {
+    // Needed before is_binlog_cache_empty.
+    thd->binlog_setup_trx_data();
+    if (!thd->is_binlog_cache_empty(true))
+    {
+      *more_trx= true;
+      DBUG_PRINT("info", ("Transaction cache is non-empty: setting "
+                          "more_transaction_with_same_gtid_next="
+                          "true."));
+    }
+  }
+  if (!(*more_trx &&
+        thd->variables.gtid_next.type == ANONYMOUS_GROUP))
+  {
+    release_anonymous_ownership();
+    thd->clear_owned_gtids();
+  }
+}
+
+void Gtid_state::update_gtids_impl_own_nothing(THD *thd)
+{
+  DBUG_ASSERT(thd->commit_error != THD::CE_COMMIT_ERROR ||
+              thd->has_gtid_consistency_violation);
+  DBUG_ASSERT(thd->variables.gtid_next.type == AUTOMATIC_GROUP);
+}
+
+void Gtid_state::update_gtids_impl_end(THD *thd, bool more_trx)
+{
+  if (!more_trx)
+    end_gtid_violating_transaction(thd);
+}
+
+enum_return_status Gtid_state::ensure_commit_group_sidnos(rpl_sidno sidno)
+{
+  DBUG_ENTER("Gtid_state::ensure_commit_group_sidnos");
+  sid_lock->assert_some_wrlock();
+  /*
+    As we use the sidno as index of commit_group_sidnos and there is no
+    sidno=0, the array size must be at least sidno + 1.
+  */
+  while ((commit_group_sidnos.size()) < (size_t)sidno + 1)
+  {
+    if (commit_group_sidnos.push_back(false))
+      goto error;
+  }
+  RETURN_OK;
+error:
+  BINLOG_ERROR(("Out of memory."), (ER_OUT_OF_RESOURCES, MYF(0)));
+  RETURN_REPORTED_ERROR;
+
+}
