@@ -89,7 +89,8 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
   if (thd->killed == THD::KILL_CONNECTION &&
       thd->wsrep_conflict_state != REPLAYING)
   {
-    WSREP_INFO("applier has been aborted, skipping apply_rbr: %lld",
+    WSREP_INFO("Applier aborted. Skipping apply event while processing"
+               " write-set: %lld",
                (long long) wsrep_thd_trx_seqno(thd));
     DBUG_RETURN(WSREP_CB_FAILURE);
   }
@@ -100,8 +101,9 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
     thd->wsrep_conflict_state= NO_CONFLICT;
   mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
 
-  if (!buf_len) WSREP_DEBUG("empty rbr buffer to apply: %lld",
-                            (long long) wsrep_thd_trx_seqno(thd));
+  if (!buf_len)
+    WSREP_DEBUG("Empty apply event found while processing write-set: %lld",
+                (long long) wsrep_thd_trx_seqno(thd));
 
   while(buf_len)
   {
@@ -111,7 +113,7 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
 
     if (!ev)
     {
-      WSREP_ERROR("applier could not read binlog event, seqno: %lld, len: %zu",
+      WSREP_ERROR("Applier could not read binlog event, seqno: %lld, len: %zu",
                   (long long)wsrep_thd_trx_seqno(thd), buf_len);
       rcode= 1;
       goto error;
@@ -130,22 +132,44 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
         delete ev;
         continue;
       }
+      /*
+         gtid_pre_statement_checks will fail on the subsequent statement
+         if the bits below are set. So we don't mark the thd to run in
+         transaction mode yet, and assume there will be such a "BEGIN"
+         log event that will set those appropriately.
+      */
+      thd->variables.option_bits&= ~OPTION_BEGIN;
+      thd->server_status&= ~SERVER_STATUS_IN_TRANS;
+      assert(event == 1);
     }
     default:
       break;
     }
 
     thd->server_id = ev->server_id; // use the original server id for logging
+    thd->unmasked_server_id = ev->common_header->unmasked_server_id;
     thd->set_time();                // time the query
-
-    Transaction_ctx *trn_ctx= thd->get_transaction();
-    wsrep_xid_init(trn_ctx->xid_state()->get_xid(),
+    wsrep_xid_init(thd->get_transaction()->xid_state()->get_xid(),
                    thd->wsrep_trx_meta.gtid.uuid,
                    thd->wsrep_trx_meta.gtid.seqno);
-    thd->lex->set_current_select(0);
+    thd->lex->set_current_select(NULL);
     if (!ev->common_header->when.tv_sec)
       my_micro_time_to_timeval(my_micro_time(), &ev->common_header->when);
-    ev->thd = thd;
+    ev->thd = thd; // because up to this point, ev->thd == 0
+
+    set_timespec_nsec(&thd->wsrep_rli->ts_exec[0], 0);
+    thd->wsrep_rli->stats_read_time += diff_timespec(&thd->wsrep_rli->ts_exec[0],
+                                                     &thd->wsrep_rli->ts_exec[1]);
+
+    /* MySQL slave "Sleeps if needed, and unlocks rli->data_lock."
+     * at this point. But this does not apply for wsrep, we just do the unlock part
+     * of sql_delay_event()
+     *
+     * if (sql_delay_event(ev, thd, rli))
+     */
+    //mysql_mutex_assert_owner(&rli->data_lock);
+    //mysql_mutex_unlock(&rli->data_lock);
+
     exec_res = ev->apply_event(thd->wsrep_rli);
     DBUG_PRINT("info", ("exec_event result: %d", exec_res));
 
@@ -159,12 +183,19 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
       delete ev;
       goto error;
     }
+
+    set_timespec_nsec(&thd->wsrep_rli->ts_exec[1], 0);
+    thd->wsrep_rli->stats_exec_time += diff_timespec(&thd->wsrep_rli->ts_exec[1],
+                                          &thd->wsrep_rli->ts_exec[0]);
+
+    DBUG_PRINT("info", ("wsrep apply_event error = %d", exec_res));
     event++;
 
     if (thd->wsrep_conflict_state!= NO_CONFLICT &&
         thd->wsrep_conflict_state!= REPLAYING)
-      WSREP_WARN("conflict state after RBR event applying: %d, %lld",
-                 thd->wsrep_query_state, (long long)wsrep_thd_trx_seqno(thd));
+      WSREP_WARN("conflict state after RBR event applying: %s, %lld",
+                 wsrep_get_query_state(thd->wsrep_query_state),
+                 (long long)wsrep_thd_trx_seqno(thd));
 
     if (thd->wsrep_conflict_state == MUST_ABORT) {
       WSREP_WARN("RBR event apply failed, rolling back: %lld",
@@ -194,7 +225,8 @@ static wsrep_cb_status_t wsrep_apply_events(THD*        thd,
   assert(thd->wsrep_exec_mode== REPL_RECV);
 
   if (thd->killed == THD::KILL_CONNECTION)
-    WSREP_INFO("applier aborted: %lld", (long long)wsrep_thd_trx_seqno(thd));
+    WSREP_INFO("applier aborted while processing write-set: %lld",
+               (long long)wsrep_thd_trx_seqno(thd));
 
   if (rcode) DBUG_RETURN(WSREP_CB_FAILURE);
   DBUG_RETURN(WSREP_CB_SUCCESS);
@@ -220,15 +252,12 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
 
   thd->wsrep_trx_meta = *meta;
 
-  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_applying_writeset);
-
-#ifdef WITH_WSREP
-#ifdef WSREP_PROC_INFO
-  if (WSREP(thd))
-    WSREP_DEBUG("wsrep: applying write set (%lld)", (long long) wsrep_thd_trx_seqno(thd));
-#endif /* WSREP_PROC_INFO */
-#endif /* WITH_WSREP */
-
+  THD_STAGE_INFO(thd, stage_wsrep_applying_writeset);
+  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+           "wsrep: applying write-set (%lld)",
+           (long long)wsrep_thd_trx_seqno(thd));
+  WSREP_DEBUG("%s", thd->wsrep_info);
+  thd_proc_info(thd, thd->wsrep_info);
 
   /* tune FK and UK checking policy */
   if (wsrep_slave_UK_checks == FALSE) 
@@ -252,18 +281,13 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
   }
   wsrep_cb_status_t rcode(wsrep_apply_events(thd, buf, buf_len));
 
-  THD_STAGE_INFO_GUARD_LEAVE();
-
-  /* Provide this for some galera test cases that check the thread state */
-  DBUG_EXECUTE_IF("thd_proc_info.wsrep_apply_cb",
-                  { thd_proc_info(thd, "wsrep: applied write set"); };);
-#ifdef WITH_WSREP
-#ifdef WSREP_PROC_INFO
-  if (WSREP(thd))
-    WSREP_DEBUG("wsrep: applied write set (%lld)", (long long) wsrep_thd_trx_seqno(thd));
-#endif /* WSREP_PROC_INFO */
-#endif /* WITH_WSREP */
-
+  THD_STAGE_INFO(thd, stage_wsrep_applied_writeset);
+  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+           "wsrep: %s write set (%lld)",
+           rcode == WSREP_CB_SUCCESS ? "applied" : "failed to apply",
+           (long long)wsrep_thd_trx_seqno(thd));
+  WSREP_DEBUG("%s", thd->wsrep_info);
+  thd_proc_info(thd, thd->wsrep_info);
 
   if (WSREP_CB_SUCCESS != rcode)
   {
@@ -274,10 +298,10 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
   while ((tmp = thd->temporary_tables))
   {
     WSREP_DEBUG("Applier %u, has temporary tables: %s.%s",
-                  thd->thread_id(), 
-                  (tmp->s) ? tmp->s->db.str : "void",
-                  (tmp->s) ? tmp->s->table_name.str : "void");
-      close_temporary_table(thd, tmp, 1, 1);    
+                thd->thread_id(), 
+                (tmp->s) ? tmp->s->db.str : "void",
+                (tmp->s) ? tmp->s->table_name.str : "void");
+    close_temporary_table(thd, tmp, 1, 1);    
   }
 
   return rcode;
@@ -285,30 +309,23 @@ wsrep_cb_status_t wsrep_apply_cb(void* const             ctx,
 
 static wsrep_cb_status_t wsrep_commit(THD* const thd)
 {
-  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_committing);
-#ifdef WITH_WSREP
-#ifdef WSREP_PROC_INFO
-  if (WSREP(thd))
-    WSREP_DEBUG("wsrep: committing (%lld)", (long long) wsrep_thd_trx_seqno(thd));
-#endif /* WSREP_PROC_INFO */
-#endif /* WITH_WSREP */
-
+  THD_STAGE_INFO(thd, stage_wsrep_committing);
+  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+           "wsrep: committing write set (%lld)",
+           (long long)wsrep_thd_trx_seqno(thd));
+  WSREP_DEBUG("%s", thd->wsrep_info);
+  thd_proc_info(thd, thd->wsrep_info);
 
   wsrep_cb_status_t const rcode(trans_commit(thd) ?
                                 WSREP_CB_FAILURE : WSREP_CB_SUCCESS);
 
-  THD_STAGE_INFO_GUARD_LEAVE();
-#ifdef WITH_WSREP
-#ifdef WSREP_PROC_INFO
-  if (WSREP(thd))
-    WSREP_DEBUG("wsrep: committed (%lld)", (long long) wsrep_thd_trx_seqno(thd));
-#endif /* WSREP_PROC_INFO */
-#endif /* WITH_WSREP */
-
-
-  /* Provide this for some galera test cases that check the thread state */
-  DBUG_EXECUTE_IF("thd_proc_info.wsrep_commit",
-                  { thd_proc_info(thd, "wsrep: committed"); };);
+  THD_STAGE_INFO(thd, stage_wsrep_committed);
+  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+           "wsrep: %s write set (%lld)",
+           (rcode == WSREP_CB_SUCCESS ? "committed" : "failed to commit"),
+            (long long)wsrep_thd_trx_seqno(thd));
+  WSREP_DEBUG("%s", thd->wsrep_info);
+  thd_proc_info(thd, thd->wsrep_info);
 
   if (WSREP_CB_SUCCESS == rcode)
   {
@@ -326,13 +343,12 @@ static wsrep_cb_status_t wsrep_commit(THD* const thd)
 
 static wsrep_cb_status_t wsrep_rollback(THD* const thd)
 {
-  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_rolling_back);
-#ifdef WITH_WSREP
-#ifdef WSREP_PROC_INFO
-  if (WSREP(thd))
-    WSREP_DEBUG("wsrep: rolling back (%lld)", (long long) wsrep_thd_trx_seqno(thd));
-#endif /* WSREP_PROC_INFO */
-#endif /* WITH_WSREP */
+  THD_STAGE_INFO(thd, stage_wsrep_rolling_back);
+  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+           "wsrep: rolling back write set (%lld)",
+           (long long)wsrep_thd_trx_seqno(thd));
+  WSREP_DEBUG("%s", thd->wsrep_info);
+  thd_proc_info(thd, thd->wsrep_info);
 
   /* Check for comments in Relay_log_info::cleanup_context */
   trans_rollback_stmt(thd);
@@ -340,12 +356,13 @@ static wsrep_cb_status_t wsrep_rollback(THD* const thd)
   wsrep_cb_status_t const rcode(trans_rollback(thd) ?
                                 WSREP_CB_FAILURE : WSREP_CB_SUCCESS);
 
-#ifdef WITH_WSREP
-#ifdef WSREP_PROC_INFO
-  if (WSREP(thd))
-    WSREP_DEBUG("wsrep: rolled back (%lld)", (long long) wsrep_thd_trx_seqno(thd));
-#endif /* WSREP_PROC_INFO */
-#endif /* WITH_WSREP */
+  THD_STAGE_INFO(thd, stage_wsrep_rolled_back);
+  snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+           "wsrep: %s write set (%lld)",
+           rcode == WSREP_CB_SUCCESS ? "rolled back" : "failed to rollback",
+           (long long)wsrep_thd_trx_seqno(thd));
+  WSREP_DEBUG("%s", thd->wsrep_info);
+  thd_proc_info(thd, thd->wsrep_info);
 
   thd->wsrep_rli->cleanup_context(thd, 0);
   thd->variables.gtid_next.set_automatic();
@@ -354,6 +371,7 @@ static wsrep_cb_status_t wsrep_rollback(THD* const thd)
 }
 
 wsrep_cb_status_t wsrep_commit_cb(void*         const     ctx,
+                                  const void*             trx_handle,
                                   uint32_t      const     flags,
                                   const wsrep_trx_meta_t* meta,
                                   wsrep_bool_t* const     exit,
@@ -361,6 +379,14 @@ wsrep_cb_status_t wsrep_commit_cb(void*         const     ctx,
 {
   THD* const thd((THD*)ctx);
 
+  /* Applier transaction delays entering CommitMonitor so
+  cache the needed params that can aid entering CommitMonitor
+  post prepare stage.
+  Replay of local transaction uses the same path as applying of
+  transaction but CommitMonitor protocol is different for it. */
+  if (trx_handle && thd->wsrep_conflict_state != REPLAYING)
+    thd->wsrep_ws_handle.opaque= const_cast<void*>(trx_handle);
+
   assert(meta->gtid.seqno == wsrep_thd_trx_seqno(thd));
 
   wsrep_cb_status_t rcode;
@@ -398,7 +424,8 @@ wsrep_cb_status_t wsrep_commit_cb(void*         const     ctx,
     if (wsrep_slave_count_change < 0)
     {
       wsrep_slave_count_change++;
-      WSREP_DEBUG("Closing applier thread, to close %d", abs(wsrep_slave_count_change));
+      WSREP_DEBUG("Closing applier thread(s). Yet to close %d",
+                  abs(wsrep_slave_count_change));
       *exit = true;
     }
     mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
@@ -406,13 +433,9 @@ wsrep_cb_status_t wsrep_commit_cb(void*         const     ctx,
 
   if (*exit == false && thd->wsrep_applier)
   {
-    /* This should be set as part of trans_begin and not explicitly.
-    Note sure why Galera thought of setting it here but it conflicts
-    as the first BEGIN statement is now intepreted wrongly as statement
-    inside an active transaction. */
-    /* From trans_begin() */
-    // thd->variables.option_bits|= OPTION_BEGIN;
-    //thd->server_status|= SERVER_STATUS_IN_TRANS;
+    /* From trans_begin(). Also check the comment at line:134 when/why these bits are unset. */
+    thd->variables.option_bits|= OPTION_BEGIN;
+    thd->server_status|= SERVER_STATUS_IN_TRANS;
     thd->wsrep_apply_toi= false;
   }
 
