@@ -22,24 +22,30 @@
 
   This function quite the same as MYSQL_BIN_LOG::write_cache(),
   with the exception that here we write in buffer instead of log file.
+
+  @params
+    cache   - IO cahce to read events from
+    buf     - buffer where to write the events, may contain some events
+              at call time, this function appends events in the end
+    buf_len - in input, this tells the position where to append events
+              (0 is in the begin)
+              in output, function sets the actual length of buffer after
+              all appends
  */
 int wsrep_write_cache_buf(IO_CACHE *cache, uchar **buf, size_t *buf_len)
 {
-  *buf= NULL;
-  *buf_len= 0;
-
   my_off_t const saved_pos(my_b_tell(cache));
 
   if (reinit_io_cache(cache, READ_CACHE, 0, 0, 0))
   {
-    WSREP_ERROR("failed to initialize io-cache");
+    WSREP_ERROR("Failed to initialize io-cache");
     return ER_ERROR_ON_WRITE;
   }
 
   uint length = my_b_bytes_in_cache(cache);
   if (unlikely(0 == length)) length = my_b_fill(cache);
 
-  size_t total_length = 0;
+  size_t total_length = *buf_len;
 
   if (likely(length > 0)) do
   {
@@ -52,17 +58,18 @@ int wsrep_write_cache_buf(IO_CACHE *cache, uchar **buf, size_t *buf_len)
       */
       if (total_length > wsrep_max_ws_size)
       {
-          WSREP_WARN("transaction size limit (%lu) exceeded: %zu",
+          WSREP_WARN("Transaction/Write-set size limit (%lu) exceeded: %zu",
                      wsrep_max_ws_size, total_length);
           goto error;
       }
 
-      uchar* tmp = (uchar *)my_realloc(PSI_NOT_INSTRUMENTED,
-                                       *buf, total_length, MYF(0));
+      uchar* tmp = (uchar *)my_realloc(key_memory_wsrep, *buf, total_length, MYF(0));
       if (!tmp)
       {
-          WSREP_ERROR("could not (re)allocate buffer: %zu + %u",
-                      *buf_len, length);
+          WSREP_ERROR("Fail to allocate/reallocate memory to hold"
+                      " write-set for replication."
+                      " Existing Size: %zu, Requested Size: %lu",
+                      *buf_len, (long unsigned) total_length);
           goto error;
       }
       *buf = tmp;
@@ -74,13 +81,13 @@ int wsrep_write_cache_buf(IO_CACHE *cache, uchar **buf, size_t *buf_len)
 
   if (reinit_io_cache(cache, WRITE_CACHE, saved_pos, 0, 0))
   {
-    WSREP_WARN("failed to initialize io-cache");
+    WSREP_WARN("Failed to initialize io-cache");
     goto cleanup;
   }
 
   if (reinit_io_cache(cache, WRITE_CACHE, saved_pos, 0, 0))
   {
-    WSREP_ERROR("failed to initialize io-cache");
+    WSREP_ERROR("Failed to initialize io-cache");
     goto cleanup;
   }
 
@@ -89,7 +96,7 @@ int wsrep_write_cache_buf(IO_CACHE *cache, uchar **buf, size_t *buf_len)
 error:
   if (reinit_io_cache(cache, WRITE_CACHE, saved_pos, 0, 0))
   {
-    WSREP_WARN("failed to initialize io-cache");
+    WSREP_WARN("Failed to initialize io-cache");
   }
 cleanup:
   my_free(*buf);
@@ -145,7 +152,7 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
 
     if (reinit_io_cache(cache, READ_CACHE, 0, 0, 0))
     {
-        WSREP_ERROR("failed to initialize io-cache");
+        WSREP_ERROR("Failed to initialize io-cache");
         return ER_ERROR_ON_WRITE;
     }
 
@@ -159,24 +166,41 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
     size_t used(0);
 
     uint length(my_b_bytes_in_cache(cache));
-    if (unlikely(0 == length)) length = my_b_fill(cache);
 
     /* Start 5.7 MySQL server will write GTID directly to binlog file.
     This means GTID event is not present in the case if GTID is set explictly.
-    Let's construct the GTID event and prefix it to wsrep-binlog-buffer. */
-    if (thd->variables.gtid_next.type != AUTOMATIC_GROUP)
+    We have cached the said GTID event which is then written before filling up
+    the cache with the real-data-change event. */
+    if (thd->wsrep_gtid_event_buf)
     {
-      Gtid_log_event gtid_event(thd, true, 0, 0);
-      uchar gtid_buf[Gtid_log_event::MAX_EVENT_LENGTH];
-      uint32 gtid_len = gtid_event.write_to_memory(gtid_buf);
+      if (thd->wsrep_gtid_event_buf_len < allocated)
+      {
+        memcpy(stack_buf, thd->wsrep_gtid_event_buf, thd->wsrep_gtid_event_buf_len);
+      }
+      else
+      {
+        uchar* tmp = (uchar *)my_realloc(key_memory_wsrep, heap_buf,
+                                         thd->wsrep_gtid_event_buf_len, MYF(0));
+        if (!tmp)
+        {
+          WSREP_ERROR("Failed to allocate/reallocate buffer to hold GTID event"
+                      " Existing Size: %zu, Requested Size: %lu",
+                      allocated, thd->wsrep_gtid_event_buf_len);
+          err = WSREP_TRX_SIZE_EXCEEDED;
+          goto cleanup;
+        }
+        heap_buf = tmp;
+        buf = heap_buf;
+        allocated = thd->wsrep_gtid_event_buf_len;
 
-      assert(Gtid_log_event::MAX_EVENT_LENGTH < STACK_SIZE);
-
-      memcpy(buf, gtid_buf, gtid_len);
-      used = gtid_len;
-      total_length = gtid_len;
+        memcpy(buf, thd->wsrep_gtid_event_buf, thd->wsrep_gtid_event_buf_len);
+      }
+      total_length += thd->wsrep_gtid_event_buf_len;
+      used = total_length;
     }
 
+    if (unlikely(0 == length)) length = my_b_fill(cache);
+
     if (likely(length > 0)) do
     {
         total_length += length;
@@ -188,7 +212,7 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
         */
         if (unlikely(total_length > wsrep_max_ws_size))
         {
-            WSREP_WARN("transaction size limit (%lu) exceeded: %zu",
+            WSREP_WARN("Transaction/Write-set size limit (%lu) exceeded: %zu",
                        wsrep_max_ws_size, total_length);
 	    err = WSREP_TRX_SIZE_EXCEEDED;
             goto cleanup;
@@ -197,12 +221,13 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
         if (total_length > allocated)
         {
             size_t const new_size(heap_size(total_length));
-            uchar* tmp = (uchar *)my_realloc(PSI_NOT_INSTRUMENTED,
-                                             heap_buf, new_size, MYF(0));
+            uchar* tmp = (uchar *)my_realloc(key_memory_wsrep, heap_buf, new_size, MYF(0));
             if (!tmp)
             {
-                WSREP_ERROR("could not (re)allocate buffer: %zu + %u",
-                            allocated, length);
+                WSREP_ERROR("Failed to allocate/reallocate memory to hold"
+                            " write-set for replication."
+                            " Existing Size: %zu, Requested Size: %lu",
+                            allocated, (long unsigned) new_size);
                 err = WSREP_TRX_SIZE_EXCEEDED;
                 goto cleanup;
             }
@@ -213,6 +238,7 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
 
             if (used <= STACK_SIZE && used > 0) // there's data in stack_buf
             {
+                DBUG_ASSERT(buf == stack_buf);
                 memcpy(heap_buf, stack_buf, used);
             }
         }
@@ -230,12 +256,15 @@ static int wsrep_write_cache_once(wsrep_t*  const wsrep,
 cleanup:
     if (reinit_io_cache(cache, WRITE_CACHE, saved_pos, 0, 0))
     {
-        WSREP_ERROR("failed to reinitialize io-cache");
+        WSREP_ERROR("Failed to reinitialize io-cache");
     }
 
     if (unlikely(WSREP_OK != err)) wsrep_dump_rbr_buf(thd, buf, used);
 
     my_free(heap_buf);
+    if (thd->wsrep_gtid_event_buf_len < STACK_SIZE) my_free(thd->wsrep_gtid_event_buf);
+    thd->wsrep_gtid_event_buf     = NULL;
+    thd->wsrep_gtid_event_buf_len = 0;
     return err;
 }
 
@@ -256,32 +285,25 @@ static int wsrep_write_cache_inc(wsrep_t*  const wsrep,
 
     if (reinit_io_cache(cache, READ_CACHE, 0, 0, 0))
     {
-      WSREP_ERROR("failed to initialize io-cache");
+      WSREP_ERROR("Failed to initialize io-cache");
       return WSREP_TRX_ERROR;
     }
 
     int err(WSREP_OK);
 
-    size_t total_length(0);
+    size_t total_length(*len);
 
     uint length(my_b_bytes_in_cache(cache));
-    if (unlikely(0 == length)) length = my_b_fill(cache);
-
-    /* Start 5.7 MySQL server will write GTID directly to binlog file.
-    This means GTID event is not present in the case if GTID is set explictly.
-    Let's construct the GTID event and prefix it to wsrep-binlog-buffer. */
-    if (thd->variables.gtid_next.type != AUTOMATIC_GROUP)
+    if (thd->wsrep_gtid_event_buf)
     {
-      Gtid_log_event gtid_event(thd, true, 0, 0);
-      uchar gtid_buf[Gtid_log_event::MAX_EVENT_LENGTH];
-      uint32 gtid_len = gtid_event.write_to_memory(gtid_buf);
+        if(WSREP_OK != (err=wsrep_append_data(wsrep, &thd->wsrep_ws_handle,
+                                              thd->wsrep_gtid_event_buf,
+                                              thd->wsrep_gtid_event_buf_len)))
+                goto cleanup;
 
-      if (WSREP_OK != (err=wsrep_append_data(wsrep, &thd->wsrep_ws_handle,
-                                            gtid_buf, gtid_len)))
-        goto cleanup;
-
-      total_length = gtid_len;
+        total_length += thd->wsrep_gtid_event_buf_len;
     }
+    if (unlikely(0 == length)) length = my_b_fill(cache);
 
     if (likely(length > 0)) do
     {
@@ -292,7 +314,7 @@ static int wsrep_write_cache_inc(wsrep_t*  const wsrep,
         */
         if (unlikely(total_length > wsrep_max_ws_size))
         {
-            WSREP_WARN("transaction size limit (%lu) exceeded: %zu",
+            WSREP_WARN("Transaction/Write-set size limit (%lu) exceeded: %zu",
                        wsrep_max_ws_size, total_length);
             err = WSREP_TRX_SIZE_EXCEEDED;
             goto cleanup;
@@ -310,9 +332,13 @@ static int wsrep_write_cache_inc(wsrep_t*  const wsrep,
 cleanup:
     if (reinit_io_cache(cache, WRITE_CACHE, saved_pos, 0, 0))
     {
-        WSREP_ERROR("failed to reinitialize io-cache");
+        WSREP_ERROR("Failed to reinitialize io-cache");
     }
 
+    if (thd->wsrep_gtid_event_buf) my_free(thd->wsrep_gtid_event_buf);
+    thd->wsrep_gtid_event_buf_len = 0;
+    thd->wsrep_gtid_event_buf     = NULL;
+
     return err;
 }
 
@@ -350,9 +376,7 @@ void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
   FILE *of= fopen(filename, "wb");
   if (of)
   {
-    size_t ret;
-    ret= fwrite(rbr_buf, buf_len, 1, of);
-    (void)ret;
+    fwrite (rbr_buf, buf_len, 1, of);
     fclose(of);
   }
   else
@@ -379,7 +403,7 @@ void wsrep_dump_rbr_direct(THD* thd, IO_CACHE* cache)
   my_off_t const saved_pos(my_b_tell(cache));
   if (reinit_io_cache(cache, READ_CACHE, 0, 0, 0))
   {
-    WSREP_ERROR("failed to initialize io-cache");
+    WSREP_ERROR("Failed to initialize io-cache");
     return ;
   }
   // open file
@@ -412,7 +436,7 @@ cleanup:
   // init back
   if (reinit_io_cache(cache, WRITE_CACHE, saved_pos, 0, 0))
   {
-    WSREP_ERROR("failed to reinitialize io-cache");
+    WSREP_ERROR("Failed to reinitialize io-cache");
   }
   // close file
   if (of) fclose(of);
