@@ -35,6 +35,8 @@ enum wsrep_exec_mode {
     LOCAL_COMMIT
 };
 
+const char* wsrep_get_exec_mode(wsrep_exec_mode state);
+
 enum wsrep_query_state {
     QUERY_IDLE,
     QUERY_EXEC,
@@ -43,6 +45,8 @@ enum wsrep_query_state {
     QUERY_ROLLINGBACK,
 };
 
+const char* wsrep_get_query_state(wsrep_query_state state);
+
 enum wsrep_conflict_state {
     NO_CONFLICT,
     MUST_ABORT,
@@ -50,16 +54,20 @@ enum wsrep_conflict_state {
     ABORTED,
     MUST_REPLAY,
     REPLAYING,
+    REPLAYED,
     RETRY_AUTOCOMMIT,
     CERT_FAILURE,
 };
 
+const char* wsrep_get_conflict_state(wsrep_conflict_state state);
+
 enum wsrep_consistency_check_mode {
     NO_CONSISTENCY_CHECK,
     CONSISTENCY_CHECK_DECLARED,
     CONSISTENCY_CHECK_RUNNING,
 };
 
+const char* wsrep_get_wsrep_status(wsrep_status status);
 
 // Global wsrep parameters
 extern wsrep_t*    wsrep;
@@ -131,7 +139,17 @@ enum enum_pxc_strict_modes {
 };
 extern ulong       pxc_strict_mode;
 
+enum enum_pxc_maint_modes {
+    PXC_MAINT_MODE_DISABLED = 0,
+    PXC_MAINT_MODE_SHUTDOWN,
+    PXC_MAINT_MODE_MAINTENANCE,
+};
+extern ulong       pxc_maint_mode;
+extern ulong       pxc_maint_transition_period;
+extern my_bool     pxc_encrypt_cluster_traffic;
+
 // MySQL status variables
+extern my_bool     wsrep_new_cluster;
 extern my_bool     wsrep_connected;
 extern my_bool     wsrep_ready;
 extern const char* wsrep_cluster_state_uuid;
@@ -173,7 +191,7 @@ extern "C" void wsrep_thd_set_exec_mode(THD *thd, enum wsrep_exec_mode mode);
 extern "C" void wsrep_thd_set_query_state(
         THD *thd, enum wsrep_query_state state);
 extern "C" void wsrep_thd_set_conflict_state(
-        THD *thd, enum wsrep_conflict_state state);
+        THD *thd, bool lock, enum wsrep_conflict_state state);
 
 extern "C" void wsrep_thd_set_trx_to_replay(THD *thd, uint64 trx_id);
 
@@ -184,14 +202,26 @@ extern "C" time_t wsrep_thd_query_start(THD *thd);
 extern "C" my_thread_id wsrep_thd_thread_id(THD *thd);
 extern "C" int64_t wsrep_thd_trx_seqno(THD *thd);
 extern "C" query_id_t wsrep_thd_query_id(THD *thd);
+extern "C" wsrep_trx_id_t wsrep_thd_next_trx_id(THD *thd);
+extern "C" wsrep_trx_id_t wsrep_thd_trx_id(THD *thd);
 extern "C" const char * wsrep_thd_query(THD *thd);
 extern "C" query_id_t wsrep_thd_wsrep_last_query_id(THD *thd);
 extern "C" void wsrep_thd_set_wsrep_last_query_id(THD *thd, query_id_t id);
 extern "C" void wsrep_thd_awake(THD *thd, my_bool signal);
 extern "C" int wsrep_thd_retry_counter(THD *thd);
 
+extern "C" void wsrep_handle_fatal_signal(int sig);
 
-extern void wsrep_close_client_connections(bool wait_to_end, bool server_shutdown);
+extern "C" void wsrep_thd_set_next_trx_id(THD *thd);
+
+extern "C" void wsrep_set_thd_proc_info(THD* thd, const char* str);
+extern "C" const char* wsrep_get_thd_proc_info(THD* thd);
+
+extern "C" void wsrep_thd_auto_increment_variables(THD*,
+                                                   unsigned long long *offset,
+                                                   unsigned long long *increment);
+
+extern void wsrep_close_client_connections(my_bool wait_to_end, bool server_shtudown);
 extern int  wsrep_wait_committing_connections_close(int wait_time);
 extern void wsrep_close_applier(THD *thd);
 extern void wsrep_wait_appliers_close(THD *thd);
@@ -215,11 +245,19 @@ extern wsrep_seqno_t wsrep_locked_seqno;
    wsrep_provider                     && \
    strcmp(wsrep_provider, WSREP_NONE))
 
+/* use xxxxxx_NNULL macros when thd pointer is guaranteed to be non-null to
+ * avoid compiler warnings (GCC 6 and later) */
+#define WSREP_NNULL(thd) \
+  (WSREP_ON && wsrep && thd->variables.wsrep_on)
+
 #define WSREP(thd) \
-  (WSREP_ON && wsrep && (thd && thd->variables.wsrep_on))
+  (thd && WSREP_NNULL(thd))
 
 #define WSREP_CLIENT(thd) \
-    (WSREP(thd) && thd->wsrep_client_thread)
+  (WSREP(thd) && thd->wsrep_client_thread)
+
+#define WSREP_EMULATE_BINLOG_NNULL(thd) \
+  (WSREP_NNULL(thd) && wsrep_emulate_bin_log)
 
 #define WSREP_EMULATE_BINLOG(thd) \
   (WSREP(thd) && wsrep_emulate_bin_log)
@@ -229,7 +267,7 @@ extern wsrep_seqno_t wsrep_locked_seqno;
 #define WSREP_LOG(fun, ...)                                       \
     {                                                             \
         char msg[1024] = {'\0'};                                  \
-        snprintf(msg, sizeof(msg) - 1, ##__VA_ARGS__);            \
+        snprintf(msg, sizeof(msg) - 1, ## __VA_ARGS__);           \
         fun("WSREP: %s", msg);                                    \
     }
 
@@ -241,9 +279,9 @@ extern wsrep_seqno_t wsrep_locked_seqno;
 
 #define WSREP_LOG_CONFLICT_THD(thd, role)                                      \
     WSREP_LOG(sql_print_information, 	                                       \
-      "%s: \n "       	                                                       \
-      "  THD: %u, mode: %s, state: %s, conflict: %s, seqno: %lld\n "          \
-      "  SQL: %s",							       \
+      "%s: \n "     	                                                       \
+      "  THD: %u, mode: %s, state: %s, conflict: %s, seqno: %lld\n "           \
+      "  SQL: %s\n",							       \
       role, wsrep_thd_thread_id(thd), wsrep_thd_exec_mode_str(thd),            \
       wsrep_thd_query_state_str(thd),                                          \
       wsrep_thd_conflict_state_str(thd), (long long)wsrep_thd_trx_seqno(thd),  \
@@ -253,16 +291,17 @@ extern wsrep_seqno_t wsrep_locked_seqno;
 #define WSREP_LOG_CONFLICT(bf_thd, victim_thd, bf_abort)		       \
   if (wsrep_debug || wsrep_log_conflicts)				       \
   {                                                                            \
-    WSREP_LOG(sql_print_information, "cluster conflict due to %s for threads:",\
+    WSREP_LOG(sql_print_information, "--------- CONFLICT DETECTED --------");  \
+    WSREP_LOG(sql_print_information, "cluster conflict due to %s for threads:\n",\
       (bf_abort) ? "high priority abort" : "certification failure"             \
     );                                                                         \
     if (bf_thd)     WSREP_LOG_CONFLICT_THD(bf_thd, "Winning thread");          \
     if (victim_thd) WSREP_LOG_CONFLICT_THD(victim_thd, "Victim thread");       \
   }
 
-#define WSREP_QUERY(thd)					\
-  (((!opt_general_log_raw) && thd->rewritten_query.length())	\
-   ? (thd->rewritten_query.c_ptr_safe()) : (thd->query().str))
+#define WSREP_QUERY(thd)                                \
+  ((!opt_general_log_raw) && thd->rewritten_query.length()      \
+   ? thd->rewritten_query.c_ptr_safe() : thd->query().str)
 
 extern void wsrep_ready_wait();
 
@@ -275,9 +314,14 @@ enum wsrep_trx_status {
 
 extern enum wsrep_trx_status
 wsrep_run_wsrep_commit(THD *thd, handlerton *hton, bool all);
+
+extern enum wsrep_trx_status wsrep_replicate(THD *thd);
+extern enum wsrep_trx_status wsrep_pre_commit(THD *thd);
+
 class Ha_trx_info;
 struct THD_TRANS;
 void wsrep_register_hton(THD* thd, bool all);
+void wsrep_interim_commit(THD* thd);
 void wsrep_post_commit(THD* thd, bool all);
 void wsrep_brute_force_killer(THD *thd);
 int  wsrep_hire_brute_force_killer(THD *thd, uint64_t trx_id);
@@ -317,6 +361,8 @@ extern PSI_mutex_key key_LOCK_wsrep_sst;
 extern PSI_cond_key  key_COND_wsrep_sst;
 extern PSI_mutex_key key_LOCK_wsrep_sst_init;
 extern PSI_cond_key  key_COND_wsrep_sst_init;
+extern PSI_mutex_key key_LOCK_wsrep_sst_thread;
+extern PSI_cond_key  key_COND_wsrep_sst_thread;
 extern PSI_mutex_key key_LOCK_wsrep_rollback;
 extern PSI_cond_key  key_COND_wsrep_rollback;
 extern PSI_mutex_key key_LOCK_wsrep_replaying;
@@ -349,4 +395,17 @@ bool wsrep_stmt_rollback_is_safe(THD* thd);
 void wsrep_init_sidno(const wsrep_uuid_t&);
 bool wsrep_node_is_donor();
 bool wsrep_node_is_synced();
+bool wsrep_replicate_GTID(THD* thd);
+
+typedef struct wsrep_key_arr
+{
+    wsrep_key_t* keys;
+    size_t       keys_len;
+} wsrep_key_arr_t;
+bool wsrep_prepare_keys_for_isolation(THD*              thd,
+                                      const char*       db,
+                                      const char*       table,
+                                      const TABLE_LIST* table_list,
+                                      wsrep_key_arr_t*  ka);
+void wsrep_keys_free(wsrep_key_arr_t* key_arr);
 #endif /* WSREP_MYSQLD_H */
