@@ -166,7 +166,7 @@ static bool sst_auth_real_set (const char* value)
 
   if (value)
   {
-    v= my_strdup(PSI_NOT_INSTRUMENTED, value, MYF(0));
+    v= my_strdup(key_memory_wsrep, value, MYF(0));
   }
   else                                          // its NULL
   {
@@ -184,7 +184,7 @@ static bool sst_auth_real_set (const char* value)
     if (strlen(sst_auth_real))
     {
       if (wsrep_sst_auth) { my_free((void*) wsrep_sst_auth); }
-      wsrep_sst_auth= my_strdup(PSI_NOT_INSTRUMENTED, WSREP_SST_AUTH_MASK, MYF(0));
+      wsrep_sst_auth= my_strdup(key_memory_wsrep, WSREP_SST_AUTH_MASK, MYF(0));
     }
     return 0;
   }
@@ -227,7 +227,7 @@ static bool            sst_needed   = false;
 
 void wsrep_sst_grab ()
 {
-  WSREP_INFO("wsrep_sst_grab()");
+  WSREP_INFO("Preparing to initiate SST/IST");
   if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
   sst_complete = false;
   mysql_mutex_unlock (&LOCK_wsrep_sst);
@@ -241,7 +241,7 @@ bool wsrep_sst_wait ()
   if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
   while (!sst_complete)
   {
-    WSREP_INFO("Waiting for SST to complete.");
+    WSREP_INFO("Waiting for SST/IST to complete.");
     mysql_cond_wait (&COND_wsrep_sst, &LOCK_wsrep_sst);
   }
 
@@ -315,8 +315,8 @@ void wsrep_sst_received (wsrep_t*            const wsrep,
     }
     else if (local_seqno > seqno)
     {
-        WSREP_WARN("SST postion is in the past: %lld, current: %lld. "
-                   "Can't continue.",
+        WSREP_WARN("SST postion is in the past: %lld."
+                   " Current position: %lld. Can't continue.",
                    (long long)seqno, (long long)local_seqno);
         unireg_abort(1);
     }
@@ -341,7 +341,7 @@ void wsrep_sst_continue ()
 {
   if (sst_needed)
   {
-    WSREP_INFO("Signalling provider to continue.");
+    WSREP_INFO("Signalling provider to continue on SST completion.");
     wsrep_sst_received (wsrep, local_uuid, local_seqno, NULL, 0);
   }
 }
@@ -419,12 +419,12 @@ static int generate_binlog_opt_val(char** ret)
   {
     assert(opt_bin_logname);
     *ret= strcmp(opt_bin_logname, "0") ?
-        my_strdup(PSI_NOT_INSTRUMENTED, opt_bin_logname, MYF(0)) :
-        my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(0));
+        my_strdup(key_memory_wsrep, opt_bin_logname, MYF(0)) :
+        my_strdup(key_memory_wsrep, "", MYF(0));
   }
   else
   {
-    *ret= my_strdup(PSI_NOT_INSTRUMENTED, "", MYF(0));
+    *ret= my_strdup(key_memory_wsrep, "", MYF(0));
   }
   if (!*ret) return -ENOMEM;
   return 0;
@@ -438,18 +438,29 @@ static wsp::process * sst_process= static_cast<wsp::process*>(NULL);
 
 static bool sst_cancelled= false;
 
-void wsrep_sst_cancel ()
+void wsrep_sst_cancel (bool call_wsrep_cb)
 {
   if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
   if (!sst_cancelled)
   {
     sst_cancelled=true;
+    /*
+      When we launched the SST process, then we need
+      to terminate it before exit from the parent (server)
+      process:
+    */
     if (sst_process)
     {
       sst_process->terminate();
       sst_process = NULL;
     }
-    if (sst_awaiting_callback)
+    /*
+      If this is a normal shutdown, then we need to notify
+      the wsrep provider about completion of the SST, to
+      prevent infinite waitng in the wsrep provider after
+      the SST process was canceled:
+    */
+    if (call_wsrep_cb && sst_awaiting_callback)
     {
       WSREP_INFO("Signalling cancellation of the SST request.");
       wsrep_gtid_t const state_id = {
@@ -463,6 +474,23 @@ void wsrep_sst_cancel ()
   mysql_mutex_unlock (&LOCK_wsrep_sst);
 }
 
+/*
+  Handling of fatal signals: SIGABRT, SIGTERM, etc.
+*/
+extern "C" void wsrep_handle_fatal_signal (int sig)
+{
+  wsrep_sst_cancel(false);
+}
+
+/*
+  This callback is invoked in the case of abnormal
+  termination of the wsrep provider:
+*/
+void wsrep_abort_cb (void)
+{
+  wsrep_sst_cancel(false);
+}
+
 static void* sst_joiner_thread (void* a)
 {
   sst_thread_arg* arg= (sst_thread_arg*) a;
@@ -478,7 +506,7 @@ static void* sst_joiner_thread (void* a)
     const size_t out_len = 512;
     char out[out_len];
 
-    WSREP_INFO("Running: '%s'", arg->cmd);
+    WSREP_INFO("Initiating SST/IST transfer on JOINER side (%s)", arg->cmd);
 
     // Launch the SST script and save pointer to its process:
 
@@ -561,7 +589,8 @@ static void* sst_joiner_thread (void* a)
       {
         err= proc.error();
         char errbuf[MYSYS_STRERROR_SIZE];
-        WSREP_ERROR("SST script aborted with error %d (%s)", err, my_strerror(errbuf, sizeof(errbuf), err));
+        WSREP_ERROR("SST script aborted with error %d (%s)",
+                    err, my_strerror(errbuf, sizeof(errbuf), err));
       }
     }
     else
@@ -737,7 +766,7 @@ static ssize_t sst_prepare_mysqldump (const char*  addr_in,
 
 static bool SE_initialized = false;
 
-ssize_t wsrep_sst_prepare (void** msg)
+ssize_t wsrep_sst_prepare (void** msg, THD* thd)
 {
   const ssize_t ip_max= 256;
   char ip_buf[ip_max];
@@ -790,6 +819,7 @@ ssize_t wsrep_sst_prepare (void** msg)
       WSREP_ERROR("Could not prepare state transfer request: "
                   "failed to guess address to accept state transfer at. "
                   "wsrep_sst_receive_address must be set manually.");
+      if (thd) delete thd;
       unireg_abort(1);
     }
   }
@@ -839,7 +869,7 @@ ssize_t wsrep_sst_prepare (void** msg)
     char* const addr_ptr(method_ptr + method_len + 1);
     strcpy (addr_ptr, addr_out);
 
-    WSREP_INFO ("Prepared SST request: %s|%s", method_ptr, addr_ptr);
+    WSREP_INFO ("Prepared SST/IST request: %s|%s", method_ptr, addr_ptr);
 
     if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
     sst_awaiting_callback = true;
@@ -911,7 +941,7 @@ static void sst_reject_queries(my_bool close_conn)
 {
     wsrep_ready_set (FALSE); // this will be resotred when donor becomes synced
     WSREP_INFO("Rejecting client queries for the duration of SST.");
-    if (TRUE == close_conn) wsrep_close_client_connections(false, false);
+    if (TRUE == close_conn) wsrep_close_client_connections(FALSE, false);
 }
 
 static int sst_donate_mysqldump (const char*         addr,
@@ -963,7 +993,7 @@ static int sst_donate_mysqldump (const char*         addr,
                      "%s",
                      host, port, mysqld_port, mysqld_unix_port,
                      wsrep_defaults_file, uuid_str,
-                     (long long)seqno, bypass ? " "WSREP_SST_OPT_BYPASS : "");
+                     (long long)seqno, bypass ? " " WSREP_SST_OPT_BYPASS : "");
 
   if (ret < 0 || ret >= cmd_len)
   {
@@ -1034,7 +1064,7 @@ static int sst_flush_tables(THD* thd)
   }
   else
   {
-    WSREP_INFO("Tables flushed.");
+    WSREP_INFO("Table flushing completed.");
     const char base_name[]= "tables_flushed";
     ssize_t const full_len= strlen(mysql_real_data_home) + strlen(base_name)+2;
     char *real_name= static_cast<char *>(alloca(full_len));
@@ -1087,7 +1117,7 @@ static void* sst_donor_thread (void* a)
   wsrep_pfs_register_thread(key_THREAD_wsrep_sst_donor);
 #endif /* HAVE_PSI_INTERFACE */
 
-  WSREP_INFO("Running: '%s'", arg->cmd);
+  WSREP_INFO("Initiating SST/IST transfer on DONOR side (%s)", arg->cmd);
 
   int  err= 1;
   bool locked= false;
@@ -1164,10 +1194,12 @@ wait_signal:
     {
       if (mysql_mutex_lock (&LOCK_wsrep_sst)) abort();
       // Print error message if SST is not cancelled:
+#if 0
       if (!sst_cancelled)
       {
          WSREP_ERROR("Failed to read from: %s", proc.cmd());
       }
+#endif
       // Clear the pointer to SST process:
       sst_process = NULL;
       mysql_mutex_unlock (&LOCK_wsrep_sst);
@@ -1258,7 +1290,7 @@ static int sst_donate_other (const char*   method,
                  wsrep_defaults_file, wsrep_defaults_group_suffix,
                  binlog_opt, binlog_opt_val,
                  uuid, (long long) seqno,
-                 bypass ? " "WSREP_SST_OPT_BYPASS : "");
+                 bypass ? " " WSREP_SST_OPT_BYPASS : "");
   my_free(binlog_opt_val);
 
   if (ret < 0 || ret >= cmd_len)
@@ -1281,7 +1313,7 @@ static int sst_donate_other (const char*   method,
   }
   mysql_cond_wait (&arg.COND_wsrep_sst_thread, &arg.LOCK_wsrep_sst_thread);
 
-  WSREP_INFO("sst_donor_thread signaled with %d", arg.err);
+  WSREP_INFO("DONOR thread signaled with %d", arg.err);
   pthread_detach(tmp);
   return arg.err;
 }
@@ -1311,6 +1343,17 @@ wsrep_cb_status_t wsrep_sst_donate_cb (void* app_ctx, void* recv_ctx,
     return WSREP_CB_FAILURE;
   }
 
+  /* Wait for wsrep-SE to initialize that also signals
+  completion of init_server_component which is important before we initiate
+  any meangiful action especially DONOR action from this node. */
+  while (!wsrep_is_SE_initialized())
+  {
+    sleep(1);
+    THD* applier_thd= static_cast<THD*>(recv_ctx);
+    if (applier_thd->killed == THD::KILL_CONNECTION)
+      return WSREP_CB_FAILURE;
+  }
+
   int ret;
   if ((ret= sst_append_auth_env(env, sst_auth_real)))
   {
@@ -1337,13 +1380,33 @@ void wsrep_SE_init_grab()
   if (mysql_mutex_lock (&LOCK_wsrep_sst_init)) abort();
 }
 
-void wsrep_SE_init_wait()
+void wsrep_SE_init_wait(THD* thd)
 {
-  while (SE_initialized == false)
+  while (SE_initialized == false && thd->killed == THD::NOT_KILLED)
   {
+    mysql_mutex_lock(&thd->LOCK_thd_data);
+    thd->current_cond= &COND_wsrep_sst_init;
+    thd->current_mutex= &LOCK_wsrep_sst_init;
+    mysql_mutex_unlock(&thd->LOCK_thd_data);
+    
     mysql_cond_wait (&COND_wsrep_sst_init, &LOCK_wsrep_sst_init);
+
+    if (thd->killed != THD::NOT_KILLED)
+    {
+      WSREP_DEBUG("SE init waiting canceled");
+      break;
+    }
+    mysql_mutex_lock(&thd->LOCK_thd_data);
+    thd->current_cond= NULL;
+    thd->current_mutex= NULL;
+    mysql_mutex_unlock(&thd->LOCK_thd_data);
   }
   mysql_mutex_unlock (&LOCK_wsrep_sst_init);
+
+  mysql_mutex_lock(&thd->LOCK_thd_data);
+  thd->current_cond= NULL;
+  thd->current_mutex= NULL;
+  mysql_mutex_unlock(&thd->LOCK_thd_data);
 }
 
 void wsrep_SE_init_done()
