@@ -35,9 +35,9 @@ int wsrep_show_bf_aborts (THD *thd, SHOW_VAR *var, char *buff)
 /* must have (&thd->LOCK_wsrep_thd) */
 void wsrep_client_rollback(THD *thd)
 {
-  WSREP_DEBUG("client rollback due to BF abort for (%u), query: %s",
-              thd->thread_id(), WSREP_QUERY(thd));
-
+  WSREP_DEBUG("Initiating rollback. Transaction (query: %s) triggered by"
+              " thread %u was BF aborted",
+              WSREP_QUERY(thd), thd->thread_id());
   my_atomic_add64(&wsrep_bf_aborts_counter, 1);
 
   thd->wsrep_conflict_state= ABORTING;
@@ -50,14 +50,16 @@ void wsrep_client_rollback(THD *thd)
 
   if (thd->locked_tables_mode && thd->lock)
   {
-    WSREP_DEBUG("unlocking tables for BF abort (%u)", thd->thread_id());
+    WSREP_DEBUG("Unlock tables, transaction triggered by thread id (%u)"
+                " was BF aborted and rolled back", thd->thread_id());
     thd->locked_tables_list.unlock_locked_tables(thd);
     thd->variables.option_bits&= ~(OPTION_TABLE_LOCK);
   }
 
   if (thd->global_read_lock.is_acquired())
   {
-    WSREP_DEBUG("unlocking GRL for BF abort (%u)", thd->thread_id());
+    WSREP_DEBUG("Release Global Read Lock, transaction triggered by thread id (%u)"
+                " was BF aborted and rolled back", thd->thread_id());
     thd->global_read_lock.unlock_global_read_lock(thd);
   }
 
@@ -72,7 +74,8 @@ void wsrep_client_rollback(THD *thd)
 
   if (thd->get_binlog_table_maps())
   {
-    WSREP_DEBUG("clearing binlog table map for BF abort (%u)", thd->thread_id());
+    WSREP_DEBUG("Clear binlog table map, transaction triggered by thread id (%u)"
+                " was BF aborted and rolled back", thd->thread_id());
     thd->clear_binlog_table_maps();
   }
   mysql_mutex_lock(&thd->LOCK_wsrep_thd);
@@ -87,66 +90,17 @@ static Relay_log_info* wsrep_relay_log_init(const char* log_fname)
 {
   uint rli_option = INFO_REPOSITORY_DUMMY;
   Relay_log_info *rli= NULL;
-  rli = Rpl_info_factory::create_rli(rli_option, false, (const char*)"", true);
+  rli = Rpl_info_factory::create_rli(rli_option, false, "wsrep", true);
+  if (!rli)
+  {
+    WSREP_ERROR("Failed to create Relay-Log for wsrep thread, aborting");
+    unireg_abort(MYSQLD_ABORT_EXIT);
+  }
   rli->set_rli_description_event(
       new Format_description_log_event(BINLOG_VERSION));
 
+  rli->current_mts_submode= new Mts_submode_wsrep();
   return (rli);
-
-#ifdef REMOVED
-  Rpl_info_handler* handler_src= NULL;
-  Rpl_info_handler* handler_dest= NULL;
-  ulong *key_info_idx= NULL;
-  const char *msg= "Failed to allocate memory for the relay log info "
-                   "structure";
-
-  DBUG_ENTER("Rpl_info_factory::create_rli");
-
-  if (!(rli= new Relay_log_info(false
-#ifdef HAVE_PSI_INTERFACE
-                                ,&key_relay_log_info_run_lock,
-                                &key_relay_log_info_data_lock,
-                                &key_relay_log_info_sleep_lock,
-                                &key_relay_log_info_data_cond,
-                                &key_relay_log_info_start_cond,
-                                &key_relay_log_info_stop_cond,
-                                &key_relay_log_info_sleep_cond
-#endif /* HAVE_PSI_INTERFACE */
-                               )))
-    goto err;
-
-  if (!(key_info_idx= new ulong[NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR]))
-     goto err;
-  key_info_idx[0]= server_id;
-  rli->set_idx_info(key_info_idx, NUMBER_OF_FIELDS_TO_IDENTIFY_COORDINATOR);
-
-  if(Rpl_info_factory::init_rli_repositories(rli, rli_option, &handler_src,
-                                             &handler_dest, &msg))
-    goto err;
-
-  if (Rpl_info_factory::decide_repository(rli, rli_option, &handler_src,
-                                          &handler_dest, &msg))
-    goto err;
-
-  DBUG_RETURN(rli);
-err:
-  delete handler_src;
-  delete handler_dest;
-  delete []key_info_idx;
-  if (rli)
-  {
-    /*
-      The handler was previously deleted so we need to remove
-      any reference to it.
-    */
-    rli->set_idx_info(NULL, 0);
-    rli->set_rpl_info_handler(NULL);
-    rli->set_rpl_info_type(INVALID_INFO_REPOSITORY);
-    delete rli;
-  }
-  WSREP_ERROR("Error creating relay log info: %s.", msg);
-  DBUG_RETURN(NULL);
-#endif /* REMOVED */
 }
 
 static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
@@ -154,7 +108,7 @@ static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
   shadow->options       = thd->variables.option_bits;
   shadow->server_status = thd->server_status;
   shadow->wsrep_exec_mode = thd->wsrep_exec_mode;
-  shadow->vio           = thd->get_protocol_classic()->get_vio();
+  shadow->vio           = thd->active_vio;
 
   // Disable general logging on applier threads
   thd->variables.option_bits |= OPTION_LOG_OFF;
@@ -164,28 +118,38 @@ static void wsrep_prepare_bf_thd(THD *thd, struct wsrep_thd_shadow* shadow)
   else
     thd->variables.option_bits&= ~(OPTION_BIN_LOG);
 
-  if (!thd->wsrep_rli) thd->wsrep_rli= wsrep_relay_log_init("wsrep_relay");
-  thd->wsrep_rli->info_thd = thd;
-  thd->init_for_queries(thd->wsrep_rli);
+#ifdef GALERA
+  /*
+    in 5.7, we declare applying to happen as with slave threads
+    this set here is for replaying, when local threads will operate
+    as slave for the duration of replaying
+  */
+  thd->slave_thread = TRUE;
+#endif /* GALERA */
 
-  thd->wsrep_rli->current_mts_submode= new Mts_submode_database();
+  if (!thd->wsrep_rli)
+  {
+    thd->wsrep_rli = wsrep_relay_log_init("wsrep_relay");
+    assert(!thd->rli_slave);
+    thd->rli_slave = thd->wsrep_rli;
+    thd->wsrep_rli->info_thd= thd;
+    thd->init_for_queries(thd->wsrep_rli);
+  }
+  thd->wsrep_rli->info_thd = thd;
 
   thd->wsrep_exec_mode= REPL_RECV;
-  thd->get_protocol_classic()->set_vio(NULL);
+  thd->set_active_vio(0);
   thd->clear_error();
 
   shadow->tx_isolation        = thd->variables.tx_isolation;
   thd->variables.tx_isolation = ISO_READ_COMMITTED;
   thd->tx_isolation           = ISO_READ_COMMITTED;
 
-  // intended shallow copy.
-  shadow->db            = thd->db();
+  shadow->db = thd->db();
+  thd->reset_db(NULL_CSTR);
 
-  LEX_CSTRING  dummy;
-  dummy.str= NULL;
-  dummy.length= 0;
-  thd->reset_db(dummy);
   shadow->user_time = thd->user_time;
+  shadow->row_count_func= thd->get_row_count_func();
 }
 
 static void wsrep_return_from_bf_mode(THD *thd, struct wsrep_thd_shadow* shadow)
@@ -193,10 +157,22 @@ static void wsrep_return_from_bf_mode(THD *thd, struct wsrep_thd_shadow* shadow)
   thd->variables.option_bits  = shadow->options;
   thd->server_status          = shadow->server_status;
   thd->wsrep_exec_mode        = shadow->wsrep_exec_mode;
-  thd->get_protocol_classic()->set_vio(shadow->vio);
+  thd->set_active_vio(shadow->vio);
   thd->variables.tx_isolation = shadow->tx_isolation;
   thd->reset_db(shadow->db);
   thd->user_time              = shadow->user_time;
+
+  assert(thd->rli_slave == thd->wsrep_rli);
+  thd->rli_slave = NULL;
+
+  delete thd->wsrep_rli->current_mts_submode;
+  thd->wsrep_rli->current_mts_submode = 0;
+  delete thd->wsrep_rli;
+  thd->wsrep_rli = 0;
+#ifdef GALERA
+  thd->slave_thread = FALSE;
+#endif /* GALERA */
+  thd->set_row_count_func(shadow->row_count_func);
 }
 
 void wsrep_replay_transaction(THD *thd)
@@ -208,7 +184,7 @@ void wsrep_replay_transaction(THD *thd)
     if (thd->wsrep_exec_mode!= REPL_RECV) {
       if (thd->get_stmt_da()->is_sent())
       {
-        WSREP_ERROR("replay issue, thd has reported status already");
+        WSREP_ERROR("Transaction replay issue, thd has reported status already");
       }
 
       /* PS reprepare observer should have been removed already
@@ -216,9 +192,29 @@ void wsrep_replay_transaction(THD *thd)
        */
       if (thd->get_reprepare_observer() && wsrep_log_conflicts)
       {
-        WSREP_WARN("dangling observer in replay transaction: (thr %u %lld %s)",
+        WSREP_WARN("Found dangling observer in replaying of transaction"
+                   " (thr %u %lld %s)",
                    thd->thread_id(), thd->query_id, thd->query().str);
       }
+
+      struct da_shadow
+      {
+          enum Diagnostics_area::enum_diagnostics_status status;
+          ulonglong affected_rows;
+          ulonglong last_insert_id;
+          char message[MYSQL_ERRMSG_SIZE];
+      };
+      struct da_shadow da_status;
+      da_status.status= thd->get_stmt_da()->status();
+      if (da_status.status == Diagnostics_area::DA_OK)
+      {
+        da_status.affected_rows= thd->get_stmt_da()->affected_rows();
+        da_status.last_insert_id= thd->get_stmt_da()->last_insert_id();
+        strmake(da_status.message,
+                thd->get_stmt_da()->message_text(),
+                sizeof(da_status.message)-1);
+      }
+
       thd->get_stmt_da()->reset_diagnostics_area();
 
       thd->wsrep_conflict_state= REPLAYING;
@@ -229,7 +225,8 @@ void wsrep_replay_transaction(THD *thd)
       close_thread_tables(thd);
       if (thd->locked_tables_mode && thd->lock)
       {
-        WSREP_DEBUG("releasing table lock for replaying (%u)",
+        WSREP_DEBUG("Release table locks owned by thread (%u)."
+                    "Transaction will be replayed",
                     thd->thread_id());
         thd->locked_tables_list.unlock_locked_tables(thd);
         thd->variables.option_bits&= ~(OPTION_TABLE_LOCK);
@@ -243,10 +240,18 @@ void wsrep_replay_transaction(THD *thd)
       MYSQL_END_STATEMENT(thd->m_statement_psi, thd->get_stmt_da());
       thd->m_statement_psi= NULL;
       thd->m_digest= NULL;
-      THD_STAGE_INFO_GUARD(thd, &stage_wsrep_replaying_trx);
-      WSREP_DEBUG("replay trx: %s %lld",
+
+      THD_STAGE_INFO(thd, stage_wsrep_replaying_trx);
+      snprintf(thd->wsrep_info, sizeof(thd->wsrep_info),
+               "wsrep: replaying write set (%lld)",
+               (long long)wsrep_thd_trx_seqno(thd));
+      WSREP_DEBUG("%s", thd->wsrep_info);
+      thd_proc_info(thd, thd->wsrep_info);
+
+      WSREP_DEBUG("Replaying transaction for SQL query: %s with write-set: %lld",
                   WSREP_QUERY(thd),
                   (long long)wsrep_thd_trx_seqno(thd));
+
       struct wsrep_thd_shadow shadow;
       wsrep_prepare_bf_thd(thd, &shadow);
 
@@ -259,8 +264,8 @@ void wsrep_replay_transaction(THD *thd)
                                     (void *)thd);
 
       wsrep_return_from_bf_mode(thd, &shadow);
-      if (thd->wsrep_conflict_state!= REPLAYING)
-        WSREP_WARN("lost replaying mode: %d", thd->wsrep_conflict_state );
+      if (thd->wsrep_conflict_state != REPLAYING)
+        WSREP_WARN("Lost replaying mode: %d", thd->wsrep_conflict_state );
 
       mysql_mutex_lock(&thd->LOCK_wsrep_thd);
 
@@ -269,41 +274,53 @@ void wsrep_replay_transaction(THD *thd)
       case WSREP_OK:
         thd->wsrep_conflict_state= NO_CONFLICT;
         wsrep->post_commit(wsrep, &thd->wsrep_ws_handle);
-        WSREP_DEBUG("trx_replay successful for: %u %llu",
+        WSREP_DEBUG("Transaction replay completed successfully (%u %llu)",
                     thd->thread_id(), (long long)thd->real_id);
         if (thd->get_stmt_da()->is_sent())
         {
-          WSREP_WARN("replay ok, thd has reported status");
+          WSREP_WARN("Replay completed with success");
         }
         else if (thd->get_stmt_da()->is_set())
         {
           if (thd->get_stmt_da()->status() != Diagnostics_area::DA_OK)
           {
-            WSREP_WARN("replay ok, thd has error status %d",
+            WSREP_WARN("Replay completed with error %d",
                        thd->get_stmt_da()->status());
           }
         }
         else
         {
-          my_ok(thd);
+          if (da_status.status == Diagnostics_area::DA_OK)
+          {
+            my_ok(thd,
+                  da_status.affected_rows,
+                  da_status.last_insert_id,
+                  da_status.message);
+          }
+          else
+          {
+            my_ok(thd);
+          }
         }
         break;
       case WSREP_TRX_FAIL:
         if (thd->get_stmt_da()->is_sent())
         {
-          WSREP_ERROR("replay failed, thd has reported status");
+          WSREP_ERROR("Transaction replay failed, thd has reported status");
         }
         else
         {
-          WSREP_DEBUG("replay failed, rolling back");
+          WSREP_DEBUG("Transaction replay with write-set: %lld failed."
+                      " Rolling back", (long long)wsrep_thd_trx_seqno(thd));
           //my_error(ER_LOCK_DEADLOCK, MYF(0), "wsrep aborted transaction");
         }
         thd->wsrep_conflict_state= ABORTED;
         wsrep->post_rollback(wsrep, &thd->wsrep_ws_handle);
         break;
       default:
-        WSREP_ERROR("trx_replay failed for: %d, schema: %s, query: %s",
-                    rcode, (thd->db().length ? thd->db().str : "(null)"),
+        WSREP_ERROR("Failed to replay transaction for: %d, schema: %s, query: %s",
+                    rcode,
+                    (thd->db().str ? thd->db().str : "(null)"),
                     WSREP_QUERY(thd));
         /* we're now in inconsistent state, must abort */
 	mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
@@ -315,7 +332,7 @@ void wsrep_replay_transaction(THD *thd)
 
       mysql_mutex_lock(&LOCK_wsrep_replaying);
       wsrep_replaying--;
-      WSREP_DEBUG("replaying decreased: %d, thd: %u",
+      WSREP_DEBUG("Decreasing replay counter: %d, thd: %u",
                   wsrep_replaying, thd->thread_id());
       mysql_cond_broadcast(&COND_wsrep_replaying);
       mysql_mutex_unlock(&LOCK_wsrep_replaying);
@@ -331,7 +348,8 @@ static void wsrep_replication_process(THD *thd)
 {
   int rcode;
   DBUG_ENTER("wsrep_replication_process");
-  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_applier_idle);
+
+  THD_STAGE_INFO(thd, stage_wsrep_applier_idle);
 
 #ifdef HAVE_PSI_INTERFACE
   wsrep_pfs_register_thread(key_THREAD_wsrep_applier);
@@ -346,8 +364,8 @@ static void wsrep_replication_process(THD *thd)
   as the first BEGIN statement is now intepreted wrongly as statement
   inside an active transaction. */
   /* From trans_begin() */
-  //thd->variables.option_bits|= OPTION_BEGIN;
-  //thd->server_status|= SERVER_STATUS_IN_TRANS;
+  thd->variables.option_bits|= OPTION_BEGIN;
+  thd->server_status|= SERVER_STATUS_IN_TRANS;
 
   rcode = wsrep->recv(wsrep, (void *)thd);
   DBUG_PRINT("wsrep",("wsrep_repl returned: %d", rcode));
@@ -366,7 +384,7 @@ static void wsrep_replication_process(THD *thd)
     /* Note: we cannot just blindly restart replication here,
      * SST might require server restart if storage engines must be
      * initialized after SST */
-    WSREP_ERROR("node consistency compromised, aborting");
+    WSREP_ERROR("Node consistency compromised. Aborting");
     wsrep_kill_mysql(thd);
     break;
   case WSREP_WARNING:
@@ -389,17 +407,15 @@ static void wsrep_replication_process(THD *thd)
     break;
   }
 
-  thd_lock_thread_count(thd);
   wsrep_close_applier(thd);
-  thd_unlock_thread_count(thd);
 
   TABLE *tmp;
   while ((tmp = thd->temporary_tables))
   {
     WSREP_WARN("Applier %u, has temporary tables at exit: %s.%s",
-                  thd->thread_id(), 
-                  (tmp->s) ? tmp->s->db.str : "void",
-                  (tmp->s) ? tmp->s->table_name.str : "void");
+               thd->thread_id(), 
+               (tmp->s) ? tmp->s->db.str : "void",
+               (tmp->s) ? tmp->s->table_name.str : "void");
   }
   wsrep_return_from_bf_mode(thd, &shadow);
 
@@ -420,7 +436,7 @@ void wsrep_create_appliers(long threads)
     if (wsrep_cluster_address && strlen(wsrep_cluster_address) &&
         wsrep_provider && strcasecmp(wsrep_provider, "none"))
     {
-      WSREP_WARN("Trying to launch slave threads before creating "
+      WSREP_ERROR("Trying to launch slave threads before creating "
                   "connection at '%s'", wsrep_cluster_address);
     }
     return;
@@ -432,7 +448,7 @@ void wsrep_create_appliers(long threads)
     if (pthread_create(
       &hThread, &connection_attrib,
       start_wsrep_THD, (void*)wsrep_replication_process))
-      WSREP_WARN("Can't create thread to manage wsrep replication");
+      WSREP_WARN("Failed to create wsrep applier/slave thread");
   }
 }
 
@@ -447,31 +463,39 @@ static void wsrep_rollback_process(THD *thd)
   mysql_mutex_lock(&LOCK_wsrep_rollback);
   wsrep_aborting_thd= NULL;
 
-  THD_STAGE_INFO_GUARD(thd, &stage_wsrep_in_rollback_thread);
+  THD_STAGE_INFO(thd, stage_wsrep_in_rollback_thread);
 
   while (thd->killed == THD::NOT_KILLED) {
 
+    THD_STAGE_INFO(thd, stage_wsrep_aborter_idle);
+
     mysql_mutex_lock(&thd->LOCK_current_cond);
-    THD_STAGE_INFO_GUARD_UPDATE(&stage_wsrep_aborter_idle);
     thd->current_mutex= &LOCK_wsrep_rollback;
     thd->current_cond= &COND_wsrep_rollback;
     mysql_mutex_unlock(&thd->LOCK_current_cond);
 
     mysql_cond_wait(&COND_wsrep_rollback,&LOCK_wsrep_rollback);
 
-    WSREP_DEBUG("WSREP rollback thread wakes for signal");
+    if (thd->killed != THD::NOT_KILLED)
+    {
+      WSREP_DEBUG("Rollback/Aborter thread cancelled");
+      break;
+    }
+
+    WSREP_DEBUG("Rollback/Aborter thread wakes for signal."
+                " Has some transaction rollback work to do.");
+
+    THD_STAGE_INFO(thd, stage_wsrep_aborter_active);
 
     mysql_mutex_lock(&thd->LOCK_current_cond);
-    THD_STAGE_INFO_GUARD_UPDATE(&stage_wsrep_aborter_active);
     thd->current_mutex= 0;
     thd->current_cond= 0;
     mysql_mutex_unlock(&thd->LOCK_current_cond);
 
     /* check for false alarms */
     if (!wsrep_aborting_thd)
-    {
-      WSREP_DEBUG("WSREP rollback thread has empty abort queue");
-    }
+      WSREP_DEBUG("Rollback/Aborter thread found empty abort queue");
+
     /* process all entries in the queue */
     while (wsrep_aborting_thd) {
       THD *aborting;
@@ -486,11 +510,12 @@ static void wsrep_rollback_process(THD *thd)
       mysql_mutex_unlock(&LOCK_wsrep_rollback);
 
       mysql_mutex_lock(&aborting->LOCK_wsrep_thd);
-      if (aborting->wsrep_conflict_state== ABORTED)
+      if (aborting->wsrep_conflict_state == ABORTED)
       {
-        WSREP_DEBUG("WSREP, thd already aborted: %llu state: %d",
-                    (long long)aborting->real_id,
-                    aborting->wsrep_conflict_state);
+        WSREP_DEBUG("Rollback/Aborter encounter already aborted thread"
+                    " thread-id: %u (%llu) conflict-state: %s",
+                    aborting->thread_id(), (long long)aborting->real_id,
+                    wsrep_get_conflict_state(aborting->wsrep_conflict_state));
 
         mysql_mutex_unlock(&aborting->LOCK_wsrep_thd);
         mysql_mutex_lock(&LOCK_wsrep_rollback);
@@ -504,8 +529,10 @@ static void wsrep_rollback_process(THD *thd)
 
       mysql_mutex_lock(&aborting->LOCK_wsrep_thd);
       wsrep_client_rollback(aborting);
-      WSREP_DEBUG("WSREP rollbacker aborted thd: (%u %llu)",
-                  aborting->thread_id(), (long long)aborting->real_id);
+      WSREP_DEBUG("Rollback/Aborter thread successfully aborted thread"
+                  " thread-id: %u (%llu) conflict-state: %s",
+                  aborting->thread_id(), (long long)aborting->real_id,
+                  wsrep_get_conflict_state(aborting->wsrep_conflict_state));
       mysql_mutex_unlock(&aborting->LOCK_wsrep_thd);
 
       mysql_mutex_lock(&LOCK_wsrep_rollback);
@@ -513,12 +540,19 @@ static void wsrep_rollback_process(THD *thd)
   }
 
   mysql_mutex_unlock(&LOCK_wsrep_rollback);
-  sql_print_information("WSREP: rollbacker thread exiting");
 
 #ifdef HAVE_PSI_INTERFACE
   wsrep_pfs_delete_thread();
 #endif /* HAVE_PSI_INTERFACE */
 
+  mysql_mutex_lock(&thd->LOCK_thd_data);
+  thd_proc_info(thd, "wsrep aborter shutting down");
+  thd->current_mutex= 0;
+  thd->current_cond=  0;
+  mysql_mutex_unlock(&thd->LOCK_thd_data);
+
+  sql_print_information("WSREP: rollbacker thread exiting");
+  thd->store_globals();
   DBUG_PRINT("wsrep",("wsrep rollbacker thread exiting"));
   DBUG_VOID_RETURN;
 }
@@ -531,7 +565,7 @@ void wsrep_create_rollbacker()
     /* create rollbacker */
     if (pthread_create( &hThread, &connection_attrib,
                         start_wsrep_THD, (void*)wsrep_rollback_process))
-      WSREP_WARN("Can't create thread to manage wsrep rollback");
+      WSREP_WARN("Failed to create wsrep rollback/aborter thread");
   }
 }
 
@@ -565,7 +599,7 @@ my_bool wsrep_thd_is_BF(void *thd_ptr, my_bool sync)
   {
     THD* thd = (THD*)thd_ptr;
     if (sync) mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-    
+
     status = ((thd->wsrep_exec_mode == REPL_RECV)    ||
 	      (thd->wsrep_exec_mode == TOTAL_ORDER));
     if (sync) mysql_mutex_unlock(&thd->LOCK_wsrep_thd);
@@ -581,7 +615,7 @@ my_bool wsrep_thd_is_BF_or_commit(void *thd_ptr, my_bool sync)
   {
     THD* thd = (THD*)thd_ptr;
     if (sync) mysql_mutex_lock(&thd->LOCK_wsrep_thd);
-    
+
     status = ((thd->wsrep_exec_mode == REPL_RECV)    ||
 	      (thd->wsrep_exec_mode == TOTAL_ORDER)  ||
 	      (thd->wsrep_exec_mode == LOCAL_COMMIT));
@@ -613,11 +647,12 @@ int wsrep_abort_thd(void *bf_thd_ptr, void *victim_thd_ptr, my_bool signal)
 
   if ( (WSREP(bf_thd) ||
          ( (WSREP_ON || bf_thd->variables.wsrep_OSU_method == WSREP_OSU_RSU) &&
-           bf_thd->wsrep_exec_mode == TOTAL_ORDER) )               &&
+           bf_thd->wsrep_exec_mode == TOTAL_ORDER) )                         &&
        victim_thd)
   {
-    WSREP_DEBUG("wsrep_abort_thd, by: %llu, victim: %llu", (bf_thd) ?
-                (long long)bf_thd->real_id : 0, (long long)victim_thd->real_id);
+    WSREP_DEBUG("BF thread %u (%llu) aborting Vicitim thread %u (%llu)",
+                bf_thd->thread_id(), (long long) bf_thd->real_id,
+                victim_thd->thread_id(), (long long) victim_thd->real_id);
     ha_wsrep_abort_transaction(bf_thd, victim_thd, signal);
   }
   else
@@ -641,3 +676,37 @@ bool wsrep_thd_has_explicit_locks(THD *thd)
   assert(thd);
   return (thd->mdl_context.wsrep_has_explicit_locks());
 }
+
+/*
+  Get auto increment variables for THD. Use global settings for
+  applier threads.
+ */
+extern "C"
+void wsrep_thd_auto_increment_variables(THD* thd,
+                                        unsigned long long* offset,
+                                        unsigned long long* increment)
+{
+  if (thd->wsrep_exec_mode == REPL_RECV &&
+      thd->wsrep_conflict_state != REPLAYING)
+  {
+    *offset= global_system_variables.auto_increment_offset;
+    *increment= global_system_variables.auto_increment_increment;
+  }
+  else
+  {
+    *offset= thd->variables.auto_increment_offset;
+    *increment= thd->variables.auto_increment_increment;
+  }
+}
+
+bool wsrep_safe_to_persist_xid(THD* thd)
+{
+  /* Rollback of transaction too also invoke persist of XID.
+  Avoid persisting XID in this use-case. */
+  bool safe_to_persist_xid= false;
+  if (thd->wsrep_conflict_state == NO_CONFLICT      ||
+      thd->wsrep_conflict_state == REPLAYING        ||
+      thd->wsrep_conflict_state == RETRY_AUTOCOMMIT)
+    safe_to_persist_xid= true;
+  return(safe_to_persist_xid);
+}
