@@ -113,7 +113,7 @@ env::ctor_common(char** e)
             if (!env_[i])
             {
                 errno_ = errno;
-                WSREP_ERROR("Failed to allocate env. var: %s", e[i]);
+                WSREP_ERROR("Failed to allocate environment variable: %s", e[i]);
                 return true;
             }
         }
@@ -124,7 +124,7 @@ env::ctor_common(char** e)
     else
     {
         errno_ = errno;
-        WSREP_ERROR("Failed to allocate env. var vector of length: %zu", len_);
+        WSREP_ERROR("Failed to allocate environment variable vector of length: %zu", len_);
         return true;
     }
 }
@@ -779,7 +779,15 @@ void process::terminate ()
 {
   if (pid_)
   {
+    /*
+      If we have an appropriated system call, then we try
+      to terminate entire process group:
+    */
+#if _XOPEN_SOURCE >= 500 || _DEFAULT_SOURCE || _BSD_SOURCE
+    if (killpg(pid_, SIGTERM))
+#else
     if (kill(pid_, SIGTERM))
+#endif
     {
       WSREP_WARN("Unable to terminate process: %s: %d (%s)",
                  str_, errno, strerror(errno));
