@@ -227,17 +227,7 @@ int ha_recover(HASH *commit_list)
 }
 
 
-/**
-  Rollback the active XA transaction.
-
-  @note Resets rm_error before calling ha_rollback(), so
-        the thd->transaction.xid structure gets reset
-        by ha_rollback() / THD::transaction::cleanup().
-
-  @return true if the rollback failed, false otherwise.
-*/
-
-static bool xa_trans_force_rollback(THD *thd)
+bool xa_trans_force_rollback(THD *thd)
 {
   /*
     We must reset rm_error before calling ha_rollback(),
@@ -257,14 +247,7 @@ static bool xa_trans_force_rollback(THD *thd)
 }
 
 
-/**
-  Reset some transaction state information and delete corresponding
-  Transaction_ctx object from cache.
-
-  @param thd    Current thread
-*/
-
-static void cleanup_trans_state(THD *thd)
+void cleanup_trans_state(THD *thd)
 {
   thd->variables.option_bits&= ~OPTION_BEGIN;
   thd->server_status&=
@@ -427,6 +410,8 @@ bool Sql_cmd_xa_commit::trans_xa_commit(THD *thd)
       else
         res= MY_TEST(ha_commit_low(thd, /* all */ true));
 
+      DBUG_EXECUTE_IF("simulate_xa_commit_log_failure", { res= true; });
+
       if (res)
         my_error(ER_XAER_RMERR, MYF(0)); // todo/fixme: consider to rollback it
 #ifdef HAVE_PSI_TRANSACTION_INTERFACE
@@ -437,8 +422,9 @@ bool Sql_cmd_xa_commit::trans_xa_commit(THD *thd)
           we need to explicitly mark the transaction as committed.
         */
         MYSQL_COMMIT_TRANSACTION(thd->m_transaction_psi);
-        thd->m_transaction_psi= NULL;
       }
+
+      thd->m_transaction_psi= NULL;
 #endif
     }
   }
@@ -468,6 +454,9 @@ bool Sql_cmd_xa_commit::execute(THD *thd)
   if (!st)
   {
     thd->mdl_context.release_transactional_locks();
+#ifdef WITH_WSREP
+    WSREP_DEBUG("XA commit failed, MDL released: %u", thd->thread_id());
+#endif /* WITH_WSREP */
     /*
         We've just done a commit, reset transaction
         isolation level and access mode to the session default.
@@ -576,6 +565,9 @@ bool Sql_cmd_xa_rollback::execute(THD *thd)
   if (!st)
   {
     thd->mdl_context.release_transactional_locks();
+#ifdef WITH_WSREP
+    WSREP_DEBUG("XA rollback failed, MDL released: %u", thd->thread_id());
+#endif /* WITH_WSREP */
     /*
       We've just done a rollback, reset transaction
       isolation level and access mode to the session default.
@@ -650,15 +642,7 @@ bool Sql_cmd_xa_start::execute(THD *thd)
 
   if (!st)
   {
-    if (thd->binlog_applier_need_detach_trx())
-    {
-      /*
-        In case of slave thread applier or processing binlog by client,
-        detach the "native" thd's trx in favor of dynamically created.
-      */
-      plugin_foreach(thd, detach_native_trx,
-                     MYSQL_STORAGE_ENGINE_PLUGIN, NULL);
-    }
+    thd->rpl_detach_engine_ha_data();
     my_ok(thd);
   }
 
@@ -739,6 +723,12 @@ bool Sql_cmd_xa_prepare::trans_xa_prepare(THD *thd)
     DBUG_ASSERT(thd->m_transaction_psi == NULL);
 #endif
 
+    /*
+      Reset rm_error in case ha_prepare() returned error,
+      so thd->transaction.xid structure gets reset
+      by THD::transaction::cleanup().
+    */
+    thd->get_transaction()->xid_state()->reset_error();
     cleanup_trans_state(thd);
     xid_state->set_state(XID_STATE::XA_NOTR);
     thd->get_transaction()->cleanup();
@@ -764,7 +754,7 @@ bool Sql_cmd_xa_prepare::execute(THD *thd)
 
   if (!st)
   {
-    if (!thd->binlog_applier_has_detached_trx() ||
+    if (!thd->rpl_unflag_detached_engine_ha_data() ||
         !(st= applier_reset_xa_trans(thd)))
       my_ok(thd);
   }
@@ -1181,6 +1171,30 @@ void transaction_cache_delete(Transaction_ctx *transaction)
 }
 
 
+/**
+  The function restores previously saved storage engine transaction context.
+
+  @param     thd     Thread context
+*/
+static void attach_native_trx(THD *thd)
+{
+  Ha_trx_info *ha_info=
+    thd->get_transaction()->ha_trx_info(Transaction_ctx::SESSION);
+  Ha_trx_info *ha_info_next;
+
+  if (ha_info)
+  {
+    for (; ha_info; ha_info= ha_info_next)
+    {
+      handlerton *hton= ha_info->ht();
+      reattach_engine_ha_data_to_thd(thd, hton);
+      ha_info_next= ha_info->next();
+      ha_info->reset();
+    }
+  }
+}
+
+
 /**
   This is a specific to "slave" applier collection of standard cleanup
   actions to reset XA transaction states at the end of XA prepare rather than
@@ -1244,39 +1258,13 @@ my_bool detach_native_trx(THD *thd, plugin_ref plugin, void *unused)
   handlerton *hton= plugin_data<handlerton *>(plugin);
 
   if (hton->replace_native_transaction_in_thd)
+  {
+    /* Ensure any active backup engine ha_data won't be overwritten */
+    DBUG_ASSERT(!thd->ha_data[hton->slot].ha_ptr_backup);
+
     hton->replace_native_transaction_in_thd(thd, NULL,
                                             thd_ha_data_backup(thd, hton));
+  }
 
   return FALSE;
 }
-
-/**
-  The function restores previously saved storage engine transaction context.
-
-  @param     thd     Thread context
-*/
-void attach_native_trx(THD *thd)
-{
-  Ha_trx_info *ha_info=
-    thd->get_transaction()->ha_trx_info(Transaction_ctx::SESSION);
-  Ha_trx_info *ha_info_next;
-
-  if (ha_info)
-  {
-    for (; ha_info; ha_info= ha_info_next)
-    {
-      handlerton *hton= ha_info->ht();
-      if (hton->replace_native_transaction_in_thd)
-      {
-        /* restore the saved original engine transaction's link with thd */
-        void **trx_backup= thd_ha_data_backup(thd, hton);
-
-        hton->
-          replace_native_transaction_in_thd(thd, *trx_backup, NULL);
-        *trx_backup= NULL;
-      }
-      ha_info_next= ha_info->next();
-      ha_info->reset();
-    }
-  }
-}
