@@ -1,6 +1,6 @@
 /*****************************************************************************
 
-Copyright (c) 2000, 2016, Oracle and/or its affiliates. All Rights Reserved.
+Copyright (c) 2000, 2017, Oracle and/or its affiliates. All Rights Reserved.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -158,6 +158,8 @@ public:
 
 	int index_last(uchar * buf);
 
+	bool has_gap_locks() const { return true; }
+
 	int rnd_init(bool scan);
 
 	int rnd_end();
@@ -203,8 +205,10 @@ public:
 
 	void position(uchar *record);
 
+#ifdef WL6742
+	/* Removing WL6742 as part of Bug #23046302 */
 	virtual int records(ha_rows* num_rows);
-
+#endif
 	ha_rows records_in_range(
 		uint			inx,
 		key_range*		min_key,
@@ -351,6 +355,21 @@ public:
 		bool			commit);
 	/** @} */
 
+	/** This function reads zip dict-related info from SYS_ZIP_DICT
+	and SYS_ZIP_DICT_COLS for all columns marked with
+	COLUMN_FORMAT_TYPE_COMPRESSED flag and updates
+	zip_dict_name / zip_dict_data for those which have associated
+	compression dictionaries.
+
+	@param	thd		Thread handle, used to determine whether it
+				is necessary to lock dict_sys mutex
+	@param	part_name	Full table name (including partition part).
+				Must be non-NULL only if called from
+				ha_partition.
+	*/
+	virtual void update_field_defs_with_zip_dict_info(THD* thd,
+		const char* part_name);
+
 	bool check_if_incompatible_data(
 		HA_CREATE_INFO*		info,
 		uint			table_changes);
@@ -596,12 +615,12 @@ extern "C" void wsrep_thd_set_exec_mode(THD *thd, enum wsrep_exec_mode mode);
 extern "C" void wsrep_thd_set_query_state(
 	THD *thd, enum wsrep_query_state state);
 extern "C" void wsrep_thd_set_conflict_state(
-	THD *thd, enum wsrep_conflict_state state);
+        THD *thd, bool lock, enum wsrep_conflict_state state);
 
 extern "C" void wsrep_thd_set_trx_to_replay(THD *thd, uint64 trx_id);
 
-extern "C"void wsrep_thd_LOCK(THD *thd);
-extern "C"void wsrep_thd_UNLOCK(THD *thd);
+extern "C" void wsrep_thd_LOCK(THD *thd);
+extern "C" void wsrep_thd_UNLOCK(THD *thd);
 extern "C" uint32 wsrep_thd_wsrep_rand(THD *thd);
 extern "C" time_t wsrep_thd_query_start(THD *thd);
 extern "C" my_thread_id wsrep_thd_thread_id(THD *thd);
@@ -610,6 +629,7 @@ extern "C" query_id_t wsrep_thd_query_id(THD *thd);
 extern "C" const char * wsrep_thd_query(THD *thd);
 extern "C" query_id_t wsrep_thd_wsrep_last_query_id(THD *thd);
 extern "C" void wsrep_thd_set_wsrep_last_query_id(THD *thd, query_id_t id);
+extern "C" void wsrep_thd_set_next_trx_id(THD *thd);
 extern "C" void wsrep_thd_awake(THD *thd, my_bool signal);
 #endif /* WITH_WSREP */
 struct trx_t;
@@ -1145,3 +1165,38 @@ innobase_build_index_translation(
 	INNOBASE_SHARE*		share);	/*!< in/out: share structure
 					where index translation table
 					will be constructed in. */
+
+/** Compression dictionary id container */
+typedef std::map<uint16, ulint, std::less<uint16>,
+	ut_allocator<std::pair<uint16, const ulint> > >
+	zip_dict_id_container_t;
+
+/** This function checks if all the compression dictionaries referenced
+in table->fields exist in SYS_ZIP_DICT InnoDB system table.
+@return true if all referenced dictionaries exist */
+bool
+innobase_check_zip_dicts(
+	const TABLE*	table,		/*!< in: table in MySQL data
+					dictionary */
+	zip_dict_id_container_t&
+			dict_ids,	/*!< out: identified zip dict ids */
+	trx_t*		trx,		/*!< in: transaction */
+	const char**	err_dict_name);	/*!< out: the name of the
+					zip_dict which does not exist. */
+
+/** This function creates compression dictionary references in
+SYS_ZIP_DICT_COLS InnoDB system table for table_id based on info
+in table->fields and provided zip dict ids. */
+void
+innobase_create_zip_dict_references(
+	const TABLE*	table,		/*!< in: table in MySQL data
+					dictionary */
+	table_id_t	ib_table_id,	/*!< in: table ID in Innodb data
+					dictionary */
+	const zip_dict_id_container_t&
+			dict_ids,	/*!< in: zip dict ids */
+	trx_t*		trx);		/*!< in: transaction */
+
+/** Free InnoDB session specific data.
+@param[in,out]	thd	MySQL thread handler. */
+void thd_free_innodb_session(THD* thd);
