@@ -1,7 +1,7 @@
 /***********************************************************************
 
-Copyright (c) 1995, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2009, 2016, Percona Inc.
+Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
+Copyright (c) 2009, 2017, Percona Inc.
 
 Portions of this file contain modifications contributed and copyrighted
 by Percona Inc.. Those modifications are
@@ -82,8 +82,6 @@ the OS actually supports it: Win 95 does not, NT does. */
 /** File handle */
 # define os_file_t	HANDLE
 
-# define os_file_invalid	INVALID_HANDLE_VALUE
-
 /** Convert a C file descriptor to a native file handle
 @param fd file descriptor
 @return native file handle */
@@ -98,8 +96,6 @@ typedef DIR*	os_file_dir_t;	/*!< directory stream */
 /** File handle */
 typedef int	os_file_t;
 
-# define os_file_invalid	(-1)
-
 /** Convert a C file descriptor to a native file handle
 @param fd file descriptor
 @return native file handle */
@@ -109,6 +105,24 @@ typedef int	os_file_t;
 
 static const os_file_t OS_FILE_CLOSED = os_file_t(~0);
 
+/** Common file descriptor for file IO instrumentation with PFS
+on windows and other platforms */
+struct pfs_os_file_t
+{
+	os_file_t   m_file;
+#ifdef UNIV_PFS_IO
+	struct PSI_file *m_psi;
+#endif
+
+	bool is_closed() const {
+		return m_file == OS_FILE_CLOSED;
+	}
+
+	void set_closed() {
+		m_file = OS_FILE_CLOSED;
+	}
+};
+
 /** The next value should be smaller or equal to the smallest sector size used
 on any disk. A log block is required to be a portion of disk which is written
 so that if the start and the end of a block get written to disk, then the
@@ -142,6 +156,21 @@ enum os_file_create_t {
 	OS_FILE_O_SYNC = 512		/*!< Open file with O_SYNC */
 };
 
+/** Options for os_file_advise_func @{ */
+enum os_file_advise_t {
+	OS_FILE_ADVISE_NORMAL = 1,      /*!< no advice on access pattern
+					(default) */
+	OS_FILE_ADVISE_RANDOM = 2,      /*!< access in random order */
+	OS_FILE_ADVISE_SEQUENTIAL = 4,  /*!< access the specified data
+					sequentially (with lower offsets read
+					before higher ones) */
+	OS_FILE_ADVISE_WILLNEED = 8,    /*!< specified data will be accessed
+					in the near future */
+	OS_FILE_ADVISE_DONTNEED = 16,   /*!< specified data will not be
+					accessed in the near future */
+	OS_FILE_ADVISE_NOREUSE = 32     /*!< access only once */
+};
+
 static const ulint OS_FILE_READ_ONLY = 333;
 static const ulint OS_FILE_READ_WRITE = 444;
 
@@ -996,7 +1025,7 @@ A simple function to open or create a file.
 @param[out]	success		true if succeed, false if error
 @return own: handle to the file, not defined if error, error number
 	can be retrieved with os_file_get_last_error */
-os_file_t
+pfs_os_file_t
 os_file_create_simple_func(
 	const char*	name,
 	ulint		create_mode,
@@ -1016,7 +1045,7 @@ A simple function to open or create a file.
 @param[out]	success		true if succeeded
 @return own: handle to the file, not defined if error, error number
 	can be retrieved with os_file_get_last_error */
-os_file_t
+pfs_os_file_t
 os_file_create_simple_no_error_handling_func(
 	const char*	name,
 	ulint		create_mode,
@@ -1029,13 +1058,27 @@ os_file_create_simple_no_error_handling_func(
 @param[in]	fd		file descriptor to alter
 @param[in]	file_name	file name, used in the diagnostic message
 @param[in]	name		"open" or "create"; used in the diagnostic
-				message */
-void
+				message
+@return true if operation is success and false */
+bool
 os_file_set_nocache(
 	int		fd,
 	const char*	file_name,
 	const char*	operation_name);
 
+/** Tries to disable OS caching on an opened file file.
+@param[in]	file		file to alter
+@param[in]	file_name	file name, used in the diagnostic message
+@param[in]	name		"open" or "create"; used in the diagnostic
+message
+@return true if operation is success and false */
+UNIV_INLINE
+bool
+os_file_set_nocache(
+	pfs_os_file_t	file,
+	const char*	file_name,
+	const char*	operation_name);
+
 /** NOTE! Use the corresponding macro os_file_create(), not directly
 this function!
 Opens an existing file or creates a new.
@@ -1053,7 +1096,7 @@ Opens an existing file or creates a new.
 @param[in]	success		true if succeeded
 @return own: handle to the file, not defined if error, error number
 	can be retrieved with os_file_get_last_error */
-os_file_t
+pfs_os_file_t
 os_file_create_func(
 	const char*	name,
 	ulint		create_mode,
@@ -1095,6 +1138,14 @@ os_file_get_last_error.
 bool
 os_file_close_func(os_file_t file);
 
+/** NOTE! Use the corresponding macro os_file_close_no_error_handling(), not
+directly this function!
+Closes a file handle.
+@param[in] file		handle to a file
+@return true if success */
+bool
+os_file_close_no_error_handling_func(os_file_t file);
+
 #ifdef UNIV_PFS_IO
 
 /* Keys to register InnoDB I/O with performance schema */
@@ -1108,7 +1159,9 @@ extern mysql_pfs_key_t	innodb_parallel_dblwrite_file_key;
 various file I/O operations with performance schema.
 1) register_pfs_file_open_begin() and register_pfs_file_open_end() are
 used to register file creation, opening, closing and renaming.
-2) register_pfs_file_io_begin() and register_pfs_file_io_end() are
+2) register_pfs_file_rename_begin() and  register_pfs_file_rename_end()
+are used to register file renaming
+3) register_pfs_file_io_begin() and register_pfs_file_io_end() are
 used to register actual file read, write and flush
 3) register_pfs_file_close_begin() and register_pfs_file_close_end()
 are used to register file deletion operations*/
@@ -1116,26 +1169,41 @@ are used to register file deletion operations*/
 				      src_file, src_line)		\
 do {									\
 	locker = PSI_FILE_CALL(get_thread_file_name_locker)(		\
-		state, key, op, name, &locker);				\
+		state, key.m_value, op, name, &locker);			\
 	if (locker != NULL) {						\
 		PSI_FILE_CALL(start_file_open_wait)(			\
 			locker, src_file, src_line);			\
 	}								\
 } while (0)
 
-# define register_pfs_file_open_end(locker, file)			\
+# define register_pfs_file_open_end(locker, file, result)		\
 do {									\
-	if (locker != NULL) {						\
-		PSI_FILE_CALL(end_file_open_wait_and_bind_to_descriptor)(\
-			locker, file);					\
+	if (locker != NULL) {				\
+		file.m_psi = PSI_FILE_CALL(				\
+		end_file_open_wait)(					\
+			locker, result);				\
 	}								\
 } while (0)
 
+# define register_pfs_file_rename_begin(state, locker, key, op, name,	\
+				src_file, src_line)                     \
+	register_pfs_file_open_begin(state, locker, key, op, name,      \
+					src_file, src_line)             \
+
+# define register_pfs_file_rename_end(locker, result)			\
+do {									\
+	if (locker != NULL) {                              \
+		 PSI_FILE_CALL(						\
+			end_file_open_wait)(				\
+			locker, result);				\
+	}								\
+}while(0)
+
 # define register_pfs_file_close_begin(state, locker, key, op, name,	\
 				      src_file, src_line)		\
 do {									\
 	locker = PSI_FILE_CALL(get_thread_file_name_locker)(		\
-		state, key, op, name, &locker);				\
+		state, key.m_value, op, name, &locker);			\
 	if (locker != NULL) {						\
 		PSI_FILE_CALL(start_file_close_wait)(			\
 			locker, src_file, src_line);			\
@@ -1153,8 +1221,8 @@ do {									\
 # define register_pfs_file_io_begin(state, locker, file, count, op,	\
 				    src_file, src_line)			\
 do {									\
-	locker = PSI_FILE_CALL(get_thread_file_descriptor_locker)(	\
-		state, file, op);					\
+	locker = PSI_FILE_CALL(get_thread_file_stream_locker)(	\
+		state, file.m_psi, op);					\
 	if (locker != NULL) {						\
 		PSI_FILE_CALL(start_file_wait)(				\
 			locker, count, src_file, src_line);		\
@@ -1176,11 +1244,14 @@ os_file_create
 os_file_create_simple
 os_file_create_simple_no_error_handling
 os_file_close
+os_file_close_no_error_handling
 os_file_rename
 os_aio
 os_file_read
 os_file_read_no_error_handling
+os_file_read_no_error_handling_int_fd
 os_file_write
+os_file_set_eof_at
 
 The wrapper functions have the prefix of "innodb_". */
 
@@ -1200,32 +1271,44 @@ The wrapper functions have the prefix of "innodb_". */
 		key, name, create_mode, access,				\
 		read_only, success, __FILE__, __LINE__)
 
-# define os_file_close(file)						\
+# define os_file_close_pfs(file)						\
 	pfs_os_file_close_func(file, __FILE__, __LINE__)
 
+# define os_file_close_no_error_handling_pfs(file)			\
+	pfs_os_file_close_no_error_handling_func(file, __FILE__, __LINE__)
+
 # define os_aio(type, mode, name, file, buf, offset,			\
 		n, read_only, message1, message2, space_id, trx)	\
 	pfs_os_aio_func(type, mode, name, file, buf, offset,		\
 			n, read_only, message1, message2, space_id,	\
 			trx, __FILE__, __LINE__)
 
-# define os_file_read(type, file, buf, offset, n)			\
+# define os_file_read_pfs(type, file, buf, offset, n)			\
 	pfs_os_file_read_func(type, file, buf, offset, n, NULL,		\
 			      __FILE__, __LINE__)
 
-# define os_file_read_trx(file, buf, offset, n, trx)			\
+# define os_file_read_trx_pfs(file, buf, offset, n, trx)		\
 	pfs_os_file_read_func(file, buf, offset, n, trx,		\
 			      __FILE__, __LINE__)
 
-# define os_file_read_no_error_handling(type, file, buf, offset, n, o)	\
+# define os_file_read_no_error_handling_pfs(type, file, buf, offset, n, o)	\
 	pfs_os_file_read_no_error_handling_func(			\
 		type, file, buf, offset, n, o, __FILE__, __LINE__)
 
-# define os_file_write(type, name, file, buf, offset, n)	\
+# define os_file_read_no_error_handling_int_fd(                         \
+		type, file, buf, offset, n, o)				\
+	pfs_os_file_read_no_error_handling_int_fd_func(                 \
+		type, file, buf, offset, n, o, __FILE__, __LINE__)
+
+# define os_file_write_pfs(type, name, file, buf, offset, n)	\
 	pfs_os_file_write_func(type, name, file, buf, offset,	\
 			       n, __FILE__, __LINE__)
 
-# define os_file_flush(file)						\
+# define os_file_write_int_fd(type, name, file, buf, offset, n)		\
+	pfs_os_file_write_int_fd_func(type, name, file, buf, offset,	\
+				n, __FILE__, __LINE__)
+
+# define os_file_flush_pfs(file)						\
 	pfs_os_file_flush_func(file, __FILE__, __LINE__)
 
 # define os_file_rename(key, oldpath, newpath)				\
@@ -1237,6 +1320,10 @@ The wrapper functions have the prefix of "innodb_". */
 # define os_file_delete_if_exists(key, name, exist)			\
 	pfs_os_file_delete_if_exists_func(key, name, exist, __FILE__, __LINE__)
 
+# define os_file_set_eof_at_pfs(file, new_len)				\
+	pfs_os_file_set_eof_at_func(file, new_len, __FILE__, __LINE__)
+
+
 /** NOTE! Please use the corresponding macro os_file_create_simple(),
 not directly this function!
 A performance schema instrumented wrapper function for
@@ -1253,7 +1340,7 @@ os_file_create_simple() which opens or creates a file.
 @return own: handle to the file, not defined if error, error number
 	can be retrieved with os_file_get_last_error */
 UNIV_INLINE
-os_file_t
+pfs_os_file_t
 pfs_os_file_create_simple_func(
 	mysql_pfs_key_t key,
 	const char*	name,
@@ -1284,7 +1371,7 @@ monitor file creation/open.
 @return own: handle to the file, not defined if error, error number
 	can be retrieved with os_file_get_last_error */
 UNIV_INLINE
-os_file_t
+pfs_os_file_t
 pfs_os_file_create_simple_no_error_handling_func(
 	mysql_pfs_key_t key,
 	const char*	name,
@@ -1317,7 +1404,7 @@ Add instrumentation to monitor file creation/open.
 @return own: handle to the file, not defined if error, error number
 	can be retrieved with os_file_get_last_error */
 UNIV_INLINE
-os_file_t
+pfs_os_file_t
 pfs_os_file_create_func(
 	mysql_pfs_key_t key,
 	const char*	name,
@@ -1340,7 +1427,21 @@ A performance schema instrumented wrapper function for os_file_close().
 UNIV_INLINE
 bool
 pfs_os_file_close_func(
-	os_file_t	file,
+	pfs_os_file_t	file,
+	const char*	src_file,
+	ulint		src_line);
+
+/** NOTE! Use the corresponding macro os_file_close_no_error_handling(), not
+directly this function!
+Closes a file handle.
+@param[in] file		handle to a file
+@param[in]	src_file	file name where func invoked
+@param[in]	src_line	line where the func invoked
+@return true if success */
+UNIV_INLINE
+bool
+pfs_os_file_close_no_error_handling_func(
+	pfs_os_file_t	file,
 	const char*	src_file,
 	ulint		src_line);
 
@@ -1360,7 +1461,7 @@ UNIV_INLINE
 dberr_t
 pfs_os_file_read_func(
 	IORequest&	type,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	void*		buf,
 	os_offset_t	offset,
 	ulint		n,
@@ -1386,7 +1487,7 @@ UNIV_INLINE
 dberr_t
 pfs_os_file_read_no_error_handling_func(
 	IORequest&	type,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	void*		buf,
 	os_offset_t	offset,
 	ulint		n,
@@ -1394,6 +1495,33 @@ pfs_os_file_read_no_error_handling_func(
 	const char*	src_file,
 	ulint		src_line);
 
+/** NOTE! Please use the corresponding macro
+os_file_read_no_error_handling_int_fd(), not directly this function!
+This is the performance schema instrumented wrapper function for
+os_file_read_no_error_handling_int_fd_func() which requests a
+synchronous  read operation on files with int type descriptors.
+@param[in, out] type            IO request context
+@param[in]      file            Open file handle
+@param[out]     buf             buffer where to read
+@param[in]      offset          file offset where to read
+@param[in]      n               number of bytes to read
+@param[out]     o               number of bytes actually read
+@param[in]      src_file        file name where func invoked
+@param[in]      src_line        line where the func invoked
+@return DB_SUCCESS if request was successful */
+
+UNIV_INLINE
+dberr_t
+pfs_os_file_read_no_error_handling_int_fd_func(
+        IORequest&      type,
+        int             file,
+        void*           buf,
+        os_offset_t     offset,
+        ulint           n,
+        ulint*          o,
+        const char*     src_file,
+        ulint           src_line);
+
 /** NOTE! Please use the corresponding macro os_aio(), not directly this
 function!
 Performance schema wrapper function of os_aio() which requests
@@ -1422,7 +1550,7 @@ pfs_os_aio_func(
 	IORequest&	type,
 	ulint		mode,
 	const char*	name,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	void*		buf,
 	os_offset_t	offset,
 	ulint		n,
@@ -1453,13 +1581,40 @@ dberr_t
 pfs_os_file_write_func(
 	IORequest&	type,
 	const char*	name,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	const void*	buf,
 	os_offset_t	offset,
 	ulint		n,
 	const char*	src_file,
 	ulint		src_line);
 
+/** NOTE! Please use the corresponding macro os_file_write(), not
+directly this function!
+This is the performance schema instrumented wrapper function for
+os_file_write() which requests a synchronous write operation
+on files with int type descriptors.
+@param[in, out] type            IO request context
+@param[in]      name            Name of the file or path as NUL terminated
+				string
+@param[in]      file            Open file handle
+@param[out]     buf             buffer where to read
+@param[in]      offset          file offset where to read
+@param[in]      n		number of bytes to read
+@param[in]      src_file        file name where func invoked
+@param[in]      src_line        line where the func invoked
+@return DB_SUCCESS if request was successful */
+UNIV_INLINE
+dberr_t
+pfs_os_file_write_int_fd_func(
+        IORequest&      type,
+        const char*     name,
+        int		file,
+        const void*     buf,
+        os_offset_t     offset,
+        ulint           n,
+        const char*     src_file,
+        ulint           src_line);
+
 /** NOTE! Please use the corresponding macro os_file_flush(), not directly
 this function!
 This is the performance schema instrumented wrapper function for
@@ -1472,7 +1627,7 @@ Flushes the write buffers of a given file to the disk.
 UNIV_INLINE
 bool
 pfs_os_file_flush_func(
-	os_file_t	file,
+	pfs_os_file_t	file,
 	const char*	src_file,
 	ulint		src_line);
 
@@ -1533,6 +1688,22 @@ pfs_os_file_delete_if_exists_func(
 	const char*	src_file,
 	ulint		src_line);
 
+/** NOTE! Use the corresponding macro os_file_flush(), not directly this
+function!
+Truncates a file at the specified position.
+@param[in]	file	file to truncate
+@param[in]	new_len	new file length
+@param[in]	src_file	file name where func invoked
+@param[in]	src_line	line where the func invoked
+@return true if success */
+UNIV_INLINE
+bool
+pfs_os_file_set_eof_at_func(
+	pfs_os_file_t	file,
+	ib_uint64_t	new_len,
+	const char*	src_file,
+	ulint		src_line);
+
 #else /* UNIV_PFS_IO */
 
 /* If UNIV_PFS_IO is not defined, these I/O APIs point
@@ -1552,26 +1723,35 @@ to original un-instrumented file I/O APIs */
 	os_file_create_simple_no_error_handling_func(			\
 		name, create_mode, access, read_only, success)
 
-# define os_file_close(file)	os_file_close_func(file)
+# define os_file_close_pfs(file)	os_file_close_func(file)
+
+# define os_file_close_no_error_handling_pfs(file)	\
+	os_file_close_no_error_handling_func(file)
 
 # define os_aio(type, mode, name, file, buf, offset,			\
 		n, read_only, message1, message2, space_id, trx)	\
 	os_aio_func(type, mode, name, file, buf, offset,		\
 		n, read_only, message1, message2, space_id, trx)
 
-# define os_file_read(type, file, buf, offset, n)			\
+# define os_file_read_pfs(type, file, buf, offset, n)			\
 	os_file_read_func(type, file, buf, offset, n)
 
-# define os_file_read_no_error_handling(type, file, buf, offset, n, o)	\
+# define os_file_read_no_error_handling_pfs(type, file, buf, offset, n, o)	\
 	os_file_read_no_error_handling_func(type, file, buf, offset, n, o)
 
-# define os_file_read_trx(file, buf, offset, n, trx)	\
+# define os_file_read_no_error_handling_int_fd(type, file, buf, offset, n, o)  \
+	os_file_read_no_error_handling_func(type, file, buf, offset, n, o)
+
+# define os_file_read_trx_pfs(file, buf, offset, n, trx)	\
 	os_file_read_func(file, buf, offset, n, trx)
 
-# define os_file_write(type, name, file, buf, offset, n)		\
+# define os_file_write_pfs(type, name, file, buf, offset, n)		\
 	os_file_write_func(type, name, file, buf, offset, n)
 
-# define os_file_flush(file)	os_file_flush_func(file)
+# define os_file_write_int_fd(type, name, file, buf, offset, n)            \
+	os_file_write_func(type, name, file, buf, offset, n)
+
+# define os_file_flush_pfs(file)	os_file_flush_func(file)
 
 # define os_file_rename(key, oldpath, newpath)				\
 	os_file_rename_func(oldpath, newpath)
@@ -1581,13 +1761,77 @@ to original un-instrumented file I/O APIs */
 # define os_file_delete_if_exists(key, name, exist)			\
 	os_file_delete_if_exists_func(name, exist)
 
+# define os_file_set_eof_at_pfs(file, new_len)				\
+	os_file_set_eof_at_func(file, new_len)
+
 #endif	/* UNIV_PFS_IO */
 
-/** Closes a file handle.
-@param[in] file		handle to a file
+#ifdef UNIV_PFS_IO
+	#define os_file_close(file) os_file_close_pfs(file)
+#else
+	#define os_file_close(file) os_file_close_pfs((file).m_file)
+#endif
+
+#ifdef UNIV_PFS_IO
+	#define os_file_close_no_error_handling(file)	\
+			os_file_close_no_error_handling_pfs(file)
+#else
+	#define os_file_close_no_error_handling(file)	\
+			os_file_close_no_error_handling_pfs((file).m_file)
+#endif
+
+#ifdef UNIV_PFS_IO
+	#define os_file_read(type, file, buf, offset, n)                \
+		os_file_read_pfs(type, file, buf, offset, n)
+#else
+	#define os_file_read(type, file, buf, offset, n)                \
+                os_file_read_pfs(type, file.m_file, buf, offset, n)
+#endif
+
+#ifdef UNIV_PFS_IO
+	#define os_file_flush(file)	os_file_flush_pfs(file)
+#else
+	#define os_file_flush(file)	os_file_flush_pfs(file.m_file)
+#endif
+
+#ifdef UNIV_PFS_IO
+	#define os_file_write(type, name, file, buf, offset, n)         \
+		os_file_write_pfs(type, name, file, buf, offset, n)
+#else
+	#define os_file_write(type, name, file, buf, offset, n)         \
+                os_file_write_pfs(type, name, file.m_file, buf, offset, n)
+#endif
+
+#ifdef UNIV_PFS_IO
+	#define os_file_read_no_error_handling(type, file, buf, offset, n, o)  \
+		 os_file_read_no_error_handling_pfs(type, file, buf, offset, n, o)
+#else
+	#define os_file_read_no_error_handling(type, file, buf, offset, n, o) \
+                 os_file_read_no_error_handling_pfs(			      \
+			type, file.m_file, buf, offset, n, o)
+#endif
+
+#ifdef UNIV_PFS_IO
+	#define os_file_set_eof_at(file, new_len)	\
+		 os_file_set_eof_at_pfs(file, new_len)
+#else
+	#define os_file_set_eof_at(file, new_len)	\
+		 os_file_set_eof_at_pfs(file.m_file, new_len)
+#endif
+
+/** Announces an intention to access file data in a specific pattern in the
+future.
+@param[in, own]	file	handle to a file
+@param[in]	offset	file region offset
+@param[in]	len	file region length
+@param[in]	advice	advice for access pattern
 @return true if success */
 bool
-os_file_close_no_error_handling(os_file_t file);
+os_file_advise(
+	pfs_os_file_t   file,   /*!< in, own: handle to a file */
+	os_offset_t     offset, /*!< in: file region offset  */
+	os_offset_t     len,    /*!< in: file region length  */
+	ulint		advice);/*!< in: advice for access pattern */
 
 /** Gets a file size.
 @param[in]	file		handle to a file
@@ -1603,7 +1847,7 @@ os_file_get_size(
 @return file size, or (os_offset_t) -1 on failure */
 os_offset_t
 os_file_get_size(
-	os_file_t	file)
+	pfs_os_file_t	file)
 	MY_ATTRIBUTE((warn_unused_result));
 
 /** Write the specified number of zeros to a newly created file.
@@ -1616,7 +1860,7 @@ os_file_get_size(
 bool
 os_file_set_size(
 	const char*	name,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	os_offset_t	size,
 	bool		read_only)
 	MY_ATTRIBUTE((warn_unused_result));
@@ -1637,17 +1881,19 @@ preserved is smaller or equal than current size of file.
 bool
 os_file_truncate(
 	const char*	pathname,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	os_offset_t	size);
 
-/***********************************************************************//**
+/** NOTE! Use the corresponding macro os_file_flush(), not directly this
+function!
 Truncates a file at the specified position.
+@param[in]	file	file to truncate
+@param[in]	new_len	new file length
 @return true if success */
-
 bool
-os_file_set_eof_at(
-	os_file_t	file,	/*!< in: handle to a file */
-	ib_uint64_t	new_len);/*!< in: new file length */
+os_file_set_eof_at_func(
+	os_file_t	file,
+	ib_uint64_t	new_len);
 
 /** NOTE! Use the corresponding macro os_file_flush(), not directly this
 function!
@@ -1841,7 +2087,7 @@ os_aio_func(
 	IORequest&	type,
 	ulint		mode,
 	const char*	name,
-	os_file_t	file,
+	pfs_os_file_t	file,
 	void*		buf,
 	os_offset_t	offset,
 	ulint		n,
@@ -1994,7 +2240,7 @@ Note: On Windows we use the name and on Unices we use the file handle.
 bool
 os_is_sparse_file_supported(
 	const char*	path,
-	os_file_t	fh)
+	pfs_os_file_t	fh)
 	MY_ATTRIBUTE((warn_unused_result));
 
 /** Decompress the page data contents. Page type must be FIL_PAGE_COMPRESSED, if
