@@ -167,6 +167,9 @@ public:
                  ha_partition *clone_arg,
                  MEM_ROOT *clone_mem_root_arg);
    ~ha_partition();
+
+   bool init_with_fields();
+
   /*
     A partition handler has no characteristics in itself. It only inherits
     those from the underlying handlers. Here we set-up those constants to
@@ -205,6 +208,15 @@ public:
                                                copied,
                                                deleted);
   }
+
+  /** This function reads zip dict-related info from partition handlers.
+  It may do nothing if individual handlers do not support COMPRESSED_COLUMNS.
+
+  @param    thd          Thread handler
+  @param    part_name    Must be always NULL.
+  */
+  virtual void update_field_defs_with_zip_dict_info(THD* thd,
+                                                    const char* part_name);
 private:
   bool get_num_parts(const char *name, uint *num_parts)
   {
@@ -345,11 +357,12 @@ public:
   }
   int update_row(const uchar *old_data, uchar *new_data)
   {
-    return Partition_helper::ph_update_row(old_data, new_data);
+    return Partition_helper::ph_update_row(old_data, new_data,
+                                           rpl_lookup_rows());
   }
   int delete_row(const uchar *buf)
   {
-    return Partition_helper::ph_delete_row(buf);
+    return Partition_helper::ph_delete_row(buf, rpl_lookup_rows());
   }
   virtual int delete_all_rows(void);
   virtual int truncate();
@@ -560,6 +573,7 @@ public:
   {
     return Partition_helper::ph_read_range_next();
   }
+  bool has_gap_locks() const;
 public:
   /*
     -------------------------------------------------------------------------
@@ -1134,6 +1148,15 @@ public:
     virtual void append_create_info(String *packet)
   */
 
+  /* For TokuDB Read Free Replication */
+    void rpl_before_write_rows();
+    void rpl_after_write_rows();
+    void rpl_before_delete_rows();
+    void rpl_after_delete_rows();
+    void rpl_before_update_rows();
+    void rpl_after_update_rows();
+    bool rpl_lookup_rows();
+
   /*
     -------------------------------------------------------------------------
     MODULE partitioning specific handler API
@@ -1156,6 +1179,9 @@ public:
     return (HA_PARTITION_FUNCTION_SUPPORTED |
             HA_FAST_CHANGE_PARTITION);
   }
+#ifdef WITH_WSREP
+  bool wsrep_is_innodb() { return m_innodb; }
+#endif /* WITH_WSREP */
 
 #ifdef WITH_WSREP
     virtual int wsrep_db_type() const;
