@@ -1,25 +1,13 @@
-/* Copyright (c) 2000, 2019, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2000, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License, version 2.0,
-   as published by the Free Software Foundation.
-
-   This program is also distributed with certain software (including
-   but not limited to OpenSSL) that is licensed under separate terms,
-   as designated in a particular file or component or in included license
-   documentation.  The authors of MySQL hereby grant you an additional
-   permission to link the program and your derivative works with the
-   separately licensed software that they have included with MySQL.
-
-   Without limiting anything contained in the foregoing, this file,
-   which is part of C Driver for MySQL (Connector/C), is also subject to the
-   Universal FOSS Exception, version 1.0, a copy of which can be found at
-   http://oss.oracle.com/licenses/universal-foss-exception.
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License, version 2.0, for more details.
+   GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -159,8 +147,10 @@ static my_bool ssl_should_retry(Vio *vio, int ret,
     /* Note: the OpenSSL error queue gets cleared in report_errors(). */
     report_errors(ssl);
 #else             /* Release build */
+# ifndef HAVE_YASSL
     /* OpenSSL: clear the error queue. */
     ERR_clear_error();
+# endif
 #endif
     should_retry= FALSE;
     ssl_set_sys_error(ssl_error);
@@ -183,14 +173,16 @@ size_t vio_ssl_read(Vio *vio, uchar *buf, size_t size)
   {
     enum enum_vio_io_event event;
 
+#ifndef HAVE_YASSL
     /*
       OpenSSL: check that the SSL thread's error queue is cleared. Otherwise
       SSL_read() returns an error from the error queue, when SSL_read() failed
       because it would block.
     */
     DBUG_ASSERT(ERR_peek_error() == 0);
+#endif
 
-    ret= SSL_read(ssl, buf, size);
+    ret= SSL_read(ssl, buf, (int)size);
 
     if (ret >= 0)
       break;
@@ -220,14 +212,16 @@ size_t vio_ssl_write(Vio *vio, const uchar *buf, size_t size)
   {
     enum enum_vio_io_event event;
 
+#ifndef HAVE_YASSL
     /*
       OpenSSL: check that the SSL thread's error queue is cleared. Otherwise
       SSL_write() returns an error from the error queue, when SSL_write() failed
       because it would block.
     */
     DBUG_ASSERT(ERR_peek_error() == 0);
+#endif
 
-    ret= SSL_write(ssl, buf, size);
+    ret= SSL_write(ssl, buf, (int)size);
 
     if (ret >= 0)
       break;
@@ -244,6 +238,22 @@ size_t vio_ssl_write(Vio *vio, const uchar *buf, size_t size)
   DBUG_RETURN(ret < 0 ? -1 : ret);
 }
 
+#ifdef HAVE_YASSL
+
+/* Emulate a blocking recv() call with vio_read(). */
+static long yassl_recv(void *ptr, void *buf, size_t len)
+{
+  return (long)vio_read(ptr, buf, len);
+}
+
+
+/* Emulate a blocking send() call with vio_write(). */
+static long yassl_send(void *ptr, const void *buf, size_t len)
+{
+  return (long)vio_write(ptr, buf, len);
+}
+
+#endif
 
 int vio_ssl_shutdown(Vio *vio, int how)
 {
@@ -298,6 +308,10 @@ void vio_ssl_delete(Vio *vio)
     vio->ssl_arg= 0;
   }
 
+#ifndef HAVE_YASSL
+  ERR_remove_thread_state(0);
+#endif
+
   vio_delete(vio);
 }
 
@@ -330,12 +344,14 @@ static int ssl_handshake_loop(Vio *vio, SSL *ssl,
   {
     enum enum_vio_io_event event;
 
+#ifndef HAVE_YASSL
     /*
       OpenSSL: check that the SSL thread's error queue is cleared. Otherwise
       SSL-handshake-function returns an error from the error queue, when the
       function failed because it would block.
     */
     DBUG_ASSERT(ERR_peek_error() == 0);
+#endif
 
     ret= func(ssl);
 
@@ -357,8 +373,7 @@ static int ssl_handshake_loop(Vio *vio, SSL *ssl,
 }
 
 
-static int ssl_do(struct st_VioSSLFd *ptr, Vio *vio,
-                  long timeout MY_ATTRIBUTE((unused)),
+static int ssl_do(struct st_VioSSLFd *ptr, Vio *vio, long timeout,
                   ssl_handshake_func_t func,
                   unsigned long *ssl_errno_holder)
 {
@@ -367,7 +382,7 @@ static int ssl_do(struct st_VioSSLFd *ptr, Vio *vio,
   my_socket sd= mysql_socket_getfd(vio->mysql_socket);
 
   /* Declared here to make compiler happy */
-#if !defined(DBUG_OFF)
+#if !defined(HAVE_YASSL) && !defined(DBUG_OFF)
   int j, n;
 #endif
 
@@ -383,14 +398,15 @@ static int ssl_do(struct st_VioSSLFd *ptr, Vio *vio,
   }
   DBUG_PRINT("info", ("ssl: 0x%lx timeout: %ld", (long) ssl, timeout));
   SSL_clear(ssl);
+  SSL_SESSION_set_timeout(SSL_get_session(ssl), timeout);
   SSL_set_fd(ssl, sd);
-#if defined(SSL_OP_NO_COMPRESSION)
+#if !defined(HAVE_YASSL) && defined(SSL_OP_NO_COMPRESSION)
   SSL_set_options(ssl, SSL_OP_NO_COMPRESSION); /* OpenSSL >= 1.0 only */
 #elif OPENSSL_VERSION_NUMBER >= 0x00908000L /* workaround for OpenSSL 0.9.8 */
   sk_SSL_COMP_zero(SSL_COMP_get_compression_methods());
 #endif
 
-#if !defined(DBUG_OFF)
+#if !defined(HAVE_YASSL) && !defined(DBUG_OFF)
   {
     STACK_OF(SSL_COMP) *ssl_comp_methods = NULL;
     ssl_comp_methods = SSL_COMP_get_compression_methods();
@@ -402,15 +418,25 @@ static int ssl_do(struct st_VioSSLFd *ptr, Vio *vio,
       for (j = 0; j < n; j++)
       {
         SSL_COMP *c = sk_SSL_COMP_value(ssl_comp_methods, j);
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
         DBUG_PRINT("info", ("  %d: %s\n", c->id, c->name));
-#else /* OPENSSL_VERSION_NUMBER < 0x10100000L */
-        DBUG_PRINT("info", ("  %d: %s\n", SSL_COMP_get_id(c), SSL_COMP_get0_name(c)));
-#endif /* OPENSSL_VERSION_NUMBER < 0x10100000L */
       }
   }
 #endif
 
+  /*
+    Since yaSSL does not support non-blocking send operations, use
+    special transport functions that properly handles non-blocking
+    sockets. These functions emulate the behavior of blocking I/O
+    operations by waiting for I/O to become available.
+  */
+#ifdef HAVE_YASSL
+  /* Set first argument of the transport functions. */
+  yaSSL_transport_set_ptr(ssl, vio);
+  /* Set functions to use in order to send and receive data. */
+  yaSSL_transport_set_recv_function(ssl, yassl_recv);
+  yaSSL_transport_set_send_function(ssl, yassl_send);
+#endif
+
   if ((r= ssl_handshake_loop(vio, ssl, func, ssl_errno_holder)) < 1)
   {
     DBUG_PRINT("error", ("SSL_connect/accept failure"));
